#ifndef _ASM_XEN_HYPERCALL_H
#define _ASM_XEN_HYPERCALL_H

#include "mipsregs.h"
#include "mips-exts.h"
#include "xen/mm.h"
#include "xen/interface/xen.h"
#include "asm/xen/hypervisor.h"

// extern unsigned long hypercall_page[];
extern struct { unsigned long _entry; } hypercall_page[];

extern struct shared_info *xen_shared_info;

extern unsigned long gpctx[NCPU];

#define __get_cpu_var(x)  (x[processor_id()])

#define load_gp(var) \
({ \
    __asm__ __volatile__ ( \
        ".set push\n" \
        ".set noreorder\n" \
        "move %0, $28\n" \
        ".set pop\n" \
        :"=r"(var) \
        ); \
})

#define store_gp(var) \
({ \
    __asm__ __volatile__ ( \
        ".set push\n" \
        ".set noreorder\n" \
        "move $28, %0\n" \
        ".set pop\n" \
        : \
        : "r"(var) \
        ); \
})

#define switch_xen_linux_stacks() \
({ \
        load_gp(__get_cpu_var(gpctx)); \
        store_gp((*(xen_shared_info->arch.xensp[processor_id()])) & ~(XEN_PAGE_SIZE - 1)); \
})

#define restore_linux_stack() store_gp(__get_cpu_var(gpctx))


#define __HYPERCALL                                                     \
        ".set\tpush\n.set\tnoreorder\njalr\t%[func]\nnop\n.set\tpop\n"
#define __HYPERCALL_ENTRY(x)                                            \
        [func] "r" ((unsigned long)(hypercall_page[__HYPERVISOR_##x])._entry)

#define __HYPERCALL_RETREG      "$2"
#define __HYPERCALL_ARG1REG     "$4"
#define __HYPERCALL_ARG2REG     "$5"
#define __HYPERCALL_ARG3REG     "$6"
#define __HYPERCALL_ARG4REG     "$7"
#define __HYPERCALL_ARG5REG     "$8"

#define __HYPERCALL_DECLS                                               \
        register unsigned long __res  asm(__HYPERCALL_RETREG);          \
        register unsigned long __arg1 asm(__HYPERCALL_ARG1REG) = __arg1; \
        register unsigned long __arg2 asm(__HYPERCALL_ARG2REG) = __arg2; \
        register unsigned long __arg3 asm(__HYPERCALL_ARG3REG) = __arg3; \
        register unsigned long __arg4 asm(__HYPERCALL_ARG4REG) = __arg4; \
        register unsigned long __arg5 asm(__HYPERCALL_ARG5REG) = __arg5;

#define __HYPERCALL_0PARAM      "=r" (__res)
#define __HYPERCALL_1PARAM      __HYPERCALL_0PARAM, "+r" (__arg1)
#define __HYPERCALL_2PARAM      __HYPERCALL_1PARAM, "+r" (__arg2)
#define __HYPERCALL_3PARAM      __HYPERCALL_2PARAM, "+r" (__arg3)
#define __HYPERCALL_4PARAM      __HYPERCALL_3PARAM, "+r" (__arg4)
#define __HYPERCALL_5PARAM      __HYPERCALL_4PARAM, "+r" (__arg5)

#define __HYPERCALL_0ARG()
#define __HYPERCALL_1ARG(a1)                                            \
        __HYPERCALL_0ARG()              __arg1 = (unsigned long)(a1);
#define __HYPERCALL_2ARG(a1,a2)                                         \
        __HYPERCALL_1ARG(a1)            __arg2 = (unsigned long)(a2);
#define __HYPERCALL_3ARG(a1,a2,a3)                                      \
        __HYPERCALL_2ARG(a1,a2)         __arg3 = (unsigned long)(a3);
#define __HYPERCALL_4ARG(a1,a2,a3,a4)                                   \
        __HYPERCALL_3ARG(a1,a2,a3)      __arg4 = (unsigned long)(a4);
#define __HYPERCALL_5ARG(a1,a2,a3,a4,a5)                                \
        __HYPERCALL_4ARG(a1,a2,a3,a4)   __arg5 = (unsigned long)(a5);

#define __HYPERCALL_CLOBBER5    "memory", "$31", "$3", \
"$9", "$10", "$11", "$12", "$13", "$14", "$15", "$24", "$25"
#define __HYPERCALL_CLOBBER4    __HYPERCALL_CLOBBER5, __HYPERCALL_ARG5REG
#define __HYPERCALL_CLOBBER3    __HYPERCALL_CLOBBER4, __HYPERCALL_ARG4REG
#define __HYPERCALL_CLOBBER2    __HYPERCALL_CLOBBER3, __HYPERCALL_ARG3REG
#define __HYPERCALL_CLOBBER1    __HYPERCALL_CLOBBER2, __HYPERCALL_ARG2REG
#define __HYPERCALL_CLOBBER0    __HYPERCALL_CLOBBER1, __HYPERCALL_ARG1REG


#define restore_sp(var) \
({ \
    __asm__ __volatile__ ( \
        ".set push\n" \
        ".set noreorder\n" \
        "move $29, %0\n" \
        ".set pop\n" \
        : \
        : "r"(var) \
        ); \
})

#define _hypercall_destroy_domain(type, name)                           \
({                                                                      \
        __HYPERCALL_DECLS;                                              \
        __HYPERCALL_0ARG();                                             \
        switch_xen_linux_stacks();                                      \
		restore_sp(*(xen_shared_info->arch.xensp[processor_id()]));		\
        asm volatile (__HYPERCALL                                       \
                      : __HYPERCALL_0PARAM                              \
                      : __HYPERCALL_ENTRY(name)                         \
                      : __HYPERCALL_CLOBBER0);                          \
        restore_linux_stack(); \
        (type)__res;                                                    \
})


#define _hypercall0(type, name)                                         \
({                                                                      \
        __HYPERCALL_DECLS;                                              \
        __HYPERCALL_0ARG();                                             \
        switch_xen_linux_stacks(); \
        asm volatile (__HYPERCALL                                       \
                      : __HYPERCALL_0PARAM                              \
                      : __HYPERCALL_ENTRY(name)                         \
                      : __HYPERCALL_CLOBBER0);                          \
        restore_linux_stack(); \
        (type)__res;                                                    \
})

#define _hypercall1(type, name, a1)                                     \
({                                                                      \
        __HYPERCALL_DECLS;                                              \
        __HYPERCALL_1ARG(a1);                                           \
        switch_xen_linux_stacks(); \
        asm volatile (__HYPERCALL                                       \
                      : __HYPERCALL_1PARAM                              \
                      : __HYPERCALL_ENTRY(name)                         \
                      : __HYPERCALL_CLOBBER1);                          \
        restore_linux_stack(); \
        (type)__res;                                                    \
})

#define _hypercall2(type, name, a1, a2)                                 \
({                                                                      \
        __HYPERCALL_DECLS;                                              \
        __HYPERCALL_2ARG(a1, a2);                                       \
        switch_xen_linux_stacks(); \
        asm volatile (__HYPERCALL                                       \
                      : __HYPERCALL_2PARAM                              \
                      : __HYPERCALL_ENTRY(name)                         \
                      : __HYPERCALL_CLOBBER2);                          \
        restore_linux_stack(); \
        (type)__res;                                                    \
})

#define _hypercall3(type, name, a1, a2, a3)                             \
({                                                                      \
        __HYPERCALL_DECLS;                                              \
        __HYPERCALL_3ARG(a1, a2, a3);                                   \
        switch_xen_linux_stacks(); \
        asm volatile (__HYPERCALL                                       \
                      : __HYPERCALL_3PARAM                              \
                      : __HYPERCALL_ENTRY(name)                         \
                      : __HYPERCALL_CLOBBER3);                          \
        restore_linux_stack(); \
        (type)__res;                                                    \
})

#define _hypercall4(type, name, a1, a2, a3, a4)                         \
({                                                                      \
        __HYPERCALL_DECLS;                                              \
        __HYPERCALL_4ARG(a1, a2, a3, a4);                               \
        switch_xen_linux_stacks(); \
        asm volatile (__HYPERCALL                                       \
                      : __HYPERCALL_4PARAM                              \
                      : __HYPERCALL_ENTRY(name)                         \
                      : __HYPERCALL_CLOBBER4);                          \
        restore_linux_stack(); \
        (type)__res;                                                    \
})

#define _hypercall5(type, name, a1, a2, a3, a4, a5)                     \
({                                                                      \
        __HYPERCALL_DECLS;                                              \
        __HYPERCALL_5ARG(a1, a2, a3, a4, a5);                           \
        switch_xen_linux_stacks(); \
        asm volatile (__HYPERCALL                                       \
                      : __HYPERCALL_5PARAM                              \
                      : __HYPERCALL_ENTRY(name)                         \
                      : __HYPERCALL_CLOBBER5);                          \
        restore_linux_stack(); \
        (type)__res;                                                    \
})

static inline long
privcmd_call(unsigned call,
	     unsigned long a1, unsigned long a2,
	     unsigned long a3, unsigned long a4,
	     unsigned long a5)
{
	__HYPERCALL_DECLS;
	__HYPERCALL_5ARG(a1, a2, a3, a4, a5);

	switch_xen_linux_stacks();
	asm volatile(__HYPERCALL
		     : __HYPERCALL_5PARAM
		     : [func] "r" (hypercall_page[call]._entry)
		     : __HYPERCALL_CLOBBER5);
	restore_linux_stack();

	return (long)__res;
}

static inline int
HYPERVISOR_xen_version(int cmd, void *arg)
{
	int (*hc_xen_version)(int cmd, void *arg);
	unsigned long long *hc_offset = (unsigned long long *)hypercall_page;

	hc_xen_version = (void *)(hc_offset[__HYPERVISOR_xen_version]);

	return hc_xen_version(cmd, arg);
}

static inline int
HYPERVISOR_console_io(int cmd, int count, char *str)
{
	int (*hc_console_io)(int cmd, int count, char *str);
	unsigned long *hc_offset = (unsigned long *)hypercall_page;
	int retval;

	hc_console_io = (void *)(*(hc_offset + __HYPERVISOR_console_io));

	switch_xen_linux_stacks();

	retval = hc_console_io(cmd, count, str);

	restore_linux_stack();

	return retval;
}

static inline int
HYPERVISOR_grant_table_op(unsigned int cmd, void *uop, unsigned int count)
{
	return _hypercall3(int, grant_table_op, cmd, uop, count);
}

static inline int
HYPERVISOR_memory_op(unsigned int cmd, void *arg)
{
	return _hypercall2(int, memory_op, cmd, arg);
}

static inline int
HYPERVISOR_destroy_domain(void)
{
	return _hypercall_destroy_domain(int, destroy_domain);
}


#endif /* _ASM_XEN_HYPERCALL_H */
