/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

#ifndef _ATOMIC_H
#define _ATOMIC_H

typedef struct { volatile int counter; } atomic_t;

#define atomic_read(v)	((v)->counter)
#define atomic_set(v,i)	((v)->counter = (i))

static __inline__ void atomic_add(int i, atomic_t * v)
{
	unsigned long temp;

	__asm__ __volatile__(
		"1:   ll      %0, %1      # atomic_add\n"
		"     addu    %0, %2                  \n"
		"     sc      %0, %1                  \n"
		"     beqz    %0, 1b                  \n"
		: "=&r" (temp), "=m" (v->counter)
		: "Ir" (i), "m" (v->counter));
}

static __inline__ void atomic_sub(int i, atomic_t * v)
{
	unsigned long temp;

	__asm__ __volatile__(
		"1:   ll      %0, %1      # atomic_sub\n"
		"     subu    %0, %2                  \n"
		"     sc      %0, %1                  \n"
		"     beqz    %0, 1b                  \n"
		: "=&r" (temp), "=m" (v->counter)
		: "Ir" (i), "m" (v->counter));
}

/*
 * Same as above, but return the result value
 */
static __inline__ int atomic_add_return(int i, atomic_t * v)
{
        unsigned long temp, result;

        __asm__ __volatile__(
        "1:     ll      %1, %2          # atomic_add_return     \n"
        "       addu    %0, %1, %3                              \n"
        "       sc      %0, %2                                  \n"
        "       beqz    %0, 1b                                  \n"
        "       addu    %0, %1, %3                              \n"
        "       sync                                            \n"
        : "=&r" (result), "=&r" (temp), "=m" (v->counter)
        : "Ir" (i), "m" (v->counter)
        : "memory");

        return result;
}

static __inline__ int atomic_sub_return(int i, atomic_t * v)
{
        unsigned long temp, result;

        __asm__ __volatile__(
        "1:     ll      %1, %2          # atomic_sub_return     \n"
        "       subu    %0, %1, %3                              \n"
        "       sc      %0, %2                                  \n"
        "       beqz    %0, 1b                                  \n"
        "       subu    %0, %1, %3                              \n"
        "       sync                                            \n"
        : "=&r" (result), "=&r" (temp), "=m" (v->counter)
        : "Ir" (i), "m" (v->counter)
        : "memory");

        return result;
}

#ifdef XLR
typedef struct { volatile int value; } phnx_atomic_t;

static __inline__ int phnx_test_and_set(phnx_atomic_t *lock)
{
	  int oldval = 0;

	    __asm__ __volatile__ (".set pus"
		".set noreorder\n"
		"move $9, %2\n"
		"li $8, 1\n"
		".word 0x71280014\n"
		"move %1, $8\n"
		".set pop\n"
		: "+m" (lock->value), "=r" (oldval)
		: "r" ((unsigned long)&lock->value)
		: "$8", "$9"
	);
	return (oldval == 0 ? 1/*success*/ : 0/*failure*/);
}

static __inline__ int swap_w(unsigned int value, int *addr)
{
  __asm__ __volatile__ (".set push\n"
			".set noreorder\n"
		    "move $8, %2\n"
		    "move $9, %3\n"
			//"swapw $8, $9\n"
			".word 0x71280014\n"
			"move %0, $8\n"
			".set pop\n"
			: "=&r"(value), "+m" (*addr)
			: "0"(value), "r" ((unsigned long)addr)
			: "$8", "$9"
			);
  return value;
}

static __inline__ int test_and_set(int *addr)
{
  int oldval = 0;
  oldval = swap_w(1, addr);
  return oldval == 0 ? 1/*success*/ : 0/*fail*/;
}
#endif /* XLR */

#endif
