/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/
#ifndef __BOOT2_H__
#define __BOOT2_H__

#include "types.h"
#include "system.h"
#include "shared_structs.h"
#include "global_shmem.h"
#include "nlm_hal_macros.h"

/* This area if for global memory across all apps running on XLR/XLS
   Currently we use this for implementing the bridge errata workaround
   as a global lock area. Refer to global_shmem.h for reserving various
   regions in this area
 */

#define GLOBAL_SHMEM_START      (0xffffffff80100000ULL - 0x1000)


enum { KUSEG_MODE, KSEG0_MODE };

extern int run_mode;
extern int commoncode_seg_enabled;
extern int verbose;

extern char *envs[];

struct phoenix_app_init {
	signed char (*init) (void);
};

struct phoenix_app_run {
	void (*run) (void);
	struct phoenix_app_run *next;
};

/* Check for the valid memory address */

static __inline__ int not_translated_vaddr(unsigned long addr)
{
	if (!((addr >= (unsigned long)KSEG0)
	      && (addr < (unsigned long)KSEG0 + (512 << 20)))
	    && !((addr >= (unsigned long)KSEG1)
		 && (addr < (unsigned long)KSEG1 + (512 << 20)))
	    && !((addr >= (unsigned long)KSEG2)
		 && (addr < (unsigned long)KSEG2 + (512 << 20)))
	    && !((addr >= (unsigned long)KSEG3)
		 && (addr < (unsigned long)KSEG3 + (512 << 20) - 1))) {
		return 1;	/* Invlalid address */
	} else {
		return 0;	/* Valid address */
	}

}

static __inline__ void userapp_load_scratch_info(int master_cpu,
						 uint32_t buddy_mask)
{
	uint64_t cpu_mask = (((uint64_t) buddy_mask << 32) | ((uint64_t) master_cpu));

	__asm__ __volatile__(
		".set push\n"
		".set noreorder\n"
		"move $28, %0\n"		/* move into gp reg (&boot1_info) */
		"dmtc0 %0, $22, 0\n"	/* move (&boot1_info) into $22    */
		"dmtc0 %1, $22, 1\n"	/* move (master cpu + buddy mask) into $22       */
	    ".set pop\n"
		:
		:"r"(boot1_info), "r"(cpu_mask)
	);
}

extern int8_t userapp_init(void);

extern void register_app_timeslice(struct phoenix_app_run *app);
extern void deregister_app_timeslice(struct phoenix_app_run *app);
extern void phnx_initialize_getopt(void);

#endif
