/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

#ifndef _CPU_H
#define _CPU_H

#define CPU_BLOCKID_IFU      0
#define CPU_BLOCKID_ICU      1
#define CPU_BLOCKID_IEU      2
#define CPU_BLOCKID_LSU      3
#define CPU_BLOCKID_MMU      4
#define CPU_BLOCKID_PRF      5

#define IFU_THR_EN_REGID             0
#define IFU_THR_SLEEP_REGID          1
#define IFU_THR_SCHED_POLICY_REGID   2
#define IFU_THR_SCHED_COUNTER_REGID  3
#define IFU_BHR_PROG_MASK_REGID      4
#define IFU_DISABLE_SB_SLEEP_REGID   5
#define IFU_DEFEATURE_REGID          6
#define IFU_SLEEP_STATE_REGID        16
#define IFU_BHR_REGID                17
#define IFU_IPG_0_REGID              20
#define IFU_IPG_1_REGID              21
#define IFU_IPG_2_REGID              22
#define IFU_IPG_3_REGID              23
#define IFU_RAS_0_REGID              24
#define IFU_RAS_1_REGID              25
#define IFU_RAS_2_REGID              26
#define IFU_RAS_3_REGID              27
#define IFU_RAS_4_REGID              28
#define IFU_RAS_5_REGID              29
#define IFU_RAS_6_REGID              30
#define IFU_RAS_7_REGID              31
#define IFU_BHT_0_31_REGID           64
#define IFU_BHT_2016_2047_REGID      127

#define LSU_CFG0_REGID       0
#define LSU_CFG1_REGID       1
#define LSU_CFG2_REGID       2
#define LSU_CFG3_REGID       3
#define LSU_CFG4_REGID       4
#define LSU_STAT_REGID       5
#define LSU_DEF_REGID        6
#define LSU_DBG0_REGID       7
#define LSU_DBG1_REGID       8
#define LSU_CERRLOG_REGID    9
#define LSU_CERROVF_REGID    10
#define LSU_CERRINT_REGID    11

#define IEU_DEFEATURE_REGID 0

#define ICU_DEFEATURE_REGID 0

#define MMU_SETUP_REGID 0


#define NETOS_PTEBASE_SCRATCH_IDX 	 3
#define HPEXEC_PTEBASE_SCRATCH_IDX 	 4

#ifndef __ASSEMBLY__

#include "irq.h"
#include "pt_regs.h"
#include "asmmacros.h"
#include "xlr_boot_lib.h"

/* Scratch register usage */
#define BINFO_SCRATCH_IDX 		 0
#define WAKEUP_SCRATCH_IDX  	 7
#define CRF_PERF0_SCRATCH_IDX    NLM_PERF0_SCRATCH
#define CRF_PERF1_SCRATCH_IDX    NLM_PERF1_SCRATCH

extern void reset_printk_base_lock();

static inline void cpu_halt(void)
{
	for (;;) {
		__asm__ volatile("wait\n");
	}
}

static inline void cpu_enable_l2_cache(void)
{
	uint64_t value = 0;

	value = read_phnx_ctrl ((CPU_BLOCKID_LSU << 8) | LSU_CFG0_REGID);
	write_phnx_ctrl (((CPU_BLOCKID_LSU << 8) | LSU_CFG0_REGID), (value | 0x10));  
}

static inline void cpu_disable_l2_cache(void)
{
	uint64_t value = 0;

	value = read_phnx_ctrl((CPU_BLOCKID_LSU << 8) | LSU_CFG0_REGID);
	write_phnx_ctrl(((CPU_BLOCKID_LSU << 8) | LSU_CFG0_REGID), (value & ~0x10));  
}

extern void flush_icache_all(void);
extern void flush_dcache_all(void);
extern void per_cpu_arch_init(void);

/*
 * prototypes for functions pertaining to exception processing
 */

extern void show_regs(struct pt_regs *regs);
extern void *set_except_vector(int n, void *addr);
extern void do_reserved(struct pt_regs *regs);
extern void trap_init(void);
void tlb_invalid(struct pt_regs *regs);
void ldr_do_syscall(struct pt_regs *regs);

struct exception_region {
    unsigned int data[1024];
};
extern struct exception_region *exception_vectors;
extern int smp_mode;
extern uint64_t find_memblk(unsigned long size);

extern void panic(const char *fmt,...);

extern void setup_tlb(unsigned long virt, unsigned long long phys, int size);
extern int setup_kseg2_tlb(unsigned long virt, unsigned long long phys, int size);
extern int setup_kseg2_tlb_uncached(unsigned long virt, unsigned long long phys, int size);

extern void dump_stack(void);

extern unsigned long create_mapping(unsigned long long phys, unsigned long size);
extern unsigned long create_mapping_uncached(unsigned long long phys, unsigned long size);
extern void dump_page_table();

static __inline__ int chip_id(void)
{
        unsigned int chip = 0;

        __asm__ volatile (""
              ".set push\n"
                          ".set mips64\n"
                          "mfc0 %0, $15\n"
              ".set pop\n":"=r" (chip));
        return (chip);
}

#endif /* __ASSEMBLY__ */


/* Stored in the ebase ptr + 2048 */
#define NETOS_XKPHYS_CACHED 0xa800000000000000
#define NETOS_EBASE_PCPU_OFF 2048
#define NETOS_GEN_EXC_DISP_HDLR_SZ 768
#define NETOS_XTLB_REFILL_HDLR_SZ 512
#ifndef __ASSEMBLY__
struct netos_pcpu_info
{
	unsigned long long sp; /* netos stack */
	unsigned long long exccnt; /* To know the nested exception */
	char gen_exc_disp_hdlr[NETOS_GEN_EXC_DISP_HDLR_SZ]; /* genexc dispatch handler is copied to this address*/
	char tlb_refill_hdlr[NETOS_XTLB_REFILL_HDLR_SZ]; /* refill handler will be copied to this address */
};
#endif
#define NETOS_PCPU_SP_OFF 		0
#define NETOS_PCPU_EXCCNT_OFF		8
#define NETOS_PCPU_EXC_DISP_HDLR_OFF   16
#define NETOS_PCPU_XTLB_REFILL_OFF 	784


/* These includes need to follow the above
 * static-defined common routines
 */
#ifdef XLP_SIM
#include "xlp_cpu.h" 
#else
#include "xlr_xls_cpu.h"
#endif

#endif
