/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

// this file contains packet descriptor data structures
// to handle multi message packets and queues of multi
// message packets

#ifndef _DESC_MSG_H
#define _DESC_MSG_H

#include "types.h"
#include "msgring.h"
#include "spinlock.h"
#include "cache.h"
#include "mips-exts.h"

#define MSG_PER_PACKET 20
#define MSG_QUEUE_SIZE 20

/* struct to hold variable size message ring  message */

typedef struct dmsg {
	unsigned long long msg[4];
} dmsg;


/* typedef data structure for packet descriptors */

typedef struct desc {
	uint64_t     *entry;
	int          wr_ptr;
	int          rd_ptr;
	int          size;
} desc;

/* init multi desc packet structure */

static inline desc* desc_init (desc *d) {
	d->entry =  (uint64_t*)cacheline_aligned_malloc(MSG_PER_PACKET*sizeof(uint64_t));
	d->wr_ptr = 0;
	d->rd_ptr = 0;
	d->size = 0;
	return d;
}

/* add entry to packet descriptor */

static inline void desc_add (desc *d, uint64_t in_msg) {
	d->entry[d->wr_ptr] = in_msg;
	d->wr_ptr = (d->wr_ptr +1) % MSG_PER_PACKET;
	d->size++;
}

/* function to dump out contents of packet */

static inline void desc_dump(desc *d) {
	int i;
	for (i = 0; i < d->size; i++) {
		printf ("pkt desc %d is 0x%llx address %p\n", i, (unsigned long long) d->entry[i], &(d->entry[i]));    
	}
}

/* function to reset packet descriptor entry */

static inline void desc_free (desc *d) {
	d->wr_ptr = 0;
	d->rd_ptr = 0;
	d->size   = 0;
}

/* typedef for queue of packet descriptors */

typedef struct desc_q {
	desc    d[MSG_QUEUE_SIZE];
	int     wr_ptr;
	int     rd_ptr;
	int     size;
} desc_q;

/* function to init queue */

static inline void desc_queue_init(desc_q *dq) {
	int i;
	
	dq->wr_ptr = 0;
	dq->rd_ptr = 0;
	dq->size = 0; 
	for (i = 0; i < 20; i++) {
		desc_init(&(dq->d[i]));
	}
}

/* function to get ptr to packet from queue */

static inline desc* desc_get_wr_pkt (desc_q *dq) { 
	desc* dp = &(dq->d[dq->wr_ptr]);
	dq->wr_ptr = (dq->wr_ptr + 1) % MSG_QUEUE_SIZE;
	dq->size++;
	return dp;
}

/* return used packet buffer to queue */

static inline desc* desc_get_rd_pkt(desc_q *dq) {
	desc* dp = &(dq->d[dq->rd_ptr]);
	dq->rd_ptr = (dq->rd_ptr + 1)  % MSG_QUEUE_SIZE;
	dq->size--;
	return dp;
}

/* get queue size */

static inline int desc_get_q_size(desc_q *dq) {
	return dq->size;
}

#endif
