#ifndef _INIT_H
#define _INIT_H

#if __GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 3)
#define __attribute_used__ __attribute__((__used__))
#else
#define __attribute_used__ __attribute__((__unused__))
#endif

#define __initdata   \
    __attribute__ ((__section__ (".init.data")))
#define __initsetup  \
    __attribute_used__ __attribute__ ((__section__ (".init.setup")))

/*
 * Used for kernel command line parameter setup
 */
struct kernel_param {
    const char *name;
    enum { OPT_STR, OPT_UINT, OPT_BOOL, OPT_INVBOOL, OPT_CUSTOM } type;
    void *var;
    unsigned int len;
};

extern struct kernel_param __setup_start, __setup_end;

#define custom_param(_name, _var) \
    static char __setup_str_##_var[] __initdata = _name; \
    static struct kernel_param __setup_##_var __attribute_used__ \
        __initsetup = { __setup_str_##_var, OPT_CUSTOM, &_var, 0 }
#define boolean_param(_name, _var) \
    static char __setup_str_##_var[] __initdata = _name; \
    static struct kernel_param __setup_##_var __attribute_used__ \
        __initsetup = { __setup_str_##_var, OPT_BOOL, &_var, sizeof(_var) }
#define invbool_param(_name, _var) \
    static char __setup_str_##_var[] __initdata = _name; \
    static struct kernel_param __setup_##_var __attribute_used__ \
        __initsetup = { __setup_str_##_var, OPT_INVBOOL, &_var, sizeof(_var) }
#define integer_param(_name, _var) \
    static char __setup_str_##_var[] __initdata = _name; \
    static struct kernel_param __setup_##_var __attribute_used__ \
        __initsetup = { __setup_str_##_var, OPT_UINT, &_var, sizeof(_var) }
#define string_param(_name, _var) \
    static char __setup_str_##_var[] __initdata = _name; \
    static struct kernel_param __setup_##_var __attribute_used__ \
        __initsetup = { __setup_str_##_var, OPT_STR, &_var, sizeof(_var) }

#endif
