/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

#ifndef _PARSER_H
#define _PARSER_H

#include "types.h"
#include "msgring.h"
#include "mac.h"
#include "iomap.h"
#include "pkt_gen.h"
#include "net_config.h"
#define GMAC_MODE  0
#define XGMAC_MODE 1
#define SPI4_MODE  2

#define L2_ETH   0x1
#define L2_FIXED 0x2
#define L2_PPP   0x3

#define PREPAD_ENABLE 16

// --------------------------------------------------------------------------------
// L2 TABLE STRUCTURE
// --------------------------------------------------------------------------------

typedef struct l2_table_entry {
	int   extraHdrProtoSize;    //  5 bits
	int   extraHdrProtoOffset;  //  6 bits
	int   extraHdrSize;         //  6 bits
	int   proto_offset;         //  6 bits
	int   l2_header_offset;     //  6 bits
	int   l2_proto;             //  2 bits

} l2_entry, *p_l2_entry;

typedef struct l2_table {
	l2_entry entry [16]; 
} l2_table, *p_l2_table;

// --------------------------------------------------------------------------------
// L3 TABLE STRUCTURE
// --------------------------------------------------------------------------------

typedef struct l3_table_entry {

	uint32_t  port;                // 4 bits
	uint32_t  port_mask;           // 1 bit
	uint32_t  l3_prot_mask;        // 1 bit
	uint32_t  l3_prot_key;         // [15:0]
	uint32_t  l2_prot;             // [1:0] 17:16
	uint32_t  ip_checksum_compute; // bit 18
	uint32_t  l4_prot_offset;      // [5:0] 25:20
	uint32_t  l3_header_offset;    // [5:0] 31:26
	//       
	uint32_t  offset_0;            // [6:0] 31:25
	uint32_t  len_0;               // [3:0] 24:21
	uint32_t  offset_1;            // [6:0] 20:14
	uint32_t  len_1;               // [3:0] 13:10
	uint32_t  offset_2;            // [5:0] 9:4 
	uint32_t  len_2;               // [3:0] 3:0

	uint32_t  l4_classify;         // reserved field, Depricated for B* silicon,  not used
} l3_entry, *p_l3_entry;

typedef struct l3_table {
	l3_entry entry [16];
} l3_table, *p_l3_table;

// --------------------------------------------------------------------------------
// L4 TABLE STRUCTURE
// --------------------------------------------------------------------------------

typedef struct l4_table_entry {
	uint32_t  l4_prot_key;           // [7:0]
	uint32_t  port;                  // 
	uint32_t  l3_addr;               // 
	uint32_t  l4_prot_mask;          // [7:0]
	uint32_t  port_mask;             // 
	uint32_t  l3_addr_mask;          // 
	uint32_t  reserved;              // [5:0] 6:1
	uint32_t  len_4;                 // [3:0] 10:7
	uint32_t  offset_4;              // [5:0] 16:11
	uint32_t  len_3;                 // [3:0] 20:17
	uint32_t  offset_3;              // [5:0] 26:21
	uint32_t  reserved_2;            // [4:0] 31:27
} l4_entry, *p_l4_entry;

typedef struct l4_table {
	l4_entry entry [8];
} l4_table, *p_l4_table;

// --------------------------------------------------------------------------------
// PARSER CONFIG
// --------------------------------------------------------------------------------

typedef struct parser_config {
	// 0x100
	int   unclassified_mask;    //   9 bits [24:15]
	int   crc_hash_poly;        //   6 bits [14:8]
	int   use_global_port;      //   1 bit [3:3]
	int   use_cam;              //   1 bit [2:2]
	int   use_hash;             //   1 bit [1:1]
	int   use_proto;            //   1 bit [0:0]
	// 0x101
	int   poly_mask;           //   7 bits [11:5]
	int   parse_depth;         //   5 bits [4:0]

} prsr_cfg_entry, *p_prsr_cfg_entry;


// --------------------------------------------------------------------------------
// TRANSLATE TABLE 
// --------------------------------------------------------------------------------

typedef struct translate_table_entry {
	int   class_id;           // 2 bits 
	int   bucket_id;          // 6 bits 
	int   use_bucket_id;      // 1 bits      
} trans_entry, *p_trans_entry;

typedef struct translate_table {
	trans_entry entry[128];
	int num_entries;
} ttable, *p_ttable;

typedef struct tcam_translate_table {
	trans_entry entry[4];
	int num_entries;
} tcam_ttable, *p_tcam_ttable;

// --------------------------------------------------------------------------------
// CAM TABLE AND MASK
// --------------------------------------------------------------------------------

typedef struct cam_table_and_mask {
	uint8_t data_0[16], mask_0[16];
	uint8_t data_1[16], mask_1[16];
	uint8_t data_2[16], mask_2[16];
	uint8_t data_3[16], mask_3[16];
}cam_table, *p_cam_table;


// --------------------------------------------------------------------------------
// PREPAD STRUCTURES
// --------------------------------------------------------------------------------

typedef struct prepad_struct {
	uint8_t entry [32];
} prepad, *p_prepad;


typedef struct parser_error_class {
	uint32_t portId;
	uint32_t pkt_count;
	uint32_t l3_cam_addr;
	uint32_t l4_cam_addr;
	uint32_t l3_csum;
	uint32_t l4_csum;
	uint32_t vector_mismatch;
	uint32_t l3_cam_hit;
	uint32_t l4_cam_hit;
	uint32_t extract_0_valid;
	uint32_t extract_1_valid;
	uint32_t extract_2_valid;
	uint32_t extract_3_valid;
	uint32_t extract_4_valid;
	uint32_t extract_5_valid;
	uint32_t l2_header_done ;
	uint32_t l3_header_done ;
	uint32_t hash_32 ;
	p_packet error_packet;
	p_prepad error_prepad;
} parser_error_t;

/* Parser Config */
extern prsr_cfg_entry   prsr_cfg;
extern p_prsr_cfg_entry p_prsr_cfg;

/* Translate table */
extern ttable   _ttable;
extern p_ttable _p_ttable;

/* L2CTable */
extern l2_table   l2_t; 
extern p_l2_table p_l2_t;

/* L3CTable */
extern l3_table   l3_t;
extern p_l3_table p_l3_t;

/* L4CTable */
extern l4_table   l4_t;
extern p_l4_table p_l4_t;

/* cam table and mask */
extern cam_table   cam_t; 
extern p_cam_table p_cam_t;

extern uint32_t  l2_protocol_index;
extern uint32_t  l2_protocol;
extern uint32_t  l3_start_index;
extern uint32_t  l3_checksum_length;
extern uint32_t  l3_ctable_index;
extern uint32_t  l3_checksum_valid;
extern uint32_t  l4_checksum_valid;
extern uint32_t  l3_protocol_index;
extern uint32_t  l3_protocol;
extern uint32_t  l4_start_index;
extern uint32_t  l4_ctable_index;
extern uint32_t  extract_vector;
extern uint32_t  field_protocol;
extern uint32_t  ethernet_format_check_pkt;
extern uint32_t  ppp_format_check_pkt;


static inline uint32_t calc_packet_offset(p_packet _my_pkt, int prepad_on) {
    uint32_t packet_offset = 0;

    if ( prepad_on ) packet_offset = 32;
    return ( packet_offset );
}

static inline uint32_t calc_l2_protocol_index(p_packet _my_pkt, int prepad_on) {
    uint32_t index;
    if ( ethernet_format_check_pkt ) {
        // the format is encoded in the last byte of the packet
        if ( _my_pkt->data[(_my_pkt->len)-1] == 0x1 ) {
            index = 16; // 1 - ethernet vlan
        } else if ( _my_pkt->data[(_my_pkt->len)-1] == 0x2 ) {
            index = 20; // 2 - ethernet llc snap
        } else if ( _my_pkt->data[(_my_pkt->len)-1] == 0x3 ) {
            index = 24; // 3 - ethernet llc snap vlan
        } else {
            index = 12; // plain ethernet
        }
    } else if ( ppp_format_check_pkt ) {
        if ( _my_pkt->data[(_my_pkt->len)-1] == 0x1 ) {
            index = 2; // uncompressed ppp
        } else {
            index = 0; // compressed ppp
        }
    } else {
        index = 12;
    }
    return (index);
} 

static inline uint32_t calc_l2_protocol(p_packet _my_pkt, int prepad_on) {
    uint32_t protocol = 0 ;
    uint32_t index = calc_l2_protocol_index(_my_pkt, prepad_on);
    uint32_t PACKET_OFFSET = calc_packet_offset(_my_pkt, prepad_on);

    protocol = ((_my_pkt->data[PACKET_OFFSET+index] & 0xff ) << 8 ) | 
               (_my_pkt->data[PACKET_OFFSET+index+1] & 0xff ); 

    return protocol;
}

static inline uint32_t calc_l3_ctable_index(p_packet _my_pkt, int prepad_on, int portId) {
    uint32_t l2_proto_entry = L2_ETH;
    uint32_t l2_protocol    = calc_l2_protocol(_my_pkt, prepad_on);
    uint32_t l3_key         = ( ( l2_proto_entry & 0x3 ) << 16 ) | l2_protocol; 
    uint32_t i;
    uint32_t hit_index;

    hit_index = 15;
    for ( i = 0; i < 16 ; i ++ ) {
        uint32_t table_entry = (( p_l3_t->entry[i].l2_prot & 0x3 ) << 16 ) | 
								p_l3_t->entry[i].l3_prot_key;
        if ( l3_key == table_entry ) {
            hit_index = i;
            break; 
        }
    }
    return (hit_index);
}

static inline uint32_t calc_l3_start_index(p_packet _my_pkt, int prepad_on) {
    uint32_t index; 

    index = calc_l2_protocol_index(_my_pkt, prepad_on) + 2 ;
    return index;
}

static inline uint32_t calc_l3_protocol_index(p_packet _my_pkt, int prepad_on) {
    uint32_t index; 

    // assuming ipv4
    index = calc_l3_start_index(_my_pkt, prepad_on) + 9;

    return (index);
}

static inline uint32_t calc_l3_protocol(p_packet _my_pkt, int prepad_on) {
    uint32_t l3_protocol = 0 ;
    uint32_t index       = calc_l3_protocol_index(_my_pkt, prepad_on) ;
    uint32_t PACKET_OFFSET = calc_packet_offset(_my_pkt, prepad_on);

    l3_protocol = _my_pkt->data[PACKET_OFFSET+index];
    return l3_protocol;
}

static inline  uint32_t calc_l3_checksum_length(p_packet _my_pkt, int prepad_on) {
    uint32_t l3_length = 0;
    uint32_t PACKET_OFFSET = calc_packet_offset(_my_pkt, prepad_on);

    // FIXME - assuming ipv4
    uint32_t ihl = 
		(_my_pkt->data[PACKET_OFFSET + calc_l3_start_index(_my_pkt, prepad_on)]) & 0x0f ; 
    l3_length    = ihl*4; 

    return (l3_length);
}

static inline uint32_t calc_l4_start_index (p_packet _my_pkt, int prepad_on) {
    uint32_t l4_start_index = calc_l3_start_index(_my_pkt, prepad_on) + 
							  calc_l3_checksum_length(_my_pkt, prepad_on);
    return l4_start_index;
}

static inline uint32_t calc_l4_ctable_index (p_packet _my_pkt, int prepad_on) {
    uint32_t PACKET_OFFSET = calc_packet_offset(_my_pkt, prepad_on);
    uint32_t l4_protocol = 
		_my_pkt->data[PACKET_OFFSET + calc_l3_protocol_index(_my_pkt, prepad_on)];
    uint32_t l4_hit_index = 8;
    uint32_t l4_entry     = 0;   
    uint32_t i;

    for ( i = 0; i < 8 ; i ++ ) {
        l4_entry = p_l4_t->entry[i].l4_prot_key;
        if ( l4_entry == l4_protocol ) {
            l4_hit_index = i;
            break;
        }
    }
    return ( l4_hit_index );
}

static inline uint32_t calc_field_protocol (p_packet _my_pkt, int prepad_on, int portId) {
    uint32_t index = (calc_l3_ctable_index(_my_pkt, prepad_on, portId) << 3) | 
					 ( calc_l4_ctable_index(_my_pkt, prepad_on) );
    return index;
}


static inline uint32_t calc_l3_checksum_valid (p_packet _my_pkt, int prepad_on) {
    return (1); // FIXME
}

static inline uint32_t calc_l4_checksum_valid (p_packet _my_pkt, int prepad_on) {
    return (1); // FIXME
}

static inline uint32_t calc_l3_cam_hit (p_packet _my_pkt, int prepad_on) {
    return(1);
}

static inline uint32_t calc_l4_cam_hit (p_packet _my_pkt, int prepad_on) {
    return(1);
}

/* L2 extraction valid */
static inline uint32_t calc_extract_0_valid (p_packet _my_pkt, int prepad_on) {
    return(1);
}

/* L3 - 0 extraction valid */
static inline uint32_t calc_extract_1_valid (p_packet _my_pkt, int prepad_on) {
    return(1);
}

/* L3 - 1 extraction valid */
static inline uint32_t calc_extract_2_valid (p_packet _my_pkt, int prepad_on) {
    return(1);
}

/* L3 - 2 extraction valid */
static inline uint32_t calc_extract_3_valid (p_packet _my_pkt, int prepad_on) {
    return(1);
}

/* L4 - 0 extraction valid */
static inline uint32_t calc_extract_4_valid (p_packet _my_pkt, int prepad_on) {
    return(1);
}

/* L4 - 1 extraction valid */
static inline uint32_t calc_extract_5_valid (p_packet _my_pkt, int prepad_on) {
    return(1);
}

static inline uint32_t calc_l2_header_done (p_packet _my_pkt, int prepad_on) {
    return(1);
}

static inline uint32_t calc_l3_header_done (p_packet _my_pkt, int prepad_on) {
    return(1);
}

static inline uint32_t get_extract_0 (p_prepad my_prepad) {
	uint32_t temp;
	temp = my_prepad->entry[16] << ( 8 * 3 ) |
		my_prepad->entry[17] << ( 8 * 2 ) |
		my_prepad->entry[18] << ( 8 * 1 ) |
		my_prepad->entry[19] << ( 8 * 0 );
	return(temp);
}

static inline uint32_t get_extract_1 (p_prepad my_prepad) {
	uint32_t temp;
	temp = my_prepad->entry[20] << ( 8 * 3 ) |
		my_prepad->entry[21] << ( 8 * 2 ) |
		my_prepad->entry[22] << ( 8 * 1 ) |
		my_prepad->entry[23] << ( 8 * 0 );
	return(temp);
}

static inline uint32_t get_extract_2 (p_prepad my_prepad) {
	uint32_t temp;
	temp = my_prepad->entry[24] << ( 8 * 3 ) |
		my_prepad->entry[25] << ( 8 * 2 ) |
		my_prepad->entry[26] << ( 8 * 1 ) |
		my_prepad->entry[27] << ( 8 * 0 );
	return(temp);
}

static inline uint32_t get_extract_3 (p_prepad my_prepad) {
	uint32_t temp;
	temp = my_prepad->entry[28] << ( 8 * 3 ) |
		my_prepad->entry[29] << ( 8 * 2 ) |
		my_prepad->entry[30] << ( 8 * 1 ) |
		my_prepad->entry[31] << ( 8 * 0 );
	return(temp);
}

static inline uint32_t get_l4_cam_addr (p_prepad my_prepad) {
	return ( my_prepad->entry[15] & 0x7 );
}

static inline uint32_t get_l3_cam_addr (p_prepad my_prepad) {
	return ( ( my_prepad->entry[15] & 0x78 ) >> 3 );
}

static inline uint32_t get_crc_32 (p_prepad my_prepad) {
	uint32_t temp;
	temp =   my_prepad->entry[11] << ( 8 * 3 ) |
	         my_prepad->entry[12] << ( 8 * 2 ) |
	         my_prepad->entry[13] << ( 8 * 1 ) |
	         my_prepad->entry[14] << ( 8 * 0 );
	return(temp);
}

static inline uint32_t get_l4_csum_valid (p_prepad my_prepad) {
	return ( ( my_prepad->entry[10] & 0x1 ) );
}

static inline uint32_t get_ip_header_csum_valid (p_prepad my_prepad) {
	return (( my_prepad->entry[10] & 0x02 ) >> 1 );
}

static inline uint32_t get_l3_header_done (p_prepad my_prepad) {
	return(( my_prepad->entry[10] & 0x04 ) >> 2);
}

static inline uint32_t get_l2_header_done (p_prepad my_prepad) {
	return(( my_prepad->entry[10] & 0x08 ) >> 3);
}

// L4 - 0 extraction valid
static inline uint32_t get_extract_4_valid (p_prepad my_prepad) {
	return(( my_prepad->entry[10] & 0x10 ) >> 4);
}

// L3 - 2 extraction valid
static inline uint32_t get_extract_3_valid (p_prepad my_prepad) {
	return(( my_prepad->entry[10] & 0x20 ) >> 5);
}

// L3 - 1 extraction valid
static inline uint32_t get_extract_2_valid (p_prepad my_prepad) {
	return(( my_prepad->entry[10] & 0x40 ) >> 6);
}

// L3 - 0 extraction valid
static inline uint32_t get_extract_1_valid (p_prepad my_prepad) {
	return(( my_prepad->entry[10] & 0x80 ) >> 7);
}

// L2 extraction valid
static inline uint32_t get_extract_0_valid (p_prepad my_prepad) {
	return( my_prepad->entry[9] & 0x01 );
}

// extra header extraction valid
static inline uint32_t get_extra_header_valid (p_prepad my_prepad) {
        return(( my_prepad->entry[9] & 0x02 ) >> 1);
}

static inline uint32_t get_l4_cam_hit (p_prepad my_prepad) {
	return(( my_prepad->entry[9] & 0x04 ) >> 2);
}

static inline uint32_t get_l3_cam_hit (p_prepad my_prepad) {
	return(( my_prepad->entry[9] & 0x08 ) >> 3);
}

static inline uint32_t get_seq_no (p_prepad my_prepad) {
	uint32_t temp;
	temp =   (my_prepad->entry[ 6] & 0x0f) << ( 8 * 2 ) |
	         my_prepad->entry[ 7] << ( 8 * 1 ) |
	         my_prepad->entry[ 8] << ( 8 * 0 );
	return(temp);
}

static inline uint32_t get_hash7 (p_prepad my_prepad) {
	return( my_prepad->entry[5] & 0x7f );
}

static inline uint32_t get_bucket (p_prepad my_prepad) {
	return( my_prepad->entry[4] & 0x3f );
}

static inline uint32_t get_class (p_prepad my_prepad) {
	return( my_prepad->entry[3] & 0x3 );
}

static inline uint32_t get_use_bucket (p_prepad my_prepad) {
	return( my_prepad->entry[2] & 0x1 );
}



static inline void deconstruct_prepad_old ( p_prepad my_prepad ) {

        printf("Deconstructing prepad \n");
        printf("hash7                = %x\n",get_hash7(my_prepad));
        printf("sequence_number      = %x\n",get_seq_no(my_prepad));
        printf("bucket               = %x\n",get_bucket(my_prepad));
        printf("l3_cam_hit           = %x\n",get_l3_cam_hit(my_prepad));
        printf("l4_cam_hit           = %x\n",get_l4_cam_hit(my_prepad));
        printf("extract_0_valid      = %x\n",get_extract_0_valid(my_prepad));
        printf("extract_1_valid      = %x\n",get_extract_1_valid(my_prepad));
        printf("extract_2_valid      = %x\n",get_extract_2_valid(my_prepad));
        printf("extract_3_valid      = %x\n",get_extract_3_valid(my_prepad));
        printf("extract_4_valid      = %x\n",get_extract_4_valid(my_prepad));
        //printf("extract_5_valid      = %x\n",get_extract_5_valid(my_prepad));
        printf("l2_header_done       = %x\n",get_l2_header_done(my_prepad));
        printf("l3_header_done       = %x\n",get_l3_header_done(my_prepad));
        printf("ip_header_csum_valid = %x\n",get_ip_header_csum_valid(my_prepad));
        printf("l4_csum_valid        = %x\n",get_l4_csum_valid(my_prepad));
        printf("crc_32               = %x\n",get_crc_32(my_prepad));
        printf("l3_cam_addr          = %x\n",get_l3_cam_addr(my_prepad));
        printf("l4_cam_addr          = %x\n",get_l4_cam_addr(my_prepad));
        printf("extract              = %x%x%x%x\n",get_extract_0(my_prepad),get_extract_1(my_prepad),
                                                   get_extract_2(my_prepad),get_extract_3(my_prepad));

}

/*
 * Extern Declarations of exported functions
 */

extern uint32_t calc_thread_id (p_packet _my_pkt, int prepad_on, int portId);
extern int check_extraction (p_packet recv_pkt, p_prepad _prepad, 
                      int id, int port, int pkt_count);
extern  void gmac_5t_extract(volatile unsigned int * _mmio, int mac_num);
extern  void spi4_5t_extract(int _spi4_id);
extern  void update_translate_table(p_ttable my_ttable, volatile unsigned int *_mmio);
extern  void update_cam_results_table(p_tcam_ttable my_ttable , volatile unsigned int * _mmio);
extern  void update_cam_table(p_cam_table my_table, volatile unsigned int * _mmio);
extern  void prepad_copy(p_prepad new_prepad, p_prepad old_prepad);
extern  void init_translate_table(p_ttable my_ttable);
extern  void init_tcam_translate_table(p_tcam_ttable my_ttable);
extern  void init_cam_table(p_cam_table my_table);
extern  void set_desc_vlan_tag(struct msgrng_msg *msg, uint16_t vlan_tag);
extern  void init_classifier (void);
extern  void parser_error_print (parser_error_t _my_error);
extern  uint32_t calc_hash_32_5t(uint32_t ipsa, uint32_t ipda, uint32_t l4_proto,
					uint32_t tcp_sp, uint32_t tcp_dp);

extern  void update_l3_table(p_l3_table my_table, volatile unsigned int *_mmio);
extern  void init_l3_table(p_l3_table my_table);
extern  void update_l4_table(p_l4_table my_table, volatile unsigned int *_mmio);
extern  void init_l4_table(p_l4_table my_table);
extern  void update_prsr_cfg(p_prsr_cfg_entry _my_entry, volatile unsigned int * _mmio);
extern  void init_prsr_cfg (p_prsr_cfg_entry _my_entry);
extern  void update_l2_table(int mac_mode, p_l2_table _my_table, volatile unsigned int * _mmio, int mac_num);
void update_l2_table_block(int mac_mode, p_l2_table _my_table, volatile unsigned int * _mmio, int mac_num, int block,  net_config *net_conf);
extern  void update_l2_table_xgs ( p_l2_table _my_table, volatile unsigned int * _mmio);
extern  void init_l2_table (p_l2_table _my_table);
extern  void parser_error_init (parser_error_t _my_error);

#endif
