/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/
#ifndef __STACKFRAME_H__
#define __STACKFRAME_H__

#include "regdef.h"
#include "asm.h"
#include "types.h"
#include "offset.h"

#define STACK_OFF_R0  		    48
#define STACK_OFF_R1     	    56
#define STACK_OFF_R2     	    64
#define STACK_OFF_R3     	    72
#define STACK_OFF_R4     	    80
#define STACK_OFF_R5     	    88
#define STACK_OFF_R6     	    96
#define STACK_OFF_R7     	   104
#define STACK_OFF_R8     	   112
#define STACK_OFF_R9     	   120
#define STACK_OFF_R10    	   128
#define STACK_OFF_R11    	   136
#define STACK_OFF_R12    	   144
#define STACK_OFF_R13    	   152
#define STACK_OFF_R14    	   160
#define STACK_OFF_R15    	   168
#define STACK_OFF_R16    	   176
#define STACK_OFF_R17    	   184
#define STACK_OFF_R18    	   192
#define STACK_OFF_R19    	   200
#define STACK_OFF_R20    	   208
#define STACK_OFF_R21    	   216
#define STACK_OFF_R22    	   224
#define STACK_OFF_R23    	   232
#define STACK_OFF_R24    	   240
#define STACK_OFF_R25    	   248
#define STACK_OFF_R26    	   256
#define STACK_OFF_R27    	   264
#define STACK_OFF_R28    	   272
#define STACK_OFF_R29    	   280
#define STACK_OFF_R30    	   288
#define STACK_OFF_R31    	   296

#define STACK_OFF_STATUS 	   304
#define STACK_OFF_HI     	   312
#define STACK_OFF_LO     	   320

#define STACK_OFF_BVADDR 	   328
#define STACK_OFF_CAUSE  	   336
#define STACK_OFF_EPC    	   344

#define K_STACK_SIZE   	           352

#define SAVE_AT                                  \
		.set	push;                            \
		.set	noat;                            \
		sd	    $1, PT_R1(sp);                   \
		.set	pop

#define SAVE_TEMP                                \
	 	mfhi	v1;                              \
		sd	    $8, PT_R8(sp);                   \
		sd	    $9, PT_R9(sp);                   \
		sd	    v1, PT_HI(sp);                   \
		mflo	v1;                              \
		sd	    $10,PT_R10(sp);                  \
		sd	    $11, PT_R11(sp);                 \
		sd	    v1,  PT_LO(sp);                  \
		sd	    $12, PT_R12(sp);                 \
		sd	    $13, PT_R13(sp);                 \
		sd	    $14, PT_R14(sp);                 \
		sd	    $15, PT_R15(sp);                 \
		sd	    $24, PT_R24(sp)

#define SAVE_STATIC                              \
		sd		$16, PT_R16(sp);                 \
		sd		$17, PT_R17(sp);                 \
		sd		$18, PT_R18(sp);                 \
		sd		$19, PT_R19(sp);                 \
		sd		$20, PT_R20(sp);                 \
		sd		$21, PT_R21(sp);                 \
		sd		$22, PT_R22(sp);                 \
		sd		$23, PT_R23(sp);                 \
		sd		$30, PT_R30(sp)

#define SAVE_SOME                                \
		.set	push;                            \
		.set	reorder;                         \
		move	k1, sp;                          \
        	move	k0, sp;                          \
		dsubu	sp, k1, PT_SIZE;                 \
		sd	    k0, PT_R29(sp);                  \
	        sd	    $3, PT_R3(sp);                   \
		sd	    $0, PT_R0(sp);			 		 \
		dmfc0	v1, CP0_STATUS;                  \
		sd	    $2, PT_R2(sp);                   \
		sd	    v1, PT_STATUS(sp);               \
		sd	    $4, PT_R4(sp);                   \
		dmfc0	v1, CP0_CAUSE;                   \
		sd	    $5, PT_R5(sp);                   \
		sd	    v1, PT_CAUSE(sp);                \
		sd	    $6, PT_R6(sp);                   \
		dmfc0	v1, CP0_EPC;                     \
		sd	    $7, PT_R7(sp);                   \
		sd		v1, PT_EPC(sp);                  \
		sd		$25, PT_R25(sp);                 \
		sd	    $28, PT_R28(sp);                 \
		sd	    $31, PT_R31(sp);                 \
		ori	    $28, sp, 0x1fff;                 \
		xori	$28, 0x1fff;                     \
		.set	pop

#define SAVE_ALL                             \
		SAVE_SOME;                           \
		SAVE_AT;                             \
		SAVE_TEMP;                           \
		SAVE_STATIC

#define RESTORE_AT                           \
		.set	push;                        \
		.set	noat;                        \
		ld		$1,  PT_R1(sp);              \
		.set	pop;

#define RESTORE_TEMP                         \
		ld		$24, PT_LO(sp);              \
		ld		$8, PT_R8(sp);               \
		ld		$9, PT_R9(sp);               \
		mtlo	$24;                         \
		ld		$24, PT_HI(sp);              \
		ld		$10,PT_R10(sp);              \
		ld		$11, PT_R11(sp);             \
		mthi	$24;                         \
		ld		$12, PT_R12(sp);             \
		ld		$13, PT_R13(sp);             \
		ld		$14, PT_R14(sp);             \
		ld		$15, PT_R15(sp);             \
		ld		$24, PT_R24(sp)

#define RESTORE_STATIC                       \
		ld		$16, PT_R16(sp);             \
		ld		$17, PT_R17(sp);             \
		ld		$18, PT_R18(sp);             \
		ld		$19, PT_R19(sp);             \
		ld		$20, PT_R20(sp);             \
		ld		$21, PT_R21(sp);             \
		ld		$22, PT_R22(sp);             \
		ld		$23, PT_R23(sp);             \
		ld		$30, PT_R30(sp)

#define RESTORE_SOME                         \
		.set	push;                        \
		.set	reorder;                     \
		dmfc0	t0, CP0_STATUS;              \
		.set	pop;                         \
		ori		t0, 0x1f;                    \
		xori	t0, 0x1f;                    \
		dmtc0	t0, CP0_STATUS;              \
		li	    v1, 0xff00;                  \
		and	    t0, v1;				         \
		ld	    v0, PT_STATUS(sp);           \
		nor	    v1, $0, v1;			         \
		and	    v0, v1;				         \
		or	    v0, t0;				         \
		dmtc0	v0, CP0_STATUS;              \
		ld	    v1, PT_EPC(sp);              \
		dmtc0	v1, CP0_EPC;                 \
		ld	    $31, PT_R31(sp);             \
		ld	    $28, PT_R28(sp);             \
		ld	    $25, PT_R25(sp);             \
		ld	    $7,  PT_R7(sp);              \
		ld	    $6,  PT_R6(sp);              \
		ld	    $5,  PT_R5(sp);              \
		ld	    $4,  PT_R4(sp);              \
		ld	    $3,  PT_R3(sp);              \
		ld	    $2,  PT_R2(sp)

#define RESTORE_SP_AND_RET           \
		ld	    sp,  PT_R29(sp);     \
		.set	mips3;				 \
		eret;					     \
		.set	mips0

#define RESTORE_SP                   \
		ld	sp,  PT_R29(sp)

#define RESTORE_ALL                  \
		RESTORE_SOME;                \
		RESTORE_AT;                  \
		RESTORE_TEMP;                \
		RESTORE_STATIC;              \
		RESTORE_SP

#define RESTORE_ALL_AND_RET          \
		RESTORE_SOME;                \
		RESTORE_AT;                  \
		RESTORE_TEMP;                \
		RESTORE_STATIC;              \
		RESTORE_SP_AND_RET

#define CLI                                     \
	mfc0    t0, CP0_STATUS;                 \
	li      t1, ST0_CU0 | 0x1f;             \
	or      t0, t1;                         \
	xori    t0, 0x1f;                       \
	mtc0    t0, CP0_STATUS


#ifndef __ASSEMBLY__

extern void ldr_flush_icache_all(void);
extern void ldr_flush_dcache_all (void);

#endif	/* __ASSEMBLY__ */

#ifndef XLP_SIM
#define NR_CPUS 32
#define PER_CPU_STACK_SIZE     8192
#define TOTAL_THREAD_SIZE       (PER_CPU_STACK_SIZE * (NR_CPUS))
#endif
#endif
