#ifndef __ASM_XEN_MM_H
#define __ASM_XEN_MM_H


/* 
 * TODO: A few of the following have to be made configurable
 *       by passing them to NetOS through Xen's start_info
 *       or shared_info structure
 */

#define XEN_PAGE_SHIFT	16
#define XEN_PAGE_SIZE   (1 << XEN_PAGE_SHIFT)

#define PGD_ORDER       0
#define PGD_T_SIZE      sizeof(unsigned long)

#define PMD_ORDER       0
#define PMD_T_SIZE      sizeof(unsigned long)

#define PTE_ORDER       0
#define PTE_T_SIZE      sizeof(unsigned long)

#define PTRS_PER_PGD    ((XEN_PAGE_SIZE << PGD_ORDER) / PGD_T_SIZE)
#define PTRS_PER_PMD	((XEN_PAGE_SIZE << PMD_ORDER) / PMD_T_SIZE)
#define PTRS_PER_PTE	((XEN_PAGE_SIZE << PTE_ORDER) / PTE_T_SIZE)

#define DOM_PAGE_SIZE   (1 << dom_page_shift)

#define PGDIR_SHIFT	    (pmd_shift + (XEN_PAGE_SHIFT + PMD_ORDER - 3))
#define PGDIR_SIZE	    (1UL << PGDIR_SHIFT)
#define PGDIR_MASK	    (~(PGDIR_SIZE-1))

#define PFN_SHIFT       12

#ifndef __ASSEMBLY__

#define pgd_index(address)	(((address) >> PGDIR_SHIFT) & (PTRS_PER_PGD-1))
#define pmd_offset(address)	(((address) >> pmd_shift) & (PTRS_PER_PMD - 1))
#define pte_offset(address)	(((address) >> dom_page_shift) & (PTRS_PER_PTE - 1))

typedef struct { unsigned long pgd; } pgd_t;
#define pgd_val(x)	((x).pgd)

typedef struct { unsigned long pmd; } pmd_t;
#define pmd_val(x)	((x).pmd)

typedef struct { unsigned long pte; } pte_t;
#define pte_val(x)	((x).pte)

extern pgd_t *current_pgd;
extern unsigned long *xen_m2p_table;
extern unsigned int pgtable_levels;
extern unsigned int dom_page_shift;
extern unsigned int pmd_shift;

static __inline__ pte_t *lookup_domain_pte(pgd_t *pgd, unsigned long mpaddr)
{
	pmd_t *pmd;
    pgd = pgd + pgd_index(mpaddr);
	pmd = (pmd_t *)pgd;

	if (pgtable_levels == 3)
		pmd = (pmd_t *)pgd_val(*pgd) + pmd_offset(mpaddr);

    return (pte_t *)pmd_val(*pmd) + pte_offset(mpaddr);
}

static __inline__ unsigned long __pa(unsigned long vaddr)
{
	unsigned long paddr, offset;

	paddr = pte_val(*(lookup_domain_pte(current_pgd, vaddr))) & ~((1 << PFN_SHIFT) - 1);
	offset = vaddr & (DOM_PAGE_SIZE - 1);
	
	return (paddr + offset);
}

static __inline__ unsigned long __va(unsigned long paddr)
{
	unsigned long vaddr, offset;

	vaddr = xen_m2p_table[(paddr >> dom_page_shift)] << dom_page_shift;
	offset = paddr & (DOM_PAGE_SIZE - 1);

	return (vaddr + offset);
}

#endif /* #ifndef __ASSEMBLY__ */

#endif /* #ifndef __ASM_XEN_MM_H */
