#include <stdio.h>
#include <stdlib.h>
#include <reent.h>
#include <errno.h>
#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include "spinlock.h"
#include "contig_alloc.h"
#include "contig_alloc_param.h"

int contig_memory_init(void *shmaddr, unsigned long long paddr, unsigned long long size)
{
	struct contig_mem *m;
	int rv = 0;
	extern int mbin_init(void *addr, int size);
	
	m = calloc(1, sizeof(struct contig_mem));
	if(m == NULL)
		return -1;

	contig_mem_ptr = m;

	m->paddr   = paddr;
	m->membase = shmaddr;
	m->sbrkaddr = shmaddr;
	m->size = size;
	m->trim_threshold = DEFAULT_TRIM_THRESHOLD;
	m->top_pad = DEFAULT_TOP_PAD;
	m->sbrk_base = (void *)-1;
	m->max_sbrked_mem = 0;
	m->max_total_mem = 0;
	memset(&m->current_mallinfo, 0, sizeof(struct mallinfo));

	rv = mbin_init(&m->av, MBIN_SZ);
	if(rv == 0) 
		m->initdone = 1;

	return rv;
}

int contig_alloc_sys_call(void *data)
{
	struct nlm_contig_alloc_call *m = data;

	//printf("contig_alloc_sys_call cmd = %d\n", m->cmd);
	switch(m->cmd)
	{

		case NLM_CONTIGALLOC_CALL:
		{
			struct nlm_contig_alloc *alloc = &m->alloc;
			alloc->memptr = (long)contig_alloc(alloc->size);
			break;
		}
		case NLM_CONTIGCALLOC_CALL:
		{
			struct nlm_contig_calloc *calloc = &m->calloc;
			calloc->memptr = (long)contig_calloc(calloc->nelem, calloc->elsize);
			break;
		}
		case NLM_CONTIGREALLOC_CALL:
		{
			struct nlm_contig_realloc *realloc = &m->realloc;
			realloc->memptr = (long)contig_realloc((void *)(long)realloc->obj, realloc->size);
			break;
		}
		case NLM_CONTIGMALIGN_CALL:
		{
			struct nlm_contig_malign *malign = &m->malign;
			malign->memptr = (long)contig_malign(malign->align, malign->size);
			break;
		}
		case NLM_CONTIGFREE_CALL:
		{
			struct nlm_contig_free *free = &m->free;
			contig_free((void *)(long)free->memptr);
			break;
		}
		case NLM_CONTIGPINFO_CALL:
		{
			struct nlm_contig_pinfo *pinfo = &m->pinfo;
			if(contig_mem_ptr == NULL) {
				m->rv = -1;
				return -1;
			}
			pinfo->physstart = contig_mem_ptr->paddr;
			pinfo->virtstart = (unsigned long long)contig_mem_ptr->membase;
			break;
		}

		default:
			m->rv = -1;
			return -1;
	}
	m->rv = 0;
	return 0;
}
