#include <stdio.h>
#include <stdlib.h>
#include <system.h>
#include <uart.h>

#include "bitops.h"

#include <nlm_hal.h>
#include "stubs.h"
#include "smp.h"

unsigned char cpu_stacks[XLx_NUM_CPUS][CPU_STACK_SIZE];
unsigned long phoenix_io_base __attribute__((weak));

#define CPU_STACK_START(cpu) \
	((unsigned long)&cpu_stacks[(cpu)][0] + CPU_STACK_SIZE - CPU_STACK_INITIAL_FRAME_SIZE)

#define ADDRESS_OF(x)   ((unsigned long)(char *)(x))
#define HEAP_SIZE       (ADDRESS_OF(_endaddr) - ADDRESS_OF(end))

unsigned long heap_start = ADDRESS_OF(end);
unsigned long heap_ptr   = ADDRESS_OF(end);

struct s_mem
{
	unsigned int size;
	unsigned int icsize;
	unsigned int dcsize;
};

unsigned int get_heap_size(void)
{
	return HEAP_SIZE;
}

/* Called from crt0.S, which gets
 * addr ranges from the linker script
 */
void get_mem_info(struct s_mem *mem)
{
	mem->size = _endaddr - _loadaddr;
}

void per_cpu_init(void)
{
	unsigned int cp0_status = 0x00;
	
	write_32bit_cp0_register(CP0_PAGEMASK, 0);
	write_32bit_cp0_register(CP0_WIRED, 0);

	write_32bit_cp0_register (CP0_COMPARE, 0);
	write_32bit_cp0_register (CP0_CAUSE, 0);
	
	cp0_status = (ST0_CU2 | ST0_CU0 | ST0_KX);
#ifdef XLP_SIM
	cp0_status |= (ST0_CU1 | ST0_FR);
#endif
#ifdef _ABI64
	cp0_status |= ST0_UX;
#endif
	write_32bit_cp0_register(CP0_STATUS, cp0_status);

#ifdef XLP_SIM
	write_32bit_cp1_ctrl_register (CP1_FCSR, 0x400000);
	cp0_status |= (ST0_CU1 | ST0_FR);
#else
	cp0_status |= 0x50000000;
#endif

	/* Extended MIPS registers */
	write_phnx_eimr(1 << IRQ_IPI_SMP_RESCHEDULE);
}

#ifdef CONFIG_XEN
#include "xen.h"
extern void xen_init(void);
#define hw_enable_cpus xen_enable_cpus
#else
static void xen_init(void) {}
#define hw_enable_cpus enable_cpus
#endif

extern void alloc_per_cpu_mem_space(void);
volatile unsigned int cpu_present_map;


/* 
 * Arch-dependent routine called from crt0.S
 */
void hardware_init_hook(void)
{
	int cpu = 0;

	/* Initialize the Status Register
	 */
	per_cpu_init();

#ifdef XLP_SIM
	nlm_hal_init();

	phoenix_io_base = xlp_io_base;
#else
	phoenix_io_base = (unsigned long)0xffffffffbef00000ULL;
#endif

	/* 
	 * Parse information from the FDT to populate 'cpu_online_map'
	 */
	if (process_fdt() < 0) {
		printf("Failed to process FDT!\n");
		cpu_halt();
	}

	alloc_per_cpu_mem_space();

	xen_init();

	/* TODO: to add common_ebase here ? */

	/* 
	 * Initializes uart read/write function locks through 
	 * the dev->open/rd/wr/close calls
	 */
	syscall_init(cpu);

	cli(); /* clear intr */
	init_irq();
	ldr_trap_init();
	sti(); /* enable intr */

	/* Wakeup selected cpus
	 */
	cpu_present_map = 0;
	set_bit(processor_id(), &cpu_present_map);

	hw_enable_cpus(cpu_online_map, (long) xlr_park);

	printf("Checking if siblings (cpu_online_map=%x) are alive...\n", cpu_online_map);
	while (cpu_present_map != cpu_online_map) {
		static uint64_t count = 0;
		++count;
		if ( (count % 1000000000) == 0) {
			printf("Waiting for siblings (cpu_online_map = %x cpu_present_map = %x, to wakeup...!\n",
			       cpu_online_map, cpu_present_map);
		}
	}
	printf("(%x) siblings woke up!\n", cpu_present_map);
}
