#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <reent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <setjmp.h>
#include <system.h>
#include <mips-exts.h>	/* for cpu id */
#include <devices.h>

#include "stubs.h"

#define MAX_DEVICES 	(sizeof(dev_ops) / sizeof(dev_ops[0]))

int errno;

static struct device_ops *dev_ops[] = {
	[0] = &dev_stdio_in,
	[1] = &dev_stdio_out,
	[2] = &dev_stdio_err,
	[3] = &dev_tmpfs
};

void syscall_init(int cpu)
{
	int i = 0;

	for (i = 0; i < MAX_DEVICES; i++)
		dev_ops[i]->init();
}

int _fstat(int fd, struct stat *buf)
{
	buf->st_mode = S_IFCHR;	/* Always pretend to be a tty */
	buf->st_blksize = 0;

	return (0);
}

/* getpid -- only one process, so just return 1.
 */
int _getpid(void)
{
	return processor_id();
}

/*
 * isatty -- returns 1 if connected to a terminal device,
 *           returns 0 if not. Since we're hooked up to a
 *           serial port, we'll say yes _AND return a 1.
 */
int _isatty(int fd)
{
	return 1;
}

/*
 * kill -- go out via exit...
 */
int _kill(int pid, int sig)
{
#if 0
	if(pid == __MYPID) 
		_exit(sig); 
#endif
	errno = EINVAL;
	return -1;
}

int _link(const char *old, const char *new)
{
	errno = EMLINK;
	return -1;
}

/*
 * lseek --  Since a serial port is non-seekable, we return an error.
 */
_off_t _lseek(int fd, _off_t pos, int whence)
{
	errno = ESPIPE;
	return ((off_t)-1);
}


int _execve(const char *name, char * const *argv, char * const *env)
{
	errno = ENOMEM;

	return -1;
}

int _fcntl(int fd, int cmd, int arg)
{
	int ret = -1;

	return ret;
}

/*
 * stat -- Since we have no file system, we just return an error.
 */
int _stat(const char *file, struct stat *pstat)
{
	errno = EIO;
	return (-1);
}

_CLOCK_T_ _times(struct tms *ptms)
{
	clock_t ret = -1;

	return ret;
}

/*
 * unlink -- since we have no file system, 
 *           we just return an error.
 */
int _unlink(const char *file)
{
	errno = ENOENT;
	return -1;
}

int _wait(int *status)
{
	errno = ECHILD;
	return -1;
}

static struct _reent _rmi_impure_ptrs[XLx_NUM_CPUS] = {
	[0] = _REENT_INIT(_rmi_impure_ptrs[0]),
	[1] = _REENT_INIT(_rmi_impure_ptrs[1]),
	[2] = _REENT_INIT(_rmi_impure_ptrs[2]),
	[3] = _REENT_INIT(_rmi_impure_ptrs[3]),
	[4] = _REENT_INIT(_rmi_impure_ptrs[4]),
	[5] = _REENT_INIT(_rmi_impure_ptrs[5]),
	[6] = _REENT_INIT(_rmi_impure_ptrs[6]),
	[7] = _REENT_INIT(_rmi_impure_ptrs[7]),
	[8] = _REENT_INIT(_rmi_impure_ptrs[8]),
	[9] = _REENT_INIT(_rmi_impure_ptrs[9]),
	[10] = _REENT_INIT(_rmi_impure_ptrs[10]),
	[11] = _REENT_INIT(_rmi_impure_ptrs[11]),
	[12] = _REENT_INIT(_rmi_impure_ptrs[12]),
	[13] = _REENT_INIT(_rmi_impure_ptrs[13]),
	[14] = _REENT_INIT(_rmi_impure_ptrs[14]),
	[15] = _REENT_INIT(_rmi_impure_ptrs[15]),
	[16] = _REENT_INIT(_rmi_impure_ptrs[16]),
	[17] = _REENT_INIT(_rmi_impure_ptrs[17]),
	[18] = _REENT_INIT(_rmi_impure_ptrs[18]),
	[19] = _REENT_INIT(_rmi_impure_ptrs[19]),
	[20] = _REENT_INIT(_rmi_impure_ptrs[20]),
	[21] = _REENT_INIT(_rmi_impure_ptrs[21]),
	[22] = _REENT_INIT(_rmi_impure_ptrs[22]),
	[23] = _REENT_INIT(_rmi_impure_ptrs[23]),
	[24] = _REENT_INIT(_rmi_impure_ptrs[24]),
	[25] = _REENT_INIT(_rmi_impure_ptrs[25]),
	[26] = _REENT_INIT(_rmi_impure_ptrs[26]),
	[27] = _REENT_INIT(_rmi_impure_ptrs[27]),
	[28] = _REENT_INIT(_rmi_impure_ptrs[28]),
	[29] = _REENT_INIT(_rmi_impure_ptrs[29]),
	[30] = _REENT_INIT(_rmi_impure_ptrs[30]),
	[31] = _REENT_INIT(_rmi_impure_ptrs[31]),
};

struct _reent *__getreent(void)
{
	return &_rmi_impure_ptrs[processor_id()];
}

int _execve_r(struct _reent *ptr, const char *name, char * const *argv, char * const *env)
{
	int ret;

	errno = 0;
	if ((ret = _execve (name, argv, env)) == -1 && errno != 0)
		ptr->_errno = errno;
	return ret;
}

int _fcntl_r(struct _reent *ptr, int fd, int cmd, int arg)
{
	int ret;

	errno = 0;
	if ((ret = _fcntl (fd, cmd, arg)) == -1 && errno != 0)
		ptr->_errno = errno;
	return ret;
}

int _fstat_r(struct _reent *ptr, int fd, struct stat *pstat)
{
	int ret;

	errno = 0;
	if ((ret = _fstat (fd, pstat)) == -1 && errno != 0)
		ptr->_errno = errno;
	return ret;
}

int _getpid_r(struct _reent *ptr)
{
	int ret;
	ret = _getpid ();
	return ret;
}

int _isatty_r(struct _reent *ptr, int fd)
{
	int ret;

	errno = 0;
	if ((ret = _isatty (fd)) == -1 && errno != 0)
		ptr->_errno = errno;
	return ret;
}

int _kill_r(struct _reent *ptr, int pid, int sig)
{
	int ret;

	errno = 0;
	if ((ret = _kill (pid, sig)) == -1 && errno != 0)
		ptr->_errno = errno;
	return ret;
}

int _link_r(struct _reent *ptr, const char *old, const char *new)
{
	int ret;

	errno = 0;
	if ((ret = _link (old, new)) == -1 && errno != 0)
		ptr->_errno = errno;
	return ret;
}

_off_t _lseek_r(struct _reent *ptr, int fd, _off_t pos, int whence)
{
	_off_t ret;

	errno = 0;
	if ((ret = _lseek (fd, pos, whence)) == (_off_t) -1 && errno != 0)
		ptr->_errno = errno;
	return ret;
}

int _open_r(struct _reent *ptr, const char *file, int flags, int mode)
{
	int fd = -1;

	for (fd = 0; MAX_DEVICES; fd++) {
		if (strcmp(dev_ops[fd]->name, file) == 0) {
			/* found the device */
			return dev_ops[fd]->open_r(ptr, file, flags, mode);
		}
	}
	/* device not found */
	ptr->_errno = ENODEV;
	return -1;
}

int _close_r(struct _reent *ptr, int fd)
{
	if (fd >= 0 && fd < MAX_DEVICES)
		return dev_ops[fd]->close_r(ptr, fd);

	ptr->_errno = ENODEV;
	return -1;
}

_ssize_t _read_r(struct _reent *ptr, int fd, void *buf, size_t cnt)
{
	if (fd >= 0 && fd < MAX_DEVICES)
		return dev_ops[fd]->read_r(ptr, fd, buf, cnt);

	ptr->_errno = ENODEV;
	return -1;
}

_ssize_t _write_r(struct _reent *ptr, int fd, const void *buf, size_t cnt)
{
	if (fd >= 0 && fd < MAX_DEVICES)
		return dev_ops[fd]->write_r(ptr, fd, buf, cnt);

	ptr->_errno = ENODEV;
	return -1;
}

void *_sbrk(ptrdiff_t nbytes)
{
	char        *base = 0;
	unsigned int avail = 0;
	unsigned int heap_size=0;

	heap_size = get_heap_size();

	if (((size_t)heap_ptr >= heap_start) && 
			((size_t)heap_ptr < (heap_start + heap_size))) {
		avail = (heap_start + heap_size) - (size_t)heap_ptr;
		base = (char *)heap_ptr;
	} 

	if ((nbytes > avail) || (heap_ptr + nbytes < heap_start))
		base = (char *)-1;
	else
		heap_ptr += nbytes;

	return base;
}

void *_sbrk_r(struct _reent *ptr, ptrdiff_t incr)
{
	char *ret;
	void *_sbrk(ptrdiff_t);

	errno = 0;
	if ((ret = (char *)(_sbrk (incr))) == (void *) -1 && errno != 0)
		ptr->_errno = errno;
	return ret;
}

int _stat_r(struct _reent *ptr, const char *file, struct stat *pstat)
{
	int ret;

	errno = 0;
	if ((ret = _stat (file, pstat)) == -1 && errno != 0)
		ptr->_errno = errno;
	return ret;
}

_CLOCK_T_ _times_r(struct _reent *ptr, struct tms *ptms)
{
	clock_t ret;

	ret = _times (ptms);
	return ret;
}

int _unlink_r(struct _reent *ptr, const char *file)
{
	int ret;

	errno = 0;
	if ((ret = _unlink (file)) == -1 && errno != 0)
		ptr->_errno = errno;
	return ret;
}

int _wait_r(struct _reent *ptr, int *status)
{
	int ret;

	errno = 0;
	if ((ret = _wait (status)) == -1 && errno != 0)
		ptr->_errno = errno;
	return ret;
}
