#include <uart.h>
#include "xlr_iomap.h"
#include "mips-exts.h"

#define KSEG0 0xffffffff80000000UL

static volatile uint32_t *xlp_uart_mmio;

static int vuart;
static int uart_init_done;

static void p_uart_init(int id)
{
	unsigned int offset = 0;

	if (id == 0)
		offset = PHOENIX_IO_UART_0_OFFSET;
	else
		offset = PHOENIX_IO_UART_1_OFFSET;

#if defined(_MIPS_ARCH_XLP)
	xlp_uart_mmio = (volatile uint32_t *)(DEFAULT_XLP_IO_BASE + offset);
#elif defined(_MIPS_ARCH_XLR)
	xlp_uart_mmio = (volatile uint32_t *)(DEFAULT_PHOENIX_IO_BASE + offset);
#else
#error "Unknown MIPS_ARCH (wrong -march passed)"
#endif

	/* Set up the baud rate */
	xlp_uart_mmio[UART_LCR]    = xlp_uart_mmio[UART_LCR] | (1<<7);
	xlp_uart_mmio[UART_DLB_1]  = UART_BR_DLB1;
	xlp_uart_mmio[UART_DLB_2]  = UART_BR_DLB2;
	xlp_uart_mmio[UART_LCR]    = xlp_uart_mmio[UART_LCR] & ~(1<<7);

	uart_init_done = 1;
}

static int p_inbyte(void)
{
	int    data = 0;
	int    lsr = 0;

	if (!uart_init_done) return 0;

	for (;;) {
		lsr = xlp_uart_mmio[UART_LSR];

		if (lsr & 0x80) {    /* parity/frame/break - push a 0! */
			data = 0;
			break;
		}
		if (lsr & 0x01) {    /* Rx Data */
			data =
				xlp_uart_mmio[UART_RHR];
			break;
		}
	}

	return (char)data;
}

#define RING_BUFFER_BASE (511 << 20)
#define RING_BUFFER_SIZE (8 << 10)
static void outbyte_ring_buffer(char c)
{
	unsigned long base = RING_BUFFER_BASE + (processor_id() * RING_BUFFER_SIZE);
	char *buf = (char *)KSEG0 + base;
	static int idx;

	buf[idx] = c;
	idx = (idx + 1) % (RING_BUFFER_SIZE);
}

static void p_outbyte(char c)
{
	int lsr = 0;

	if (!uart_init_done) {

		outbyte_ring_buffer(c);
		return;
	}

	for(;;) {
		lsr = xlp_uart_mmio[UART_LSR];
		if (lsr & 0x20) break;
	}
	xlp_uart_mmio[UART_THR] = c;
}

// -----------
//  extern API
// -----------

extern void v_outbyte(char c);
extern int v_inbyte(void);
extern void vuart_update_params(uint64_t sharedcfg, int id);

int inbyte(void)
{
	if (vuart) return v_inbyte();

	return p_inbyte();
}

void outbyte(char c)
{
	if (vuart) {
		v_outbyte(c);
		return;
	}

	p_outbyte(c);
}

void uart_update_params(uint64_t sharedcfg, int id)
{
	if (sharedcfg) vuart = 1;

	if (vuart)
		vuart_update_params(sharedcfg, id);
	else
		p_uart_init(id);
}
