#include <stdio.h>
#include <inttypes.h>
#ifndef __LINUX__
#include "mips-exts.h"
#endif

#include "queue.h"
#include "vuart.h"

#define KSEG0 0xffffffff80000000UL

/* Read from FDT file */
static unsigned long uart_sharedcfg_base;
static int port;
static int uart_init_done;

queue_t *vuart_queues;

static void __uart_init(void *vuart_baseaddr, int nr_vuart, int initq)
{
	int i;

	vuart_queues = (queue_t *)vuart_baseaddr;

	if (initq) {
		for (i = 0; i < nr_vuart; i += 2 ) {
			/* 2 = 1 producer Q + 1 consumer Q */
			init_queue(&vuart_queues[i], MAX_ELEMENTS);
			init_queue(&vuart_queues[i+1], MAX_ELEMENTS);
		}
	}
}

static int __inbyte(int vuart, int qid)
{
	int c;
	queue_t *queue = &vuart_queues[(vuart * 2) + qid];

	if (queue->init_done != 1)
		return EOF;

	while (dequeue(queue, &c) != 0)
		;

	return c;
}

static int __inbyte_nowait(int vuart, int qid)
{
	int c;
	queue_t *queue = &vuart_queues[(vuart * 2) + qid];

	if (queue->init_done != 1)
		return EOF;

	if (dequeue(queue, &c) != 0)
		return EOF;

	return c;
}

static void __outbyte(char c, int vuart, int qid)
{
	queue_t *queue = &vuart_queues[(vuart * 2) + qid];

	if (queue->init_done != 1)
		return;

	while (enqueue(queue, c) != 0)
		;

	return;
}

#define RING_BUFFER_BASE (511 << 20)
#define RING_BUFFER_SIZE (8 << 10)
static void outbyte_ring_buffer(char c)
{
	unsigned long base = RING_BUFFER_BASE + (processor_id() * RING_BUFFER_SIZE);
	char *buf = (char *)KSEG0 + base;
	static int idx = 0;

	buf[idx] = c;
	idx = (idx + 1) % (RING_BUFFER_SIZE);
}

static void vuart_init(void)
{
	__uart_init((void *)uart_sharedcfg_base, NR_VUART, 0);

	uart_init_done = 1;
}

// -----------
//  extern API
// -----------

void v_outbyte(char c)
{
	if (!uart_init_done) {

		outbyte_ring_buffer(c);
		return;
	}

	__outbyte(c, port, DOMU_WRITE_Q);
}

int v_inbyte(void)
{
	if (!uart_init_done) return EOF;

	return __inbyte(port, DOMU_READ_Q);
}

void vuart_update_params(uint64_t sharedcfg, int id)
{
	uart_sharedcfg_base = KSEG0 + (sharedcfg & 0x1fffffff);
	port = id;

	printf("(NetOS) cpu-%d q=%p\n", processor_id(), &vuart_queues[(port * 2) + DOMU_WRITE_Q]);

	vuart_init();

	printf("sharedcfg=%llx, uart_sharedcfg_base=%lx, port = %d\n",
	       (unsigned long long)sharedcfg, uart_sharedcfg_base, port);
}
