#include <stdio.h>
#include <string.h>
#include <system.h>
#include <mipsregs.h>
#include <xen/mm.h>
#include <xen/interface/xen.h>
#include <xen/interface/domctl.h>
#include <asm/xen/hypercall.h>
#include "bitops.h"

extern unsigned char _endaddr[];

#define ALIGN(val, alignment) ((val + alignment - 1) & ~(alignment - 1))
#define PTR_ALIGN(val) ALIGN(val, sizeof(void *))

pgd_t *current_pgd;
unsigned long *xen_m2p_table;
unsigned int pgtable_levels;
unsigned int dom_page_shift = 22;
unsigned int pmd_shift;

unsigned long gpctx[NCPU];

static int xc_wakeup_secondary(int dom_id, unsigned long cpumask, unsigned long entry)
{
	struct xen_domctl domctl;
	struct vcpu_guest_context ctxt;
	int vcpu, ret, nr_vcpu;
	
	memset((char *)&domctl, 0, sizeof(domctl));
	memset((char *)&ctxt, 0, sizeof(ctxt));
	domctl.interface_version = XEN_DOMCTL_INTERFACE_VERSION;
	domctl.cmd = XEN_DOMCTL_bootvcpu;
    domctl.domain = dom_id;

	ctxt.reg31 = entry;
    set_xen_guest_handle(domctl.u.vcpucontext.ctxt, &ctxt);

	nr_vcpu = num_ones(cpumask);

	for (vcpu = 1; vcpu < nr_vcpu; ++vcpu) {
		domctl.u.vcpucontext.vcpu = vcpu;
		ret = privcmd_call(__HYPERVISOR_domctl2, (unsigned long)&domctl, 0, 0, 0, 0);
		if (ret != 0) {
			printf("(%s): ret value of privcmd_call = %d\n", __func__, ret);
			while (1)
				;
		}
	}

	return 0;
}

extern void (*cpu_park_func) (unsigned long);
extern unsigned long prom_pre_boot_secondary_cpus[];

void xen_enable_cpus(unsigned long cpumask, unsigned long fptr)
{
	cpu_park_func = (void *)fptr;

	xc_wakeup_secondary(dom_id, cpumask, (unsigned long)prom_pre_boot_secondary_cpus);
}

struct shared_info *xen_shared_info = NULL;
struct start_info *xen_start_info;

static char *banner = "Booting paravirtualized NetOS\n";

void xen_init(void)
{
	xen_start_info = (struct start_info *)ALIGN((unsigned long)&_endaddr, XEN_PAGE_SIZE);

	xen_shared_info = (struct shared_info *)xen_start_info->shared_info;

	current_pgd = (pgd_t *) xen_start_info->pt_base;

	xen_m2p_table = (unsigned long *) xen_start_info->mfn_list;

	pgtable_levels = xen_shared_info->arch.pgtable_levels;

	dom_page_shift = xen_shared_info->arch.dom_page_shift;

	if (pgtable_levels == 3)
		pmd_shift = (dom_page_shift + (XEN_PAGE_SHIFT + PTE_ORDER - 3));
	else
		pmd_shift = dom_page_shift = xen_shared_info->arch.dom_page_shift;

	printf("%s", banner);
}
