#include <stdarg.h>
#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "xlr_boot_lib.h"
#include "elf-ext.h"
#include "system.h"
#include "types.h"
#include "mips-exts.h"

#include "boot2.h"
#include "cpu.h"
#include "boot_mgr.h"

#include "sys/module.h"
#include "atomic.h"
#include "rmidelay.h"
#include "msgring.h"
#include "nlm_hal_macros.h"
#include "shell.h"
#include "uart.h"
#include "nlm_hal.h"
#include "stackframe.h"
#include "bitops.h"

#include <cache.h>
#ifndef XLP_SIM
#include "mipscopreg.h"
#include "interrupt.h"
#endif


static void override_app_args(unsigned char *argstr);
static void free_lib_args();

extern uint32_t msgrng_tm_get_cpu_mask(void);
extern void time_init(void);
extern void console_init(void);
extern void xlr_cpu_status(int argc, char *argv[]);
extern void stop_vcpu(int argc, char *argv[]);
extern void xlr_park(unsigned long args);

extern int xlr_load_userapp(unsigned int cpu, int argc, char *argv[],char **envs);
extern int xlr_launch_userapp(int cpu);
extern void setup_buddy_cpu_section(int , uint32_t);
#ifdef NETOS_LOADER_TODO
extern void nmi_init(void);
#endif /* NETOS_LOADER_TODO */
extern void xlr_lib_set_userapp_mask(unsigned long mask);
extern void memcpy_virt_phys(unsigned char *dest, uint64_t src, int nbytes);
extern void xlr_free_memory(uint64_t start, uint64_t size);
extern uint64_t xlr_alloc_memory(uint64_t *size);
extern void loader_wakeup_cpu(unsigned int cpu, unsigned long fn, unsigned long args);

int file_read=0;
int global_debug = 0;
#define MAX_ENV  32
#define MAX_ARGS 32
int smp_mode = 0;

static int nargs;
char *envs[MAX_ENV];
static char *args[MAX_ARGS];

uint32_t cpu_online_map;
unsigned long cpu_io_base = (unsigned long)(DEFAULT_CPU_IO_BASE);

struct thread_info threads_info[MAX_CPUS] __section(".thread.info");

typedef void (*entry_fn) (int, char *[], char *[], uint64_t);
extern void shutdown(void);

extern void xlp_dev_init();

static struct processor *this_processor;

unsigned int nlm_cpu_frequency;
unsigned long common_ebase = 0x0;

#ifndef XLP_SIM

/* Struct for temp. allocation
 * of sp/gp for secondary CPUs
 */
struct xlr_stack_pages {
	unsigned long stack[(TOTAL_THREAD_SIZE)/sizeof(long)];
};

struct xlr_stack_pages xlr_stack_pages_temp
__attribute__((__section__(".data.init_task"),
			__aligned__(PER_CPU_STACK_SIZE)));

unsigned long xlr_secondary_entry_point;

/* This function just prepares for waking up the cpus. Actual wakeup is done
   only when the cpu mask in know (at application launch time)
   */
static void xlr_put_non_vcpu0_in_park(void)
{
	void *base;
	extern char xlr_boot_smp_nmi;

	printf("Setting up NMI vector for siblings wakeup\n");

	base = (void *)(unsigned long)0xffffffffbfc00000ULL;
	memcpy(base, &xlr_boot_smp_nmi, 0x100); /* TODO check the copy size */
}

void enable_cpus(unsigned long thread_bitmask, unsigned long park_func)
{
	int myself = processor_id();

	/* exclude this cpu from the mask */
	thread_bitmask &= ~(1UL << myself);

	xlr_secondary_entry_point = park_func;

	/*Wakeup All Other CPUS and make them poll on xlr_wakeup_info.*/
	xlr_put_non_vcpu0_in_park();

	send_nmi_ipi((uint32_t)thread_bitmask);
}
#endif


void phnx_initialize_getopt(void)
{
	optind = 0;
}

struct processor *get_processor_info(void)
{
	return (this_processor);
}

uint64_t dwnld_addr = 0;
uint64_t dwnld_size = (32<<20);
int  verbose;

/* do not reset mac and ring (to start linux after rmios) */
static int noreset = 0;

static int lib_argc;
static char *lib_str_args = NULL;
static char *lib_argv[MAX_ARGS];

/* NTEOS_LOADER_TODO : where is this structure initialized */
static struct bootfile_ bootfile = {
	.name = 0,
	.start = 0,
	.max_size = BOOTFILE_MAX_SIZE,
	.size = 0,
	.valid = 0
};

struct boot1_info *boot1_info;

static struct boot1_info hack_boot1_info =  /* This is needed until FDT is
											   integrated into netos */
{
	.size                   = 0x0000000000000138,
	.psb_version            = 0x4bb393dc00000106,
	.mac_addr               = 0x0000000102030405,
	.board_version          = 0x0000000000000014,
	.global_shmem_addr      = 0xffffffff800ff000,
	.global_shmem_size      = 0x0000000000001000,
	.board_major_version    = 0x0000000000000014,
	.board_minor_version    = 0x0000000000000000,
	.board_manf_revision    = 0x000000000000ffff,
	.board_serial_number    = 0x000000000000ffff,
};


static char *userapp_usage = 
"----------------------------------------------------------\n"
"userapp <options>\n"
"   -A     args to be passed to the application.\n"
"   -h     help\n"
"   -T     The Uart that the application uses (uart0/uart1)\n"
"----------------------------------------------------------\n";

static char *userapp_os_usage = 
"userapp_os <options>\n"
"----------------------------------------------------------\n"
"   -a     elfload addr (default is last downloaded bootfile)\n"
"   -e     userapp addr (physical) to elfload from\n"
"   -M     max legitimate physical addr (in hex) \n"
"   -T     uart that the application uses (uart0/uart1)\n"
"   -A     args to be passed to the application.\n"
"   -h     help\n"
"----------------------------------------------------------\n";
struct bootfile_ *get_bootfile()
{
	return (&bootfile);
}

enum {
	PROM_PRIMARY_REENTER,
	PROM_SECONDARY_REENTER,
	PROM_BOOT_SECONDARY
};

uint32_t cpus_in_kuseg_mode = 0x0;

static uint32_t userapp_cpu_map = 0xffffffff;
static atomic_t userapp_init_smp_prep_count;

static void drain_messages(void) {
#ifndef XLP_SIM
	const int num_retries = 8;
	int bucket = phnx_thr_id();
	struct msgrng_msg msg;
	int i = 0, size = 0, code = 0, stid = 0;

	/* wait and drain all messages 
	 */
	for (i = 0; i < num_retries; i++) {
		while (!message_receive(bucket, &size, &code, &stid, &msg)) ;
		rmi_udelay(10000);
	}
#endif
}

static void userapp_init_smp_prep(void)
{
	struct cpu *this_cpu = get_cpu_info();

	drain_messages();

	this_cpu->cpu_init(this_cpu->cpu_data);

	/* disable 64-bit addressing 
	 */
	{
		unsigned int status = read_32bit_cp0_register(CP0_STATUS);
		write_32bit_cp0_register(CP0_STATUS, (status & ~ST0_KX));
	}
}

static void smp_call_function(void *fn, void *args, uint32_t cpu_mask)
{
	int i = 0;
	int count = 0;

	cpu_mask &= ~(1 << processor_id());
	for (i = 0; i < 32; i++) {
		if ((cpu_online_map & (1 << i)) && (cpu_mask & (1 << i))) {
			count++;

			loader_wakeup_cpu(i, (unsigned long)fn, (unsigned long)args);
		}
	}
}

static void smp_call_function_and_barrier_sync_helper(void (*fn) (void))
{
	if (fn)
		(*fn) ();
	atomic_add(1, &userapp_init_smp_prep_count);
}

static void smp_call_function_and_barrier_sync(void *fn,
					       uint32_t cpu_mask,
					       const char *func_name,
					       int init_count, int wait_count)
{
	int i = 0;
	const int num_retries = 32;
	int num_delay_loops = 10000000;

#if defined(PHOENIX_SIM) || defined(XLP_SIM)
	num_delay_loops = 700000;
#endif

	/* rendevouz all cpus, by sending messages and wait for them to ack */
	atomic_set(&userapp_init_smp_prep_count, init_count);
	smp_call_function(smp_call_function_and_barrier_sync_helper, fn,
			  cpu_mask);

	for (i = 0; i < num_retries; i++) {

		if (atomic_read(&userapp_init_smp_prep_count) == wait_count)
			break;
		
		rmi_udelay((unsigned long int)num_delay_loops);
		
		if (atomic_read(&userapp_init_smp_prep_count) == wait_count)
			break;
	}

	printf("count = %d, total = %d\n",
	       atomic_read(&userapp_init_smp_prep_count), wait_count);

	if (atomic_read(&userapp_init_smp_prep_count) != wait_count) {
		printf("Only %d of %d cpus ack'ed %s, oh well...\n",
		       atomic_read(&userapp_init_smp_prep_count), wait_count,
		       func_name);
	} else {
		printf("All slave cpus (%d) ack'ed %s\n",
		       atomic_read(&userapp_init_smp_prep_count), func_name);
	}

	atomic_set(&userapp_init_smp_prep_count, 0);
}

static struct bucket_size bucket_sizes_reset = {
	{
		32, 32, 32, 32, 32, 32, 32, 32,
		32, 32, 32, 32, 32, 32, 32, 32,
		32, 32, 32, 32, 32, 32, 32, 32,
		32, 32, 32, 32, 32, 32, 32, 32,
		32, 32, 32, 32, 32, 32, 32, 32,
		32, 32, 32, 32, 32, 32, 32, 32,
		32, 32, 32, 32, 32, 32, 32, 32,
		32, 32, 32, 32, 32, 32, 32, 32,
		32, 32, 32, 32, 32, 32, 32, 32,
		32, 32, 32, 32, 32, 32, 32, 32,
		32, 32, 32, 32, 32, 32, 32, 32,
		32, 32, 32, 32, 32, 32, 32, 32,
		32, 32, 32, 32, 32, 32, 32, 32,
		32, 32, 32, 32, 32, 32, 32, 32,
		32, 32, 32, 32, 32, 32, 32, 32,
		32, 32, 32, 32, 32, 32, 32, 32,
	}
};

static struct stn_cc cc_table_reset;
static struct stn_cc cc_table_reset_cpu_0 = { {
	{4, 4, 4, 4, 0, 0, 0, 0},
	{4, 4, 4, 4, 0, 0, 0, 0},
	{4, 4, 4, 4, 0, 0, 0, 0},
	{4, 4, 4, 4, 0, 0, 0, 0},
	{4, 4, 4, 4, 0, 0, 0, 0},
	{4, 4, 4, 4, 0, 0, 0, 0},
	{4, 4, 4, 4, 0, 0, 0, 0},
	{4, 4, 4, 4, 0, 0, 0, 0}
}
};

static void message_ring_cpu_reset(void)
{
	cc_table_cpu_0 = cc_table_reset_cpu_0;
	cc_table_cpu_1 = cc_table_reset;
	cc_table_cpu_2 = cc_table_reset;
	cc_table_cpu_3 = cc_table_reset;
	cc_table_cpu_4 = cc_table_reset;
	cc_table_cpu_5 = cc_table_reset;
	cc_table_cpu_6 = cc_table_reset;
	cc_table_cpu_7 = cc_table_reset;

	cc_table_xgs_0 = cc_table_reset;
	cc_table_xgs_1 = cc_table_reset;
	cc_table_gmac = cc_table_reset;
	cc_table_dma = cc_table_dma;
	cc_table_sec = cc_table_sec;

	bucket_sizes = bucket_sizes_reset;
}
void flush_tlb_all(void)
{
#if !defined(PHOENIX_SIM) && !defined(XLP_SIM)
        unsigned long flags = 0;
        unsigned long old_ctx = 0;
        unsigned int entry = 0;
        unsigned int tlbsize=10;

        unsigned int config1=0;

        config1 = read_c0_config1();
        tlbsize = ((config1 >> 25) & 0x3f) + 1;

        cpu_save_and_disable_int(flags);

        /* Save old context and create impossible VPN2 value */
        old_ctx = (get_cop_entryhi_reg() & 0xff);

        set_cop_entrylo_0_reg(0);
        set_cop_entrylo_1_reg(0);
        for (entry = 0; entry < tlbsize; entry++)
        {
                set_cop_entryhi_reg((unsigned long)KSEG0 + (PAGE_SIZE << 1) * entry);
                set_cop_index_reg(entry);
                mips_tlb_write_indexed();
        }
        set_cop_entryhi_reg(old_ctx);

        cpu_restore_int(flags);
#endif
}

static void kseg0_prep(void)
{
#ifdef NETOS_LOADER_ENHANCE
	struct net_device *netdev = 0;
#endif /* NETOS_LOADER_ENHANCE */
	uint32_t usable_cpu_mask = 0;

	struct processor *this_processor = get_processor_info();

	usable_cpu_mask = boot1_info->userapp_cpu_map;	
	printf("Preparing %x bitmask of cpus to run \n", usable_cpu_mask);
    
	/* flush I and D cache: 
	 *  If elf file loaded has exception vectors
	 *  do not take any exceptions here.
	 */
	ldr_flush_icache_all();
	ldr_flush_dcache_all();

#ifdef NETOS_LOADER_ENHANCE

	/* soft reset mac */
	netdev = PhoenixMacNetDevice();
	if (!netdev) {
		printf("No network device to cleanup\n");
	}else {
		if (!noreset) {
			printf("Resetting %s ...\n", netdev->name);
			if (!netdev->soft_reset(netdev)) {
				printf("Unable to reset %s!\n",netdev->name);
			} else {
				printf("...done\n");
			}
		}
	}
#endif /* NETOS_LOADER_ENHANCE */

	smp_call_function_and_barrier_sync(&userapp_init_smp_prep,
					   usable_cpu_mask, "userapp init", 1,
					   num_ones(usable_cpu_mask));
	drain_messages();

	if (!noreset)
		/* reset message ring config */
		message_ring_cpu_reset();

	smp_call_function_and_barrier_sync(this_processor->cpu_ipc_init,
					   (usable_cpu_mask & 0x11111111), "message ring init", 
					   1, num_ones(0x11111111 & usable_cpu_mask));

	if (!noreset) {
		this_processor->cpu_ipc_init(this_processor);
		/* at this point messages in only a few paths work */
#ifdef NETOS_LOADER_ENHANCE
		if (netdev) {
			netdev->init_msgring(netdev);
		}
#endif /* NETOS_LOADER_ENHANCE */
	}

	if (processor_id() == 0) {
		/* Reset the first available index of tlb to 0, 
		 * so we have enough tlbs. Write 0 to wired register. 
		 * That should do the trick.
		 */
		flush_tlb_all();
		write_32bit_cp0_register(CP0_WIRED, 0);
	}

#ifdef NETOS_LOADER_ENHANCE
	if(usb_stop)
		usb_stop();
#endif /* NETOS_LOADER_ENHANCE */
}

void xlr_lib_prelaunch_init_kseg(unsigned long data)
{
	/*We are called from vcpu0 in case of generic bootloader.*/
	struct cpu *this_cpu = get_cpu_info();
	unsigned int status = read_32bit_cp0_register(CP0_STATUS);
	kseg0_prep();
	write_32bit_cp0_register(CP0_STATUS, (status & ~ST0_KX));
	this_cpu->cpu_init(this_cpu->cpu_data);
	if(dwnld_addr)
		xlr_lib_release_memory(dwnld_addr, dwnld_size);
	return;
}

static void userapp_run(int argc, char *argv[])
{
	int ch, res, i=0;

	if (argc > 32) {
		printf("Too Many Args! Passing only first 32.\n");
		argc = 32;
	}

	lib_argc = argc;
	for(i=0; i<argc; i++){
		lib_argv[i] = argv[i];
	}

	phnx_initialize_getopt();
	while ((ch = getopt(argc, argv, "A:hT:a:")) != -1) {
		switch(ch) {
		case 'A':
			override_app_args((unsigned char *)optarg);
			break;
		case 'h':
			printf("%s", userapp_usage);
			free_lib_args();
			return;
		case 'T':
			envs[1] = optarg;
			if (strcmp(((char *)(envs[1])),"uart1") == 0)
				envs[1] = "ttyS1";
			else
				envs[1] = "ttyS0";
			break;
		case 'a':
		default:
			break;
		}
	}
	
	res = xlr_load_userapp(0x0, lib_argc, lib_argv, envs);
	free_lib_args();
	if(res)
		return;

	res = xlr_launch_userapp(0x0);	
	if(res < 0)
		printf("\nLaunching KSEG App Failed.\n");
	return;

	/* Should never come here */
	printf("FATAL: Application incompatible with current bootloader revision.\n");
	printf("Please upgrade bootloader in order to use this application\n");
	shutdown();
}


static void override_app_args(unsigned char *argstr)
{
	int i=0;
	unsigned char *args = (unsigned char *)strdup((const char *)argstr);
	int no_of_args = 0;

	lib_argc = 1;
	lib_str_args = (char *)args;
	
	lib_argv[0] = "userapp";
	no_of_args++;
	
	while(*args && *args == ' ')
		args++;

	if(*args){
		no_of_args++;
		lib_argv[1] = (char *)args;
	}

	while(*args){
		if(*args == ' '){
			if((*(args+1) && *(args+1) != ' ')){
				lib_argv[no_of_args] = (char *)args + 1;
				no_of_args++;				
			}
			*args = '\0';
		}
		if(no_of_args > 31){
			args++;
			while(*args && *args != ' '){
				args++;
			}
			if(*args)
				*args = '\0';
			break;
		}
		args++;
	}
	
	for(i=no_of_args; i<32; i++)
		lib_argv[i] = NULL;

	lib_argc = no_of_args;
}

static void free_lib_args()
{
	int i=0;
	if(lib_str_args){
		free(lib_str_args);
		lib_str_args = NULL;
	}

	lib_argc = 0;
	for(i=0; i<MAX_ARGS; i++)
		lib_argv[i] = NULL;

}

#define ADDRESS_OF(x) ((unsigned long)(char *)(x))

/* Split the FDT-passed online map value
 * into a master CPU mask and a buddy CPU mask.
 * Called only for command 'smpload'.
 */
static inline uint32_t bitpos(uint32_t cpumask,
		uint32_t *buddymask) 
{
	uint32_t shifting_ones;
	uint32_t retmask = 0x1;

	if (!cpumask) {
		printf("ERROR! No CPU Mask, Unable to construct Master/Buddy masks!\n");
		return -1;
	}
	
	for(shifting_ones = 0x1; cpumask; shifting_ones <<= 1) {
		if (cpumask & shifting_ones) {
			break;
		}
		else {
			/* this bit is a 0 */	
			retmask <<= 1;
		}
	}
	
	*buddymask = cpumask & ~retmask;

	return retmask;
}

void userapp_os_run(int argc, char *argv[])
{
	uint32_t kuseg_master_cpumask, buddy_mask = 0;
	uint64_t loadaddr = 0, max_phys = MAX_PHYS;
	struct bootfile_ *this_file = get_bootfile();
	char *user_app_addr = 0;
	int ch=0, i=0, flag=0;
	
	/* Always 1 
	 */
	int override = 1;

	/* fdt_process extracts fw_arg1 value prior
	 */
	if (fw_arg1) {
		loadaddr = fw_arg1;
		flag = 1;
	}

	if (strcmp(argv[0],"smpload") == 0) {

		/* By default, set 'kuseg_master_cpumask'
		 * to the first '1' bit in cpu_online_map
		 */
		kuseg_master_cpumask = bitpos(cpu_online_map, &buddy_mask);
		if (kuseg_master_cpumask <= 0) {
			return;
		}
		smp_mode = 1;
	}
	else {
		/* ampload 
		 */
		kuseg_master_cpumask = cpu_online_map;
		buddy_mask = 0;
	}
		
	printf("Using MasterCpu mask=0x%x, BuddyCPU mask=0x%x\n", 
			kuseg_master_cpumask, buddy_mask);


	phnx_initialize_getopt();

	if (argc > 32) {
		printf("Too many arguments, Passing only first 32.\n");
		argc = 32;
	}

	lib_argc = argc;
	for (i=0; i<argc; i++) {
		lib_argv[i] = argv[i];
	}

	while ((ch = getopt(argc, argv, "honvA:M:m:a:b:c:e:T:")) != -1) {

		switch (ch) {
			case 'A':
				override_app_args((unsigned char *)optarg);
				break;
			case 'a':
				loadaddr = strtoull(optarg, 0, 16);
				flag = 1;
				break;
			case 'e':
				flag = 1;
				user_app_addr = strdup(optarg);
				break;
			case 'v':
				verbose = 1;
				break;
			case 'M':
				max_phys = strtoull(optarg, 0, 16);
				break;
			case 'T':
				envs[1] = optarg;
				if (strcmp(((char *)(envs[1])),"uart1") == 0)
					envs[1] = "ttyS1";
				else
					envs[1] = "ttyS0";
				break;
			case 'h':
				printf("%s", userapp_os_usage);
				free_lib_args();
				return;
			default:
				break;
		}
	}

	printf("Loadaddr = %llx, flags = %d\n", (unsigned long long)loadaddr, flag);

	if (!flag && !this_file->valid) {
		printf("\nNo image/address found to load!\n");
		free_lib_args();
		return;
	}

	if (max_phys < (512<<20)) {
		max_phys = MAX_PHYS;
		printf("\nAddress must be above 512MB\n");
		printf("Taking default max_phys - %#llx\n",
		       (unsigned long long)max_phys);
	}

	if (!override) {
		if((kuseg_master_cpumask&0xf) || (buddy_mask & 0xf)){
			printf("\nUser Override option \"-o\" to use vcpu0-3\n");
			free_lib_args();
			return;
		}
	}

	/* Call Library To load and launch the application
   	 * All the checks on availability of vcpus will be 
	 * done by library.
	 */
	xlr_load_kuseg_app(this_file,kuseg_master_cpumask, loadaddr, user_app_addr, 
			   buddy_mask, override, lib_argc, lib_argv, envs, verbose,max_phys);
	
	free_lib_args();
	
	this_file->busy = 0;
	this_file->loaded = 1;
	file_read = 0;
	verbose = 0;
	envs[1] = "ttyS0";
}


static void userapp_mask_cpus(int argc, char *argv[]) {
	
	uint32_t mask = 0;

	if (argc != 2) {
		printf("bad args, usage: userapp_mask_cpus <cpu_bitmap>\n");
		return;
	}
	if (!argv[1])
		return;

	if (sscanf(argv[1], "%x", &mask) != 1) {
		printf("Unable to parse the mask, %s\n", argv[1]);
		return;
	}

	boot1_info->userapp_cpu_map = userapp_cpu_map = mask;
	xlr_lib_set_userapp_mask(mask);
}

static void userapp_noreset(int argc, char *argv[]) 
{
	
	if (argc != 1) {
		printf("bad args, usage: userapp_noreset\n");
		return;
	}
	if (!argv[0])
		return;

	noreset = 1;
}

static void enable_debug(int argc, char *argv[]) {

	if (argc == 2 && argv[1]) {
		global_debug = atoi(argv[1]);
	} else {
		global_debug = 1;
	}
	printf("global_debug = %d\n", global_debug);

#ifdef NETOS_LOADER_ENHANCE
	if(global_debug ==1) {
		enable_pbuf_dbg();
	}
	else {
		disable_pbuf_dbg();
	}		
#endif /* NETOS_LOADER_ENHANCE */
	return;
}

#include "md5.h"
static struct MD5Context md5c;
void md5sum(unsigned long addr, uint32_t count, unsigned char *signature) {
	
	MD5Init(&md5c);
	MD5Update(&md5c, addr, count);
	MD5Final(signature, &md5c);
}

static void md5sum_phys(uint64_t addr, uint32_t count, unsigned char *signature) 
{
	uint32_t len = 0;
	uint16_t buf_len;
	char buffer[512];
	int i;

	MD5Init(&md5c);

	while (len < count) {
		buf_len =  (count -len) < 512 ? count - len : 512;

		for (i = 0; i < buf_len; i++) {
			buffer[i] = lb_40bit_phys_cached(addr + len + i);
		}

		MD5Update(&md5c, buffer, buf_len);
		len += buf_len;
	}

	MD5Final(signature, &md5c);
}

void alloc_dwnld_buffer(uint64_t size)
{
	struct bootfile_ *this_file = get_bootfile();

	if (dwnld_addr == 0x0D000000ULL) {
		printf("Can't allocate requested size buffer\n");
		printf("[Warning]: Using Fixed [0x0D000000] Download Address.\n");
		printf("[Warning]: Size of buffer - 32MB\n");
		dwnld_size = (32<<20);
		return;
	}

	if(dwnld_addr)
		xlr_free_memory(dwnld_addr, dwnld_size);
	
	dwnld_addr = xlr_alloc_memory(&size);
	if(!dwnld_addr){
		dwnld_addr = xlr_alloc_memory(&dwnld_size);
		if(!dwnld_addr){
			printf("No buffer available!!!\n");
			dwnld_addr = 0;
			dwnld_size = 0;
		}
		printf("Allocated buffer of size %dMB\n",(int)(dwnld_size>>20));
	} else {
		dwnld_size = size;
		printf("Allocated %dMB download buffer @ 0x%lx\n",
				(int)(dwnld_size>>20), dwnld_addr);
	}

	this_file->valid 	= 0;
	this_file->max_size = dwnld_size;
	this_file->start 	= dwnld_addr;
}

static void show_max_load_usage(void)
{
	printf("usage: max_load_size [-s <size (in hex)>] [-d] [-h]\n");
	printf("       -s <size> - Allocate size buffer for downloading an image\n"
	       "                   only 32MB and 128MB size options are valid\n");
	printf("       -d        - Display current buffer size@addr\n");
	printf("       -h        - Show help\n");
	return;
}

static void set_max_load_size(int argc, char *argv[])
{
	int ch;
	uint64_t size;

	if(argc < 2){
		show_max_load_usage();
		return;
	}
	phnx_initialize_getopt();
	while ((ch = getopt(argc, argv, "s:dh")) != -1) {
		switch(ch){
		case 's':
			size = strtoull(optarg, 0, 16);
			if(!((size == (32*1024*1024))  ||
			     (size == (128*1024*1024)))){
				printf("Invalid args\n");
				show_max_load_usage();
				return;
			}
			alloc_dwnld_buffer(size);
			break;
		case 'd':
			if(!dwnld_addr){
				printf("No buffer allocated yet\n");
			}else{
				printf("Current download buffer %#llx @"
				       "%#llx\n",
				       (unsigned long long)dwnld_size,
				       (unsigned long long)dwnld_addr);
			}
			break;
		case 'h':
		default:
			show_max_load_usage();
			break;
		}
	}
}

static void checksum(int argc, char *argv[]) {

	uint64_t addr = 0;
	uint32_t count = 0;
	unsigned char signature[16];
	int i = 0;
	int phys_addr = 1;
	int argv_start = 1;

	/* User requested physical address? */
	if (argc > 2 && strcmp(argv[1], "-v") == 0) {
		argv_start++;
		phys_addr = 0;
	} 

	/* Only two parameters are needed */
	if (argc - argv_start != 2) {
		printf("bad args, usage: checksum [-v] <address> <count>\n");
		return;
	}

	addr = strtoull(argv[argv_start], 0, 16);
	count = strtoul(argv[argv_start + 1], 0, 10);

	if (!phys_addr) {
		if (!(addr >= (unsigned long)KSEG0 && 
		      addr < (unsigned long)KSEG0 + (512 << 20)) &&
		    !(addr >= (unsigned long)KSEG1 && addr < (unsigned long)KSEG1 
		      + (512 << 20))) {
			printf("[%s]: checksum address = %llx, arg = %s is not "
			       "in kseg0 or kseg1\n",
			       __FUNCTION__, (unsigned long long) addr, argv[argv_start]);
			return;
		}

		if (count > (512 << 20)) {
			printf("bad count (%u), should be < 512MB\n", count);
			return;
		}
	} 

	printf("Calculating MD5 checksum of addr = %llx, count = %u \n", 
	       (unsigned long long) addr, count);
	if (!phys_addr)
		md5sum((unsigned long)addr, count, signature);
	else
		md5sum_phys(addr, count, signature);

	printf("MD5 Sum: ");
	fflush(stdout);
	for (i = 0; i < sizeof(signature); i++)
		printf("%02x", signature[i]);
	printf("\n");
}

static void ls_shell_cmd(int argc, char *argv[]) {
	
	struct bootfile_ *this_file = get_bootfile();
	if(this_file == NULL)
		return;
	if (this_file->valid) {
		printf("%s: start = 0x%010llx (physical), size = %u (max_size=%u)\n",
		       this_file->name, this_file->start, this_file->size,
		       this_file->max_size);
	}
}

static void load_and_run(int argc, char *argv[]) {

	if (!elfload_shell_cmd(argc, argv)) {
		userapp_run(argc, argv);
	} else {
		printf("FATAL(run): Couldn't elfload the image\n");
	}
}

int8_t userapp_init(void) {
	
	boot1_info = (struct boot1_info *) &hack_boot1_info;

#ifdef XLP_SIM
	nlm_cpu_frequency = nlm_hal_cpu_freq();
#else
	nlm_cpu_frequency = 1000000000; /* TODO */
#endif
	/* remove the cpu part after reading the ebase */
	common_ebase = (long)(int)read_c0_ebase();
	common_ebase = common_ebase & (~(( 1 << 12) - 1));
	
	args[0] = "ELF Kernel";
	args[1] = "";
	nargs = 2;

	envs[0] = "BOARD=ARIZONA";
	envs[1] = "ttyS0";
	envs[2] = NULL;

	register_shell_cmd("userapp", 
			userapp_run, "launch kseg0 application on CPU0");

	register_shell_cmd("ampload", 
			userapp_os_run, "launch kuseg application(s) in AMP mode");
	register_shell_cmd("smpload", 
			userapp_os_run, "launch kuseg application(s) in SMP mode");

	register_shell_cmd("run", 
			load_and_run, "load elf file, call command 'userapp'");
	register_shell_cmd("userapp_mask_cpus", 
			userapp_mask_cpus, "set active-cpu bitmask");
	register_shell_cmd("userapp_noreset", 
			userapp_noreset, "do not reset msgring0/gmac0 for 'userapp'");
	register_shell_cmd("enable_debug", 
			enable_debug, "currently enables GMAC debug");
	register_shell_cmd("checksum", 
			checksum, "MD5 checksum calculator");
	register_shell_cmd("ls", 
			ls_shell_cmd, "display 'bootfile' info");
	register_shell_cmd("max_load_size", 
			set_max_load_size, "sets max image download size");

	userapp_cpu_map = cpu_online_map;

	boot1_info->io_base 			= PTR2U64(cpu_io_base);
	boot1_info->cpu_online_map 		= cpu_online_map;
	boot1_info->psb_os_cpu_map 		= cpus_in_kuseg_mode;
	boot1_info->userapp_cpu_map 	= userapp_cpu_map;

	elf_initialize();

	return 0;
}

void __trunctfdf2(void)
{
	abort();
}

void __extenddftf2(void)
{
	return;
} 
