/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/

#include "asm.h"
#include "stackframe.h"
#include "mipsregs.h"

#include "cpu.h"

#define get_pcpu_info(reg) \
	mfc0     k0, CP0_PRID, 1; 	\
	srl      k0, k0, 12;	\
	sll	 k0, k0, 12;	\
	daddiu   k0, k0, NETOS_EBASE_PCPU_OFF; \
	move	reg, k0;

#define netos_exc_entry(reg)	\
	get_pcpu_info(k0);	\
	ld	k1, NETOS_PCPU_EXCCNT_OFF(k0);	\
	daddiu	k1, k1, 1;	\
	sd	k1, NETOS_PCPU_EXCCNT_OFF(k0);	\
	daddiu	k1, k1, -1;	\
	beqz	k1, 1f;	\
	nop;	\
	move	reg, sp;		\
	b	2f;			\
	nop;				\
1:	ld	k1, NETOS_PCPU_SP_OFF(k0);	\
	move	reg, k1;	\
2:	daddiu  reg, reg, -K_STACK_SIZE;
	

#define netos_exc_exit(reg) \
	get_pcpu_info(k0);	\
	ld	k1, NETOS_PCPU_EXCCNT_OFF(k0);	\
	daddiu	k1, k1, -1;	\
	sd	k1, NETOS_PCPU_EXCCNT_OFF(k0);	\
	move	reg, sp;

#define save_exc_stack_frame        \
	.set push;                 \
	.set reorder;              \
	.set noat;                 \
	.set mips64;			\
	netos_exc_entry(k0);	\
	sd	$0, STACK_OFF_R0(k0);	  \
	sd	$1, STACK_OFF_R1(k0);	  \
	sd	$2, STACK_OFF_R2(k0);	  \
        sd	$3, STACK_OFF_R3(k0);	  \
	sd	$4, STACK_OFF_R4(k0);	  \
	sd	$5, STACK_OFF_R5(k0);	  \
	sd	$6, STACK_OFF_R6(k0);	  \
	sd	$7, STACK_OFF_R7(k0);	  \
	sd	$8, STACK_OFF_R8(k0);	  \
	sd	$9, STACK_OFF_R9(k0);	  \
	sd	$10,STACK_OFF_R10(k0);	  \
	sd	$11, STACK_OFF_R11(k0);   \
	sd	$12, STACK_OFF_R12(k0);   \
	sd	$13, STACK_OFF_R13(k0);   \
	sd	$14, STACK_OFF_R14(k0);   \
	sd	$15, STACK_OFF_R15(k0);   \
	sd	$16, STACK_OFF_R16(k0);   \
	sd	$17, STACK_OFF_R17(k0);   \
	sd	$18, STACK_OFF_R18(k0);   \
	sd	$19, STACK_OFF_R19(k0);   \
	sd	$20, STACK_OFF_R20(k0);   \
	sd	$21, STACK_OFF_R21(k0);   \
	sd	$22, STACK_OFF_R22(k0);   \
	sd	$23, STACK_OFF_R23(k0);   \
	sd	$24, STACK_OFF_R24(k0);   \
	sd	$25, STACK_OFF_R25(k0);   \
	sd	$28, STACK_OFF_R28(k0);   \
	sd	$29, STACK_OFF_R29(k0);	  \
	sd	$30, STACK_OFF_R30(k0);   \
	sd	$31, STACK_OFF_R31(k0);   \
	dmfc0	k1, CP0_BADVADDR;         \
	sd	k1, STACK_OFF_BVADDR(k0); \
	dmfc0	k1, CP0_STATUS;         \
	sd	k1, STACK_OFF_STATUS(k0); \
	dmfc0	k1, CP0_CAUSE;          \
	sd	k1, STACK_OFF_CAUSE(k0);  \
	dmfc0	k1, CP0_EPC;            \
	sd	k1, STACK_OFF_EPC(k0);	  \
	mflo	k1;		          \
	sd	k1,  STACK_OFF_LO(k0);	  \
	mfhi	k1;		          \
	sd	k1,  STACK_OFF_HI(k0);	  \
	move	sp, k0;			\
	CLI;	\
	.set pop;			   

#define restore_exc_stack_frame                  \
	.set	push;		             \
	.set    noat;                        \
	.set	reorder;	             \
	.set mips64;			\
	dmfc0	t0, CP0_STATUS;            \
	ori	t0, 0x1f;	             \
	xori	t0, 0x1f;	             \
	dmtc0	t0, CP0_STATUS;            \
	netos_exc_exit(k0);		\
	ld	$1,  STACK_OFF_R1(k0);       \
	ld	$4,  STACK_OFF_R4(k0);	     \
	ld	$5,  STACK_OFF_R5(k0);	     \
	ld	$6,  STACK_OFF_R6(k0);	     \
	ld	$7,  STACK_OFF_R7(k0);	     \
	ld	$8, STACK_OFF_R8(k0);	     \
	ld	$9, STACK_OFF_R9(k0);	     \
	ld	$10,STACK_OFF_R10(k0);	     \
	ld	$11, STACK_OFF_R11(k0);      \
	ld	$12, STACK_OFF_R12(k0);      \
	ld	$13, STACK_OFF_R13(k0);      \
	ld	$14, STACK_OFF_R14(k0);      \
	ld	$15, STACK_OFF_R15(k0);      \
	ld	$16, STACK_OFF_R16(k0);      \
	ld	$17, STACK_OFF_R17(k0);      \
	ld	$18, STACK_OFF_R18(k0);      \
	ld	$19, STACK_OFF_R19(k0);      \
	ld	$20, STACK_OFF_R20(k0);      \
	ld	$21, STACK_OFF_R21(k0);      \
	ld	$22, STACK_OFF_R22(k0);      \
	ld	$23, STACK_OFF_R23(k0);      \
	ld	$24, STACK_OFF_R24(k0);   \
	ld	$25, STACK_OFF_R25(k0);   \
	ld	$28, STACK_OFF_R28(k0);   \
	ld	$29, STACK_OFF_R29(k0);	  \
	ld	$30, STACK_OFF_R30(k0);   \
	ld	$31, STACK_OFF_R31(k0);   \
	ld	k1, STACK_OFF_EPC(k0);	     \
	dmtc0	k1, CP0_EPC;	             \
	ld	k1, STACK_OFF_LO(k0);	     \
	mtlo	k1;		             \
	ld	k1, STACK_OFF_HI(k0);	     \
	mthi	k1;		             \
	dmfc0	v0, CP0_STATUS;			\
	li	v1, 0xff00;	             \
	and	v0, v1;		             \
	ld	k1, STACK_OFF_STATUS(k0);    \
	nor	v1, $0, v1;	             \
	and	k1, v1;		             \
	or	k1, v0;		             \
	ld	$2,  STACK_OFF_R2(k0);	     \
	ld	$3,  STACK_OFF_R3(k0);	     \
	dmtc0	k1, CP0_STATUS;            \
	.set pop;			   

	.text
	.align 4
	.set push
	.set reorder

EXPORT(ldr_ret_from_exception)
	.set	noat
	restore_exc_stack_frame
	eret
	.set	at

NESTED(ldr_except_vec_tlbrefill, 0, sp)
	get_pcpu_info(k0)
	daddiu k0, k0, NETOS_PCPU_XTLB_REFILL_OFF
	jr k0
	nop
END(ldr_except_vec_tlbrefill)

NESTED(ldr_except_vec_xtlbrefill, 0, sp)
	get_pcpu_info(k0)
	daddiu k0, k0, NETOS_PCPU_XTLB_REFILL_OFF
	jr k0
	nop
END(ldr_except_vec_xtlbrefill)

NESTED(ldr_except_vec_cacheerr, 0, sp)
	CLI
	j	ldr_cache_error_panic
	nop
END(ldr_except_vec_cacheerr)

NESTED(ldr_except_vec_genex, 0, sp)
	get_pcpu_info(k0)
	daddiu k0, k0, NETOS_PCPU_EXC_DISP_HDLR_OFF
	jr k0
	nop
END(ldr_except_vec_genex)

NESTED(ldr_except_vec_vecint, 0, sp)
	CLI
	j ldr_vecint_panic
	nop
END(ldr_except_vec_vecint)

/* Dispatch handler for all 32 exceptions, called from ldr_except_vec_genex */
/* it should not return here, as it is getting executed from copied mem */
/* should not use k0 and k1, refill exception can come */
	.align 8
NESTED(ldr_gen_exc_dispatch_handler, K_STACK_SIZE, sp)
ldr_gen_exc_dispatch_handler:
	.set noat
	save_exc_stack_frame
	.set at
	move a0, sp
	mfc0    t0, CP0_CAUSE
	andi    t0, t0, 0x7c
	sll     t0, t0, 1
	dla	t1, ldr_exception_handlers
	daddu	t1, t1, t0
	ld      t0, 0(t1)
	dla	ra, ldr_ret_from_exception
	jr      t0 
	nop

	.global ldr_gen_exc_dispatch_handler_exit
ldr_gen_exc_dispatch_handler_exit:	
END(ldr_gen_exc_dispatch_handler)	

	.align	8
NESTED(ldr_irq_handler, K_STACK_SIZE, sp)

	/* Read EIRR :	 */
	.word   0x40304806        /* dmfc0 s0, eirr */

	/* Read EIMR :   */
	.word   0x40314807        /* dmfc0 s1, eimr */

	and s0, s0, s1
		
	/* If no interrupts, return */
	beqz    s0, 5f
	nop

1:
	/* retrieve the highest priority interrupt */
	.word       0x72118824        // dclz s1 s0
	dsubu       a0, zero, s1
	daddiu      a0, a0, 63        // a0 now has irq#
	move        a1, sp            // a1 now has sp

	li          s0, 1
	dsllv       s0, s0, a0
	.word       0x40b04806        /* dmtc0 s0, eirr */  /* blanking out that bit ? */

	jal         do_phnx_irq
	nop
5:  	j           ldr_ret_from_exception
	nop

	.set pop
END(ldr_irq_handler)

#ifndef XLP_SIM
NESTED(xlr_boot_smp_nmi, 0, sp)
	.set push
	.set noat
	.set mips64
	.set noreorder

	/* save k0 and k1 */
	mtc0	k0, $22, 6
	mtc0	k1, $22, 7

	/* Clear the  NMI and BEV bits */
	mfc0    k0, CP0_STATUS
	li      k1, 0xffb7ffff
	and     k0, k0, k1
	/* set KX flag here */
#ifdef CONFIG_64BIT
	ori	k0, ST0_KX
#endif
	mtc0    k0, CP0_STATUS

	/* setup the stack and gp before jumping to start function.
	   TLB Setting - TODO
	   */
	mfc0	t0, $15, 1 /* Read ebase */
	andi	t0, 0x1f
	dla 	t1, xlr_stack_pages_temp
	li   t2, PER_CPU_STACK_SIZE
	mul  t3, t2, t0
	nop
	nop
	nop
	nop
	nop
	nop
	daddu	gp, t1, t3
	daddu	sp, gp, t2
	daddi	sp, sp, -64

	dla	k1, xlr_secondary_entry_point
	ld	k0, 0(k1)
	jr       k0
	nop
	.set pop
END(xlr_boot_smp_nmi)
#endif
