#include <offset.h>
#include <asm.h> 
#include <mipsregs.h>
#include <stackframe.h>
#include "tlb.h"
#include "cpu.h"

/* netos ends at 1 GB, 0x40000000 */
#define NETOS_EADDR_SHIFT 30

/* For page size 64K */
#define NETOS_PAGE_MASK 0x1e000
#define NETOS_PTE_SHIFT 16

/* To get pte page from pmd 16 + 13 (64kpage / 8 per entry = 8k = 13bit) */
#define NETOS_PMDIR_SHIFT 29

/* To get pmd page from pgd 29 + 13 = 42 */
#define NETOS_PGDIR_SHIFT 42

/* Software data kept in the pte entry - width */
#define NETOS_PTE_SDATA_SHIFT 6

#define LONG_SZ 8
/* each entry contails 8 bytes */
#define NETOS_PGENTRY_ORDER  3 

/* each entry contains 16 bytes */
#define HPEXEC_PGENTRY_ORDER 4

#define HPEXEC_PAGEMASK_OFF 34

/* pgd for netos and hpexec are stored in scratch regs */
#define GET_NETOS_PGD(reg) \
	dmfc0 reg, CP0_OSSCRATCH, NETOS_PTEBASE_SCRATCH_IDX

#define GET_HPEXEC_PGD(reg) \
	dmfc0 reg, CP0_OSSCRATCH, HPEXEC_PTEBASE_SCRATCH_IDX

/* pgd ptr is in reg */
#define GET_NETOS_PMD_PAGE(reg, tmp) \
	dmfc0    tmp, CP0_BADVADDR; \
	dsrl     tmp, tmp, NETOS_PGDIR_SHIFT; \
	dsll	 tmp, tmp, NETOS_PGENTRY_ORDER; \
	daddu	 tmp, reg, tmp; \
	ld	 reg, 0(tmp)

/* pmd ptr is in reg */
#define GET_NETOS_PTE_PAGE(reg, tmp) \
	dmfc0    tmp, CP0_BADVADDR; \
	dsll	 tmp, tmp, 64 - NETOS_PGDIR_SHIFT; \
	dsrl	 tmp, tmp, 64 - NETOS_PGDIR_SHIFT; \
	dsrl	 tmp, tmp, NETOS_PMDIR_SHIFT; \
	dsll	 tmp, tmp, NETOS_PGENTRY_ORDER;	\
	daddu	 tmp, reg, tmp;	\
	ld	 reg, 0(tmp)

/* pte ptr is in reg */
#define GET_NETOS_PTE(reg, tmp) \
	dmfc0    tmp, CP0_BADVADDR; \
	dsll	 tmp, tmp, 64 - NETOS_PMDIR_SHIFT; \
	dsrl	 tmp, tmp, 64 - NETOS_PMDIR_SHIFT; \
	dsrl	 tmp, tmp, NETOS_PTE_SHIFT; \
	dsrl	 tmp, tmp, 1; \
	dsll	 tmp, tmp, 1; \
	dsll	 tmp, tmp, NETOS_PGENTRY_ORDER; \
	daddu	 tmp, reg, tmp; \
	ld	 reg, 0(tmp); \
	dsrl	 reg, reg, NETOS_PTE_SDATA_SHIFT; \
	dmtc0	 reg, CP0_ENTRYLO0; \
	ld	 reg, LONG_SZ(tmp); \
	dsrl	 reg, reg, NETOS_PTE_SDATA_SHIFT; \
	dmtc0	 reg, CP0_ENTRYLO1; \
	li	 reg, NETOS_PAGE_MASK; \
	mtc0  	 reg, CP0_PAGEMASK; \
	tlbwr; \
	eret

/* pte ptr is in reg  */
/* entry_lo0 is stored as: 0:33->entrylo0, 63:34->page_mask */
/* (higher 2 bits of pagemask are always 0) */

#define GET_HPEXEC_PTE(reg, tmp) \
	dmfc0 tmp, CP0_BADVADDR; 	\
	dsrl  tmp, tmp, MIN_PAGE_BITS + 1;	\
	dsll  tmp, tmp, HPEXEC_PGENTRY_ORDER;	\
	daddu tmp, reg, tmp;	\
	ld    reg, 0(tmp);	\
	dmtc0 reg, CP0_ENTRYLO0;	\
	ld    reg, LONG_SZ(tmp);	\
	dmtc0 reg, CP0_ENTRYLO1;	\
	ld    reg, 0(tmp);	\
	dsrl  reg, reg, HPEXEC_PAGEMASK_OFF;	\
	mtc0  reg, CP0_PAGEMASK;	\
	tlbwr;	\
	eret 

	.text
	.align 8

LEAF(netos_tlb_refill_handler) 

	.set push
	.set noreorder
	.set noat
	.set mips64
	wait
	.set pop

END(netos_tlb_refill_handler)

LEAF(netos_xtlb_refill_handler)
netos_xtlb_refill_handler:

	.set push
	.set noreorder
	.set noat
	dmfc0 k0, CP0_BADVADDR
	dsrl  k0, NETOS_EADDR_SHIFT
	beqz  k0, 1f /* for netos */
	nop
	dmfc0 k0, CP0_BADVADDR 
	dsrl  k0, MAX_VIRT_BITS 
	bgtz  k0, 2f         
	nop
	GET_HPEXEC_PGD(k0)
	GET_HPEXEC_PTE(k0, k1)
1:
	GET_NETOS_PGD(k0)
	GET_NETOS_PMD_PAGE(k0, k1)
	GET_NETOS_PTE_PAGE(k0, k1)
	GET_NETOS_PTE(k0, k1)
2:
	dmtc0 $0, CP0_ENTRYLO0
	dmtc0 $0, CP0_ENTRYLO1
	mtc0  $0, CP0_PAGEMASK
	tlbwr
	eret 


	.set pop
	.global netos_xtlb_refill_handler_exit
netos_xtlb_refill_handler_exit:
END(netos_xtlb_refill_handler)

