#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "types.h"
#include "stackframe.h"
#include "mipsregs.h"
#include "mips-exts.h"
#include "cache.h"
#include "xlr_boot_lib.h"
#include "mm.h"
#include "cpu.h"
#include "irqparam.h"
#include "irq.h"
#include "nlm_hal_macros.h"

#ifdef CONFIG_XEN
#include <asm/xen/hypercall.h>
#endif

unsigned long ldr_exception_handlers[32];

extern char ldr_except_vec_tlbrefill[], ldr_except_vec_xtlbrefill[];
extern char ldr_except_vec_genex[], ldr_except_vec_cacheerr[];
extern char ldr_except_vec_vecint[];
extern int xlr_wakeup_ipi[32];
extern char netos_tlb_refill_handler[];
extern char netos_xtlb_refill_handler[], netos_xtlb_refill_handler_exit[];
extern char ldr_gen_exc_dispatch_handler[], ldr_gen_exc_dispatch_handler_exit[];
extern void ldr_irq_handler(void);
extern void read_cache_config(void);
void destroy_domain(void);
static spinlock_t trap_init_lock;
extern void tlb_flush_intr_handler(int irq);

void uart_int_handler(int irq, struct pt_regs *regs)
{
	return;
}

void print_regs(struct pt_regs *regs, int cpu)
{
	/* Saved main processor registers */
	printf("\n(NetOS@%d)$0 :0x%016llx 0x%016lx 0x%016lx 0x%016lx", cpu,
	       0ULL, regs->regs[1], regs->regs[2], regs->regs[3]);
	printf("\n(NetOS@%d)$4 :0x%016lx 0x%016lx 0x%016lx 0x%016lx", cpu,
	       regs->regs[4], regs->regs[5], regs->regs[6], regs->regs[7]);
	printf("\n(NetOS@%d)$8 :0x%016lx 0x%016lx 0x%016lx 0x%016lx", cpu, regs->regs[8],
	       regs->regs[9],  regs->regs[10], regs->regs[11]);
	printf("\n(NetOS@%d)$12 :0x%016lx 0x%016lx 0x%016lx 0x%016lx", cpu,
	       regs->regs[12], regs->regs[13], regs->regs[14], regs->regs[15]);
	printf("\n(NetOS@%d)$16 :0x%016lx 0x%016lx 0x%016lx 0x%016lx", cpu,
	       regs->regs[16], regs->regs[17], regs->regs[18], regs->regs[19]);
	printf("\n(NetOS@%d)$20 :0x%016lx 0x%016lx 0x%016lx 0x%016lx", cpu,
	       regs->regs[20], regs->regs[21], regs->regs[22], regs->regs[23]);
	printf("\n(NetOS@%d)$24 :0x%016lx 0x%016lx 0x%016lx 0x%016lx", cpu,
	       regs->regs[24], regs->regs[25], regs->regs[26], regs->regs[27]);
	printf("\n(NetOS@%d)$28 :0x%016lx 0x%016lx 0x%016lx 0x%016lx\n", cpu,
	       regs->regs[28], regs->regs[29], regs->regs[30], regs->regs[31]);
	/* Saved cp0 registers */
	printf("(NetOS@%d)Hi : 0x%016lx\n", cpu, regs->hi);
	printf("(NetOS@%d)Lo : 0x%016lx\n", cpu, regs->lo);
	printf("(NetOS@%d)Badva : 0x%016lx\n", cpu, regs->cp0_badvaddr);
	printf("(NetOS@%d)epc   : 0x%016lx\n", cpu, regs->cp0_epc);
	printf("(NetOS@%d)Status: 0x%016lx\n", cpu, regs->cp0_status);
	printf("(NetOS@%d)Cause : 0x%016lx\n", cpu, regs->cp0_cause);
}

void ldr_do_reserved(struct pt_regs *regs)
{
	extern void dump_all_valid_tlb_entries(void);
	int cpu = processor_id();

	print_regs(regs, cpu);

	printf("(NetOS) Caught exception %ld - on cpu %d should not happen\n",
	       (regs->cp0_cause >> 2) & 0x1f, cpu);

	dump_all_valid_tlb_entries();

	enable_destroy_ipi();

	cpu_halt();
}

void *set_except_vector(int n, void *addr)
{
	ldr_exception_handlers[n] = (unsigned long)addr;

	return addr;
}

irqreturn_t netos_irq_handler(int irq, struct pt_regs *regs)
{
	switch (irq) 
	{
	case IRQ_UART:
		printf("Received UART Interrupt= %d\n", irq);
		uart_int_handler(irq, regs);
		break;
	case IRQ_WAKEUP_CPU_IPI:
	case IRQ_STOP_CPU_IPI:
	case IRQ_RECLAIM_CPU_RSRC_IPI:
	case IRQ_JOIN_CPU_IPI:
		xlr_lib_intr_handler(irq);
		break;
#ifdef CONFIG_XEN
	case IRQ_DESTROY_DOMAIN_IPI:
		printf("Received Destroy Domain Interrupt on cpu %d.\n", processor_id());
		regs->cp0_epc = (long)destroy_domain;
		break;
#endif
	case IRQ_TLB_FLUSH_IPI:
		tlb_flush_intr_handler(irq);
		break;
	case IRQ_MAC:
		printf("Received MAC Interrupt = %d\n", irq);
	default:
		printf(" Received Unhandled Interrupt = %d!\n", irq);
		break;
	}
	return 0;
}

struct exception_region *exception_vectors = (struct exception_region *)KSEG0;

void ldr_trap_init(void)
{
	int i = 0;
	int cpu = processor_id();
	unsigned long ebase = (unsigned long)&exception_vectors[cpu], pcpuptr;
	unsigned long long netos_sp;
	unsigned int size;
#ifdef CONFIG_XEN
	unsigned int pgw;
#endif
	static int cpu_cnt = 1;

	/* Stack grows upper to down, cpu_cnt is one because of this */
	spin_lock(&trap_init_lock);
	netos_sp = excsp + (NETOS_EXC_STACK_SZ * cpu_cnt) - 64;
	cpu_cnt++;
	spin_unlock(&trap_init_lock);
	//printf("[cpu %d] netos sp = %lx\n", cpu, (long)netos_sp);

	pcpuptr = ebase + NETOS_EBASE_PCPU_OFF;
	*(unsigned long long *)(pcpuptr + NETOS_PCPU_SP_OFF) = netos_sp;
	*(unsigned long long *)(pcpuptr + NETOS_PCPU_EXCCNT_OFF) = 0;

	/* Copy the refill and exception dispatch codes */
	size =  netos_xtlb_refill_handler_exit - netos_xtlb_refill_handler;
	if(size > NETOS_XTLB_REFILL_HDLR_SZ) {
		printf("Refill handler doesnot fit in, size %d \n", size);
		cpu_halt();
	}
	memcpy((char *)(pcpuptr + NETOS_PCPU_XTLB_REFILL_OFF), netos_xtlb_refill_handler, size);

	size = ldr_gen_exc_dispatch_handler_exit - ldr_gen_exc_dispatch_handler;
	if(size > NETOS_GEN_EXC_DISP_HDLR_SZ) {
		printf("Refill handler doesnot fit in, size %d \n", size);
		cpu_halt();
	}
	memcpy((char *)(pcpuptr + NETOS_PCPU_EXC_DISP_HDLR_OFF), ldr_gen_exc_dispatch_handler, size);


	read_cache_config();

	for (i = 0; i <= 31; i++)
		ldr_exception_handlers[i] = (unsigned long)ldr_do_reserved;

	/* Set the Interrupt - Exception Handler (Ex Code 0) */
	ldr_exception_handlers[0] = (unsigned long)ldr_irq_handler;

	/* Set the tlbm/tlbs/tlbr- Exception Handler (Ex Code 1,2,3 ) */
	ldr_exception_handlers[1] = (unsigned long)tlb_invalid;
	ldr_exception_handlers[2] = (unsigned long)tlb_invalid;
	ldr_exception_handlers[3] = (unsigned long)tlb_invalid;

	/* Set the syscall exception handler , Ex code 8 */
	ldr_exception_handlers[8] = (unsigned long)ldr_do_syscall;

	/* Copy the generic exception handler code to it's final destination. */
	memcpy((void *)(ebase), ldr_except_vec_tlbrefill, 0x80);
	memcpy((void *)(ebase + 0x80), ldr_except_vec_xtlbrefill, 0x80);
	memcpy((void *)(ebase + 0x100), ldr_except_vec_cacheerr, 0x80);
	memcpy((void *)(ebase + 0x180), ldr_except_vec_genex, 0x80);
	memcpy((void *)(ebase + 0x200), ldr_except_vec_vecint, 0x80);

	/* set up the ebase register */
	write_32bit_cp0_register_sel(CP0_EBASE, (uint32_t) (ebase & 0x3ffff000),1);

	/* Flush I-Cache */
	ldr_flush_icache_all();

	/* Flush D-Cache 
	 * Note: Need to do this only because the 
	 * cacheerr exception vector may be still 
	 * in D-cache when cpu takes the exception
	 */
	ldr_flush_dcache_all();

#ifdef CONFIG_XEN
	/* Uses scratch reg for netos pte */
	write_64bit_cp0_register_sel(CP0_OSSCRATCH, (long)current_pgd, NETOS_PTEBASE_SCRATCH_IDX);
	/* Disable page walker and flush all tlbs */
	pgw = read_32bit_cp0_register_sel(CP0_CONFIG, 6);
	//printf("page walker config = %x\n", pgw);
	pgw &= ~( 1 << 1 | 1 << 3);
	write_32bit_cp0_register_sel(CP0_CONFIG, pgw, 6);
#endif

}

void ldr_cache_error_panic()
{
	printf ("\nUnhandled cache error!!");
	printf("\ncpu_%d: PANIC!!",processor_id());

	cpu_halt();
}

void ldr_vecint_panic()
{
	printf ("\n(NetOS) cpu-%d: Unhandled Vectored Interrupt !!", processor_id());

	cpu_halt();
}

void ldr_do_syscall(struct pt_regs *regs)
{
	extern int nlm_sys_call(int index, void *arg);
	int rv;

	rv = nlm_sys_call(regs->regs[2], (void *)(long)regs->regs[4]);
	/* return status in v0 */
	regs->regs[2] = rv;
	/* increment epc by 4*/
	regs->cp0_epc += 4;
}

int config_exception_sys_call(void *arg)
{
	struct except_param *s = arg;

	switch(s->cmd)
	{
	case NLM_SET_EXCEPT_VECTOR:
		set_except_vector(s->index, (void *)s->handler);
		break;
	default:
		s->rv = -1;
		return -1;
	}
	s->rv = 0;

	return 0;
}

#ifdef CONFIG_XEN
/*
 * A wrapper to call the destroy_domain hypercall.
 */
void destroy_domain(void)
{
	unsigned int pgw;
	/* enable the pagewalker */

	pgw = read_32bit_cp0_register_sel(CP0_CONFIG, 6);
	pgw |= ( 1 << 1 | 1 << 3);
	write_32bit_cp0_register_sel(CP0_CONFIG, pgw, 6);

	HYPERVISOR_destroy_domain();
}
#endif

void dump_all_valid_tlb_entries(void)
{
	unsigned long long entryhi, entrylo0, entrylo1;
	unsigned int pagemask;
	int i;
	int nentries = ((read_c0_config1() >> 25) & 0x3f) + 1;

	for(i = 0; i < nentries; i++) {
		__asm__ __volatile__(".set push             \n"
				     ".set noreorder        \n"
				     ".set mips64           \n"
				     ".set noat             \n"
				     "     mtc0  %4, $0     \n"
				     "     tlbr             \n"
				     "      nop         \n"
				     "     dmfc0  %0, $10   \n"
				     "     dmfc0  %1, $2    \n"
				     "     dmfc0  %2, $3    \n"
				     "     mfc0   %3, $5    \n"
				     ".set pop              \n"
				     :"=r"(entryhi),
				      "=r"(entrylo0),
				      "=r"(entrylo1),
				      "=r"(pagemask)
				     :"r"(i));

		if(entrylo0 & 0x2) {
			printf("... Tlb index %d asid %ld vpn %lx seg %ld lo0f %lx lo0pfn %lx mask %x\n", i,
			       (long)(entryhi & 0xff), (long)(entryhi & 0x3fffffffffffe000ULL), (long)(entryhi >> 62),
			       (long)(entrylo0 & 0x3f), (long)((entrylo0 >> 6) << 12), pagemask);
		}
	}
}
