#include <stdarg.h>

#include "physaddrspace.h"
#include "xlr_boot_lib.h"
#include "elf.h"
#include "cpu.h"
#include "boot2.h"
#include <string.h>
#include "system.h"
#include "memblk.h"
#include "boot_mgr.h"
#include "xlr_boot_lib.h"
#include "types.h"
#include "mipsregs.h"
#include "msgring.h"
#include "mips-exts.h"
//#include "elf32.h"
#include "libelf.h"
#include "rmidelay.h"

#include "sys/module.h"
#include "bitops.h"
#ifdef XLP_SIM
#include "nlm_hal_pic.h"
#else
#include "pic.h"
#endif

extern void xlr_park(unsigned long args);
extern uint32_t cpu_online_map;
extern void flush_tlb_all(void);

#ifdef NETOS_LOADER_TODO

static void xlr_put_non_vcpu0_in_park(void)
{
#ifdef NETOS_LOADER_TODO
	volatile int i=0;
#endif /* NETOS_LOADER_TODO */
	int max_cpus;

#if defined(PHOENIX_SIM)
    max_cpus = 32;
#else
    max_cpus = num_ones(cpu_online_map);
#endif

#ifdef NETOS_LOADER_TODO
	for(i=1;i<max_cpus;i++) {
#if defined(PHOENIX_SIM)
		if (cpu_online_map & (1 << i)) 
#endif
			wakeup_cpu_msgring((int)i,(unsigned long)xlr_lib_entry, 0);
	}
#endif /* NETOS_LOADER_TODO */
	/* delay, so others can come in park */
	rmi_udelay(10000);
}


static void xlr_lib_prelaunch_init_kuseg(unsigned long data)
{
	/*PrelaunchInit For KUSEG applications.*/
	return;
}
#endif /* NETOS_LOADER_TODO */



void stop_vcpu(int argc, char *argv[])
{
	unsigned long mask=0;
	char ch;

	phnx_initialize_getopt();
	while ((ch = getopt(argc, argv, "hm:")) != -1) {
			switch (ch) {
					case 'm':
							mask = strtoul(optarg, 0, 16);
							break;
					case 'h':
							printf("Usage: stop_vcpu [options]\n"
									"-m    cpu mask\n"
									"-h    help\n");
							return;
					default:
							break;
			}
	}
	if(!mask){
		printf("\nNo Mask Specified.\n");
		return;
	}
	xlr_lib_stop_vcpu(mask);
	return;	
}

void get_curr_cpu_status(unsigned long *running, unsigned long *stopped,
			 unsigned long *total_loader_cpu)
{
#ifdef NETOS_LOADER_TODO
    *running = ~(lib_load.recent_avail_cpu)& lib_load.total_avail_cpu ;
    *stopped = lib_load.recent_avail_cpu;
    *total_loader_cpu = lib_load.total_avail_cpu;
#else
	*running = cpu_online_map;
    *stopped = ~cpu_online_map;
    *total_loader_cpu = 0xffffffff;
#endif /* NETOS_LOADER_TODO */
}

void xlr_cpu_status(int argc, char *argv[])
{
    unsigned long running = 0, stopped = 0, total = 0;

    get_curr_cpu_status(&running, &stopped, &total);

    printf("\n");
    printf("vCPUs in Stopped state   = 0x%08lx\n", stopped);
    printf("vCPUs in Running state   = 0x%08lx\n", running);
    printf("vCPUs under NetOS Ctrl. = 0x%08lx\n\n", total);
}

// loader should fill in per cpu info in this structure.
void fill_per_cpu_info(struct per_cpu_info_loader *ptr)
{
	ptr->num_tlbs = 0;
	ptr->ext_tlb_ptr = 0;
}

unsigned long long mmap_to_kseg(void *addr)
{
		return (long long)(long)addr;
}

void *kseg_to_mmap(unsigned long long addr)
{
		return (void *)(long)addr;
}

void send_ipi(uint32_t mask, uint32_t ipi)
{
	uint32_t i;
	uint32_t pid, tid;
	uint32_t val;

#ifndef XLP_SIM
	phoenix_reg_t *mmio = phoenix_io_mmio(PIC);
#endif

	for(i=0; i<XLR_NUM_CPU; i++){
		if(mask & (1U<<i)){
#if defined(XLP_SIM)
		        val = tid = pid = 0; /* For compiler sake */
			/* XLP TODO: currently setting the nmi & node id to zero.
			   To be correctly determined by a proper macro */
		        nlm_hal_pic_send_ipi(0, ipi, 0, i);
#else
			pid = i >> 2;
			tid = i % 4;
			val = (pid << 20) | (tid << 16) | ipi;
			phoenix_write_reg(mmio, PIC_IPI, val);
#endif
		}
	}
}

#ifndef XLP_SIM
void send_nmi_ipi(uint32_t mask)
{
    uint32_t i;
    uint32_t pid, tid;
    uint32_t val, ipi;

    ipi = 0; /* does not matter */


    phoenix_reg_t *mmio = phoenix_io_mmio(PIC);

    for(i=0; i<XLR_NUM_CPU; i++){
        if(mask & (1U<<i)){
            pid = i >> 2;
            tid = i % 4;
            val = (pid << 20) | (tid << 16) | (1 << 8) | ipi;
            phoenix_write_reg(mmio, PIC_IPI, val);
        }
    }
}
#endif

