/*************************************************************************
  Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
  reserved.
  Redistribution and use in source and binary forms, with or without
  modification, are permitted provided that the following conditions are met:
  1. Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.
  2. Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in
  the documentation and/or other materials provided with the
  distribution.
  THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
  ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
  IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
  PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
  LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
  CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
  SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
  INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
  CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
  ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
  THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/
#include <unistd.h>
#include <interface.h>

#include "xlr_boot_lib.h"
#include "elf.h"
#include "libelf.h"
#include "elf-ext.h"
#include "memblk.h"
#include "physaddrspace.h"
#include "nlm_hal_macros.h"
#include "shell.h"
#include "bitops.h"

#define Message(a,b...)

#define _2MB (2<<20)
#define BIT_SHARED 3
#define VERSION_MAGIC 0x900d900d
#define VERSION_1_6 0x10600		/*major<<16|minor<<8|subminor */
#define PRINT    {printf ("%s %d \n", __func__, __LINE__);}
#define LEAVE    {printf ("%s %d Leaving\n", __func__, __LINE__);}
#define ENTER    {printf ("%s %d Entering\n", __func__, __LINE__);}
#define NETOS_BL_KUSEG_MAP_ENTRIES 5
#define nlm_roundup(x, y)   (nlm_align((x)+(y)-1, (y)))
#define nlm_align(x, y)     ((x) & (~((y)-1)))

/* Global variables */
int lib_verbose = 0;
int nseg = 0;
Elf_ExtPhdr segment_map[MAX_ELF_SEGMENTS];
static int coredump_support = 1;
unsigned long kuseg_master_cpu_mask = 0;
static unsigned long kuseg_buddy_cpu_mask = 0;
unsigned int kseg_master_cpu;
static unsigned long lib_avail_kuseg_mem = 0;
static unsigned long lib_avail_kseg_mem = 0;
int kseg2_image = 0;		/* shared kseg2-kseg0 image, needed for MMU enabled vxWorks */
static struct xlr_lib_shared_mem *lib_shared_mem = NULL;
static uint64_t memory_reclaimed = 0;
volatile unsigned long success_mask = 0;
uint64_t max_app_phys_addr;
volatile struct xlr_vcpu_wakeup_info xlr_wakeup_info[MAX_CPUS];
static volatile struct vcpu_extended_info vcpu_extended_info;
struct per_cpu_info_loader per_cpu_info[MAX_CPUS];
static char *argv[] = { "bootstrap_psb_os", "" };
unsigned long long ptebase, excsp;
uint32_t cpu_buddy_map;

#ifdef NETOS_LOADER_TODO
static unsigned char *persistent_data = NULL;
#endif
static unsigned long persistent_data_length = 0;
struct proc_section_info *proc_section_info = (struct proc_section_info *) (NMI_BL_SHARED_REGION);

static uint32_t *proc_stack_info = 
        (uint32_t *) (NMI_BL_SHARED_REGION + (MAX_CPUS * sizeof (struct proc_section_info)));

uint64_t xlr_lib_kseg0_start = -1, xlr_lib_kseg0_size = -1;
struct kuseg_mem_map {
	uint64_t addr;
	uint64_t size;
} netos_bl_kuseg_mem_map[NETOS_BL_KUSEG_MAP_ENTRIES] = {
	{
	0, 0}, {
	0, 0}, {
	0, 0}, {
	0, 0}, {
0, 0},};

/* Global Extern Variables */
extern uint32_t cpu_online_map;
extern void coalesce_memblk();
extern void get_current_avail_memory_map(struct psb_mem_map *);
extern void memcpy64_phys_phys(uint64_t dst, uint64_t src, size_t n);
extern void __memset_64(uint64_t dst, int c, size_t n);
extern void phnx_initialize_getopt(void);
extern void *fdt;
extern void memcpy_virt_phys(unsigned char *dest, uint64_t src, int nbytes);
extern void memcpy_phys_virt(uint64_t dest, unsigned char *src, int nbytes);
extern int elf_download(struct elf_binary *elf, const int lib_run_mode, const int loading_nmi_image);
extern char app_mode[];

/* Global Extern Functions */
extern void loader_wakeup_cpu(unsigned int, unsigned long, unsigned long);
extern void xlr_send_stop_ipi(unsigned long);
extern void xlr_start_app(void);
extern int printf(const char *fmt, ...);
extern unsigned long long mmap_to_kseg(void *);
extern void *kseg_to_mmap(unsigned long long);
extern void send_ipi(uint32_t, uint32_t);
extern void xlr_lib_entry(unsigned long args);
extern void copy_per_cpu_tlb_info(int master_cpu, int buddy_cpu);
extern uint64_t free_per_cpu_tlb_list(int cpu_index, int flag);

/* Forward Declerations */
struct xlr_kuseg_mem_blk *free_kuseg_mem_blks();

/********************************************************************************************/

/* This lock is shared with hyperexec through wakeup_info structure */
lib_spinlock_t print_lock = {
	.lock = 0
};

unsigned int
acquire_print_lock(void)
{
	unsigned int flags = 0;
	lib_spin_lock_irqsave(&print_lock, flags);
	return flags;
}

void
release_print_lock(unsigned int flags)
{
	lib_spin_unlock_irqrestore(&print_lock, flags);
}

void
copy_data_uncached(unsigned long long dst, const void *src, unsigned int n)
{
	int i = 0;
	for (i = 0; i < n; ++i)
		sb_40bit_phys_uncached(dst + i, *((uint8_t *) src + i));
}

void
xlr_free_memory(uint64_t start, uint64_t size)
{
	if (!size)
		return;
	add_memblk(start, size);
	coalesce_memblk();
}

uint64_t
xlr_alloc_memory(uint64_t * size)
{
	if (*size > (512 * 1024 * 1024))
		return 0;
	*size = tlb_align_size(*size);
	if (!*size)
		return 0;
	max_app_phys_addr = MAX_PHYS;
	return find_memblk(*size);
}

static __inline__ int
count_one(uint32_t mask)
{
	int i = 0;
	int count = 0;

	for (i = 0; i < XLR_NUM_CPU; i++)
		if (mask & (1U << i))
			count++;
	return count;
}

/********************************************************************************************/
static void
reverse_shmem_list(struct xlr_lib_shared_mem *shmem)
{
	uint64_t phys[MAX_FRAGMENTS];
	uint64_t size[MAX_FRAGMENTS];
	uint64_t max_entries = shmem->entries;
	int i = 0;
	if (max_entries < 1)
		return;
	for (i = 0; i < max_entries; i++) {
		phys[i] = shmem->addr[i];
		size[i] = shmem->size[i];
	}
	for (i = 0; i < max_entries; i++) {
		shmem->addr[i] = phys[max_entries - i - 1];
		shmem->size[i] = size[max_entries - i - 1];
	}
	return;
}

struct xlr_lib_shared_mem *
xlr_lib_get_curr_shmem(void)
{
	if (lib_shared_mem->entries)
		return lib_shared_mem;
	return NULL;
}

struct xlr_lib_shared_mem *
xlr_lib_reserve_shmem(uint64_t size, uint64_t max_phys, int fragmented)
{
	uint64_t paddr;
	uint64_t orig_size;
	int i = 0;
	int max_fragments = 0;
	struct xlr_vcpu_wakeup_info *t = (struct xlr_vcpu_wakeup_info *)
		xlr_vcpu_wakeup_info;
	unsigned long alloc_size = 0;
	if (!max_phys)
		max_app_phys_addr = MAX_PHYS;
	else
		max_app_phys_addr = max_phys;

	if (!size) {
		printf("\nInvalid size specified.\n");
		return NULL;
	}

	if (size & (_2MB - 1)) {
		printf("\nShared memory size has to be 2 MB aligned.\n");
		return NULL;
	}

	if (lib_shared_mem->entries) {
		printf("Memory is already reserved\n");
		return lib_shared_mem;
	}

	if (!size) {
		printf("\nCouldnt find requested size chunk\n");
		return 0;
	}

	orig_size = size;

	if (fragmented)
		max_fragments = MAX_FRAGMENTS;
	else
		max_fragments = 1;

	for (i = 0; size && i < max_fragments;) {
		alloc_size = tlb_lower_align_size(size);
	  retry:
		paddr = find_memblk(alloc_size);
		if (!paddr) {
			if (alloc_size == _2MB) {
				/*No memory!! */
				break;
			}
			alloc_size = alloc_size / 4;
			goto retry;
		}
		lib_shared_mem->addr[i] = paddr;
		lib_shared_mem->size[i] = alloc_size;
		lib_shared_mem->entries++;
		size = size - alloc_size;
		i++;
	}

	if (size) {
		/*Couldn't allocate requested size memory */
		for (i = 0; i < lib_shared_mem->entries; i++) {
			add_memblk(lib_shared_mem->addr[i], lib_shared_mem->size[i]);
		}
		printf("Couldnt find requested size chunk\n");
		coalesce_memblk();
		memset(lib_shared_mem, 0, sizeof (struct xlr_lib_shared_mem));
		return NULL;
	}

	/*convert in to descending order */
	reverse_shmem_list(lib_shared_mem);

	for (i = 0; i < XLR_NUM_CPU; i++) {
		t->app_shared_mem_addr = lib_shared_mem->addr[0];
		t->app_shared_mem_size = lib_shared_mem->size[0];
		t->app_shared_mem_orig = t->app_shared_mem_size;
		t++;
	}
	lib_shared_mem->tot_size = orig_size;
	lib_avail_kuseg_mem -= orig_size;
	return lib_shared_mem;
}

void
xlr_reserve_mem(int argc, char *argv[])
{
	uint64_t size = 0;
	int i = 0;
	uint64_t max_phys = MAX_PHYS;
	char ch;
	int fragment = 1;
	struct xlr_lib_shared_mem *shmem = NULL;
	char *help =
		"------------------------------------------------------------------------------\n"
		"xlr_reserve_mem <options>\n"
		"------------------------------------------------------------------------------\n"
		"   -s      Size (in hex)\n"
		"   -M      Max legitimate physical addr (in hex)> \n"
		"   -c      Allocate only contiguous memory chunk\n"
		"   -p      Print current allocated buffer (list)\n"
		"   -h      Print help\n";

	if (argc < 2) {
		printf("Invalid number of argument\n");
		printf(help);
		return;
	}

	phnx_initialize_getopt();

	while ((ch = getopt(argc, argv, "pchs:M:")) != -1) {
		switch (ch) {
		case 's':
			size = strtoull(optarg, 0, 16);
			break;
		case 'M':
			max_phys = strtoull(optarg, 0, 16);
			break;
		case 'c':
			fragment = 0;
			break;
		case 'p':
			shmem = xlr_lib_get_curr_shmem();
			if (!shmem) {
				printf("No shared memory\n");
				return;
			}
			for (i = 0; i < shmem->entries; i++) {
				printf("segment %d - %#llx @ %#llx\n", i,
					   (unsigned long long) shmem->size[i],
					   (unsigned long long) shmem->addr[i]);
			}
			return;
		case 'h':
			printf(help);
			return;
		default:
			break;
		}
	}

	if (!size) {
		printf("Invalid Memory size\n");
		printf(help);
		return;
	}

	if (max_phys < (512 << 20)) {
		max_phys = MAX_PHYS;
		printf("Can't allocate shared memory below 512MB\n");
		return;
	}

	shmem = xlr_lib_reserve_shmem(size, max_phys, fragment);

	if (!shmem)
		return;

	for (i = 0; i < shmem->entries; i++) {
		printf("segment %d - %#llx @ %#llx\n", i,
			   (unsigned long long) shmem->size[i],
			   (unsigned long long) shmem->addr[i]);
	}
	return;
}

/********************************************************************************************/

int
is_this_kseg0_addr(unsigned long long phys)
{
	if ((phys < (0x10000000ULL)))
		return 1;
	return 0;
}

void
xlr_lib_get_avail_mem(unsigned long *kuseg_mem, unsigned long *kseg_mem)
{
	*kuseg_mem = lib_avail_kuseg_mem;
	*kseg_mem = lib_avail_kseg_mem;
}

static void
fill_vcpu_extended_info(struct vcpu_extended_info *ptr)
{

	ptr->magic = EXTENDED_BOOT_MAGIC;
	ptr->ver_major = 0;
	ptr->ver_minor = 1;
	ptr->size = sizeof (struct vcpu_extended_info);

#ifdef NETOS_LOADER_TODO
	ptr->per_cpu_info_size = sizeof (struct per_cpu_info_loader);
	ptr->per_cpu_info_ptr = mmap_to_kseg(per_cpu_info);
	ptr->uart0_lock = mmap_to_kseg(&device_lock.uart0_lock);
	ptr->uart1_lock = mmap_to_kseg(&device_lock.uart1_lock);
	ptr->global_lock = mmap_to_kseg(&device_lock.global_lock);
	ptr->bridge_io_lock = mmap_to_kseg(&device_lock.bridge_io_lock);
	for (i = 0; i < 8; i++)
		ptr->gmac_lock[i] = mmap_to_kseg(&device_lock.gmac_lock[i]);
	ptr->xgmac_lock[0] = mmap_to_kseg(&device_lock.xgmac_lock[0]);
	ptr->xgmac_lock[1] = mmap_to_kseg(&device_lock.xgmac_lock[1]);
	ptr->spi4_lock[0] = mmap_to_kseg(&device_lock.spi4_lock[0]);
	ptr->spi4_lock[1] = mmap_to_kseg(&device_lock.spi4_lock[1]);
	for (i = 0; i < NR_CPUS; i++) {
		fill_per_cpu_info((((struct per_cpu_info_loader *)
							(long) ptr->per_cpu_info_ptr) + i));
	}
#endif

	ptr->syscall = nlm_sys_call;
	ptr->fdtptr = fdt;			/* Store FDT Pointer for applications */
	ptr->dom_id = fw_arg2;		/* Store domain id passed firmware/hypervisor */
}


void
add_mem_map(uint64_t addr, uint64_t size)
{
	static int entries = 0;

	netos_bl_kuseg_mem_map[entries].addr = addr;
	netos_bl_kuseg_mem_map[entries].size = size;

	entries++;
}

void
xlr_load_init(void)
{
	int i;
#ifdef NETOS_LOADER_TODO
	uint64_t *t1;
	uint64_t xlr_lib_kuseg_add, xlr_lib_kuseg_size;
	int entries;
	int flag = 0;
#endif							/* NETOS_LOADER_TODO */

	register_shell_cmd("xlr_reserve_mem", (void *) xlr_reserve_mem, "reserve a block of physical memory");
	memset((void *) &xlr_wakeup_info[0], 0, MAX_CPUS * sizeof (struct xlr_vcpu_wakeup_info));
	memset((void *) &vcpu_extended_info, 0, sizeof (struct vcpu_extended_info));
	memset(&per_cpu_info[0], 0, MAX_CPUS * sizeof (struct per_cpu_info_loader));

	fill_vcpu_extended_info((struct vcpu_extended_info *) &vcpu_extended_info);

	/*Initialize netos lib / loader call data structures. */
	system_call_init();

	lib_shared_mem = (struct xlr_lib_shared_mem *) &(vcpu_extended_info.loader_shmem);

	for (i = 0; i < MAX_CPUS; i++) {
		xlr_wakeup_info[i].vcpu_extended_info_ptr = (struct vcpu_extended_info *) &vcpu_extended_info;
	}

	init_kuseg_memblks();

	/* NETOS_HACK : Allocate mem blocks
	 * based on bootloader kuseg mem map
	 * and not on boot1_info. This Hack will
	 * be cleaned up by FDT integration
	 */
	for (i = 0; i < NETOS_BL_KUSEG_MAP_ENTRIES; ++i) {

		if (!netos_bl_kuseg_mem_map[i].size)
			continue;

		add_memblk(netos_bl_kuseg_mem_map[i].addr, netos_bl_kuseg_mem_map[i].size);

		lib_avail_kuseg_mem += netos_bl_kuseg_mem_map[i].size;
	}

	if (!lib_avail_kuseg_mem) {
		printf ("!!!No Memory blocks available for application loading!!!\n");
	}

#ifdef NETOS_LOADER_TODO
	if (!persistent_data) {
		for (i = 0; i < boot1_info->psb_mem_map->nr_map; ++i) {
			if (is_this_kseg0_addr(boot1_info->psb_mem_map->map[i].addr)) {
				xlr_lib_kseg0_start = boot1_info->psb_mem_map->map[i].addr;
				xlr_lib_kseg0_size = boot1_info->psb_mem_map->map[i].size;
				lib_avail_kseg_mem = xlr_lib_kseg0_size;
			} else {
				add_memblk(boot1_info->psb_mem_map->map[i].addr,
						   boot1_info->psb_mem_map->map[i].size);
				lib_avail_kuseg_mem +=
					boot1_info->psb_mem_map->map[i].size;
			}
		}
	}
#endif
}

/*Implementation must call below routine before exiting*/
void
xlr_load_exit()
{
	uint64_t *data;
	struct xlr_kuseg_mem_blk *start;
	struct xlr_kuseg_mem_blk *tmp;
	int cnt = 0;
	unsigned int size = 0;
	unsigned long long tot_kuseg_mem = 0;
	unsigned long long tot_kseg_mem = 0;
	/*Below routine will free all kuseg memblks from kuseg_mem_blk list. */
	tmp = start = (struct xlr_kuseg_mem_blk *) free_kuseg_mem_blks();

	for (cnt = 0; tmp; cnt++, tmp = tmp->next) ;
	/*We have to store cnt kuseg entries. Each entry requires 16 bytes */
	size = cnt * 16 + 8;		/*64bit phys, 64bit len and 64bit for no. of entries */
	if (xlr_lib_kseg0_start != -1)
		size += (16 + 8);		/*64bit phys, 64bit len and 64bit for no. of entries */
	size += 8;					/*Magic No. */
	if (size < 1024)
		size = 1024;
	data = (uint64_t *) malloc(size);
	persistent_data_length = size;
	memset((void *) data, 0, size);
	*data = XLR_PERSISTENT_MEMORY_MAGIC;
	data++;
	if (xlr_lib_kseg0_start != -1) {
		*data = KSEG0_MEM_MAP;
		data++;
		*data = 1;
		data++;
		*data = xlr_lib_kseg0_start;
		data++;
		*data = xlr_lib_kseg0_size;
		data++;
		tot_kseg_mem += xlr_lib_kseg0_size;
	}
	if (cnt) {
		*data = KUSEG_MEM_MAP;
		data++;
		*data = cnt;
		data++;
		while (start) {
			*data = start->phys;
			data++;
			*data = start->size;
			data++;
			tot_kuseg_mem += start->size;
			tmp = start;
			start = start->next;
			free(tmp);
		}
	}
}

/********************************************************************************************/
static int
xlr_lib_elfload(struct xlr_lib_boot_file *file, int lib_run_mode, int loading_nmi_image)
{
	int res;
	char *file_start_xkphys_address =
		(char *) (NLH_XKPHYS | ((uint64_t) CCA_CACHED << 59) | (file->start & 0xffffffffffULL));
	struct elf_binary elf = { 0 };

	res = elf_init(&elf, file_start_xkphys_address, 0);
	if (res) {
		printf("%s %d, Error in initialising \n", __func__, __LINE__);
		return res;
	}
	return elf_download(&elf, lib_run_mode, loading_nmi_image);
}


void
setup_coredump_info(struct xlr_vcpu_wakeup_info *t, int cpu)
{
	int i = 0;
	struct proc_section_info tmp;
	unsigned long global_stack = 0;

	if (!coredump_support)
		return;

	for (i = 0; i < MAX_ELF_SEGMENTS; i++) {
		tmp.vaddr[i] = t->seg_info[i].vaddr;
		tmp.size[i] = t->seg_info[i].memsz;
	}
	for (i = 0; i < MAX_TLB_MAPPINGS; i++) {
		tmp.tlb_vaddr[i] = t->map[i].virt;
		tmp.tlb_phys[i] = t->map[i].phys;
		tmp.tlb_size[i] = t->map[i].page_size;
	}
	tmp.mode = t->mode;

	copy_data_uncached((unsigned long long) (unsigned long)
					   (proc_section_info + cpu),
					   (const void *) &tmp, (unsigned int) sizeof (tmp));
	global_stack = xlr_vcpu_wakeup_info->gp;
	copy_data_uncached((unsigned long long) (unsigned long)
					   proc_stack_info,
					   (const void *) &global_stack,
					   (unsigned int) sizeof (global_stack));
}

int
xlr_load_kseg_app(struct xlr_lib_boot_file *file,
				  uint32_t master, int argc,
				  char *argv[], int loading_nmi_image)
{
	int res;
	int i = 0;

	max_app_phys_addr = MAX_PHYS;
	//loading_nmi_image = nmiload;

	/*In case of generic bootloader kseg_master_cpu is always 0x1. */
	kseg_master_cpu = master;

	if (!loading_nmi_image) {
		if (((1U << kseg_master_cpu) & cpu_online_map) !=
			(1U << kseg_master_cpu)) {
			printf
				("\nCant launch KSEG application as specified master is not free.\n");
			return -1;
		}

		for (i = 0; i < XLR_NUM_CPU; i++) {
			if ((cpu_online_map & (1U << i))) {
				struct xlr_vcpu_wakeup_info *b =
					(struct xlr_vcpu_wakeup_info *)
					xlr_vcpu_wakeup_info + i;
				b->valid = 0;
			}
		}

		if (xlr_lib_kseg0_start == -1) {
			printf
				("\nAlready One KSEG0 application is running, cant launch another.\n");
			return -1;
		}
	}
	res = xlr_lib_elfload(file, KSEG0_APP, loading_nmi_image);
	//loading_nmi_image = 0;
	/* cpu_online_map will be updated in "userapp_run" */
	return res;
}

int
xlr_load_kuseg_app(struct xlr_lib_boot_file *file, uint32_t master,
				   uint64_t loadaddr, char *user_app_addr,
				   uint32_t buddy, int override,
				   int nargs, char *cmdline_args[],
				   char *env_args[], int verbose, uint64_t max_phys)
{
	unsigned long mask;
	int argc;
	int i, k;
	int ret = 0;
	int this_cpu = processor_id();

	lib_verbose = verbose;

	if (max_phys) {
		max_app_phys_addr = max_phys;
	} else {
		max_app_phys_addr = MAX_PHYS;
	}

	if ((master & cpu_online_map) != master) {
		printf
			("\nAll specified master vcpus are not under loader control.");
		printf("\nVCPU Mask is %#lx\n", (unsigned long) cpu_online_map);
		return -1;
	}

	if ((buddy & cpu_online_map) != buddy) {
		printf("\nAll specified buddy vcpus are not under loader\
        control.\n");
		printf("\nTrying to launch an application with buddy mask\
        %#lx\n", (unsigned long) buddy & cpu_online_map);
		buddy = buddy & cpu_online_map;
	}

	if (((master | buddy) & (cpu_online_map)) != (master | buddy)) {
		printf("\n%#lx vcpus are already in use\n",
			   ~cpu_online_map & (unsigned long) cpu_online_map);
		return -1;
	}
#if 0
	if (xlr_lib_kseg0_start == -1) {
		printf
			("\nKseg Application Is Running, You cant launch any Kuseg Application Now.\n");
		return -1;
	}
#endif

	/*Reset Validity of All vcpus who are not there in cpu_online_map */
	for (i = 0; i < XLR_NUM_CPU; i++) {
		if ((cpu_online_map & (1U << i))) {
			struct xlr_vcpu_wakeup_info *t =
				(struct xlr_vcpu_wakeup_info *)
				xlr_vcpu_wakeup_info + i;
			t->valid = 0;
		}
	}

	if (buddy && (count_one(master) != 1)) {
		printf("buddy mask specified, only one master thread allowed\n");
		return -1;
	}

	/*@ Here, we are sure that all the required cpus are free */

	if (buddy & master) {
		/*Remove master frm buddy list. */
		buddy = buddy & (~master);
	}

	/*Or of kuseg_master and kuseg_buddy cpus. */
	mask = master | buddy;
	cpu_online_map = cpu_online_map & (~mask);
	kuseg_master_cpu_mask = master;
	kuseg_buddy_cpu_mask = buddy;
	argc = 1;
	if (loadaddr) {
		argv[1] = NULL;
		argc++;
		file->start = loadaddr;
	}
	if (user_app_addr) {
		file->start = strtoull(user_app_addr, 0, 16);
	}

	/* Do the elfload for all master cpus
	 */
	success_mask = 0;

	ret = xlr_lib_elfload(file, KUSEG_APP, 0);
	if (ret) {
		if (!success_mask) {
			cpu_online_map |= mask;
			return -1;
		} else {
			cpu_online_map |= (mask & ~success_mask);
		}
	}

	/*Update boot1_info psb_os_cpu_map - this contains vcpus being used */
	boot1_info->psb_os_cpu_map = (~cpu_online_map) & (cpu_online_map);

	/*scuccess mask contains mask of master vcpus for which elfload is done successfully */
	printf("\nApplication Launched Successfully On cpu %#lx\n",
		   success_mask | buddy);
	if(buddy != 0)
		cpu_buddy_map =  success_mask | buddy;

	/*@ here all master cpus xlr_vcpu_wakeup_info structure is filled up with 
	   tlb entries, now setup argc,env,mode etc in per vcpu data structure and 
	   call wakeup. To wakeup cpus we will use callback provided by implementation,
	   it must call "wakeup" defined in xlr_boot_lib_launch.c */

	for (i = 0; i < XLR_NUM_CPU; i++) {
		struct xlr_vcpu_wakeup_info *t;

		if ((success_mask & (1U << i)) == 0)
			continue;

		unsigned char *argp = NULL;
		unsigned char *envp = NULL;
		int j = 0;

		t = (struct xlr_vcpu_wakeup_info *) xlr_vcpu_wakeup_info + i;

		t->argc = nargs;
		memset(t->argv_buf, 0, XLR_ARGV_BUF_SIZE);
		for (j = 0, argp = t->argv_buf; j < nargs; ++j) {
			t->argv[j] = argp;
			strcpy((char *) argp, cmdline_args[j]);
			argp += (strlen(cmdline_args[j]) + 1);
		}

		memset(t->env_buf, 0, XLR_ENV_BUF_SIZE);
		for (j = 0, envp = t->env_buf; env_args[j] && j < 31; ++j) {
			t->envs[j] = envp;
			strcpy((char *) envp, env_args[j]);
			envp += (strlen(env_args[j]) + 1);
			t->envs[j + 1] = NULL;
		}
		t->env = j;

		t->master_cpu = i;
		t->buddy_mask = buddy;
		t->psb_os_cpu_map = mask;
		t->master_mask = success_mask;

		t->mode = KUSEG_APP;

		t->printk_lock = (lib_spinlock_t *) & print_lock;

		for (j = 0; j < XLR_NUM_CPU; j++) {

			if ((buddy & (1U << j)) == 0)
				continue;

			struct xlr_vcpu_wakeup_info *b =
				(struct xlr_vcpu_wakeup_info *) xlr_vcpu_wakeup_info + j;

			/* Copy All Required Info from Master */
			b->master_cpu = t->master_cpu;
			b->buddy_mask = t->buddy_mask;
			b->psb_os_cpu_map = t->psb_os_cpu_map;

			b->mode = t->mode;

			b->printk_lock = t->printk_lock;

			b->map_count = t->map_count;
			printf("[%s] i=%d, j=%d, map_count=%d, &b->map_count=%p\n",
				   __func__, i, j, t->map_count, &b->map_count);

			b->valid = t->valid;
			b->master_mask = t->master_mask;

			for (k = 0; k < MAX_TLB_MAPPINGS; k++)
				b->map[k] = t->map[k];

			for (k = 0; k < MAX_ELF_SEGMENTS; k++)
				b->seg_info[k] = t->seg_info[k];

			setup_coredump_info(b, j);
			copy_per_cpu_tlb_info(i, j);
		}

		setup_coredump_info(t, i);

		if (i != this_cpu) {
			/* We will jump after wakin up all other vcpus */
			loader_wakeup_cpu(i, (unsigned long) t->func, 0xdeadUL);
		}
	}

	if (success_mask & (1U << this_cpu)) {
		struct xlr_vcpu_wakeup_info *t;
		t = (struct xlr_vcpu_wakeup_info *) xlr_vcpu_wakeup_info +
			this_cpu;

		t->args = 0xdeadUL;

#ifdef DEAD_CODE
		if (!xlr_start_app) {
			printf("\nCant launch an application on Cpu %d\n", this_cpu);
			cpu_online_map |= (1U << this_cpu);
			return -1;
		}
#endif
		setup_coredump_info(t, this_cpu);

		xlr_start_app();

	} else if (buddy & (1U << processor_id())) {

		xlr_lib_entry(0UL);
	}

	return 0;
}

static void
get_rest_vcpu_mask(unsigned int master)
{
	int i = 0;
	uint32_t psb_os_mask = 0;
	unsigned long current_kuseg_vcpus = ~cpu_online_map & (cpu_online_map);

	// create a psb_os mask such that if one of the threads of a core
	// is active, mark all other threads active in that core

	psb_os_mask = 0;
	for (i = 0; i < XLR_NUM_CPU; i += 4) {
		if ((current_kuseg_vcpus & (0xfU << i)) != 0) {
			psb_os_mask |= (0xfU << i);
		}
	}

	printf("===========boot1_info================\n");
	printf
		("cpu_online_map=%x, userapp_cpu_map %x\n",
		 (unsigned int) cpu_online_map, (unsigned int) cpu_online_map);
	printf
		("psb_os_active_mask=%x, psb_os_mask=%x\n",
		 (unsigned int) current_kuseg_vcpus, (unsigned int) psb_os_mask);

	boot1_info->psb_os_cpu_map = current_kuseg_vcpus;
	boot1_info->userapp_cpu_map =
		(cpu_online_map & ~psb_os_mask) & cpu_online_map;
	boot1_info->userapp_cpu_map |= (1U << master);

	// below is for backward compatibility
	boot1_info->cpu_online_map = boot1_info->userapp_cpu_map;

	printf
		("boot1_info: userapp_cpu_map=%llx, psb_os_cpu_map=%llx\n",
		 (unsigned long long) boot1_info->userapp_cpu_map,
		 (unsigned long long) boot1_info->psb_os_cpu_map);
	printf
		("            cpu_online_map = %#llx\n",
		 (unsigned long long) boot1_info->cpu_online_map);
}

void
xlr_lib_set_userapp_mask(unsigned long mask)
{
	cpu_online_map = mask;
}

static void
reset_fields(struct xlr_vcpu_wakeup_info *t)
{
	int i = 0;
	t->valid = t->func = t->args = t->master_cpu = t->master_mask =
		t->mode = t->buddy_mask = t->psb_os_cpu_map = t->env = t->argc =
		t->kseg_master = 0;

	for (i = 0; i < XLR_MAX_ARGV; i++)
		t->argv[i] = NULL;

	for (i = 0; i < XLR_MAX_ENV; i++)
		t->envs[i] = NULL;

	memset(&t->argv_buf[0], 0, XLR_ARGV_BUF_SIZE);
	memset(&t->env_buf[0], 0, XLR_ENV_BUF_SIZE);
	memset(&t->seg_info[0], 0, sizeof (struct core_segment_info));
	memset(&t->map[0], 0,
		   sizeof (struct cpu_tlb_mapping) * MAX_TLB_MAPPINGS);
}

void
lib_reclaim_resources(struct xlr_vcpu_wakeup_info *t, int cpu)
{
	unsigned long mask;
	int i = 0;
	struct xlr_vcpu_wakeup_info *b;

	mask = t->master_mask;
	if (t->mode == KSEG0_APP) {
		/*KSEG0 App.
		   -If this is a last cpu then reclaim kseg0 memory
		   -Else reclaim only cpu and mark this cpu as a non buddy
		   of the remaining cpus in this set.
		 */
		if ((t->master_mask | t->buddy_mask) == (1 << cpu)) {
			/*Last cpu. */
			xlr_lib_kseg0_start = t->map[0].phys;
			xlr_lib_kseg0_size = t->map[0].page_size;
			memory_reclaimed = xlr_lib_kseg0_size;
		} else if (t->master_cpu == cpu) {
			for (i = 0; i < MAX_CPUS; i++) {
				if (i == cpu)
					continue;
				if ((t->buddy_mask) & (1 << i)) {
					/*Remove master cpu from all buddy's list. */
					b = (struct xlr_vcpu_wakeup_info
						 *) (xlr_vcpu_wakeup_info) + i;
					b->master_cpu = -1;
					b->kseg_master = -1;
					b->buddy_mask = b->buddy_mask & ~(1 << cpu);
					b->master_mask = b->master_mask & ~(1 << cpu);
				}
			}
		} else {
			for (i = 0; i < MAX_CPUS; i++) {
				if (i == cpu)
					continue;
				if ((t->buddy_mask | t->master_mask) & (1 << i)) {
					/*Remove this buddy from all other cpus list */
					b = (struct xlr_vcpu_wakeup_info
						 *) (xlr_vcpu_wakeup_info) + i;
					b->buddy_mask = b->buddy_mask & ~(1 << cpu);
				}
			}
		}
	} else if (count_one(mask) && t->buddy_mask == 0x0) {
		/*MORE THAN ONE MASTER CPUS - NO BUDDY CPUS */
		if (t->master_mask == (1 << cpu)) {
			/*Last cpu */
			/*Reclaim entire memory. */
			/*Reclaim cpu */
			memory_reclaimed += free_per_cpu_tlb_list(cpu, FREE_TLB_ALL);
			coalesce_memblk();
		} else {
			for (i = 0; i < MAX_CPUS; i++) {
				if (i == cpu)
					continue;
				if (t->master_mask & (1 << i)) {
					b = (struct xlr_vcpu_wakeup_info
						 *) (xlr_vcpu_wakeup_info) + i;
					b->master_mask = b->master_mask & ~(1 << cpu);
				}
			}
			/*claim non-shared memory segments only */
			memory_reclaimed +=
				free_per_cpu_tlb_list(cpu, FREE_TLB_NONSHARED);
			coalesce_memblk();
		}
	} else {
		/*MASTER- BUDDY MODE */

		if ((t->master_mask | t->buddy_mask) == (1 << cpu)) {
			/*Last cpu */
			/*Reclaim memory */
			/*Reclaim cpu */
			memory_reclaimed += free_per_cpu_tlb_list(cpu, FREE_TLB_ALL);
			coalesce_memblk();
		} else if (t->master_cpu == cpu) {
			/*`cpu` is master */
			for (i = 0; i < MAX_CPUS; i++) {
				if (i == cpu)
					continue;
				if ((t->buddy_mask | t->master_mask) & (1 << i)) {
					b = (struct xlr_vcpu_wakeup_info
						 *) (xlr_vcpu_wakeup_info) + i;
					b->master_cpu = -1;
					b->buddy_mask = b->buddy_mask & ~(1 << cpu);
					b->master_mask = b->master_mask & ~(1 << cpu);
				}
			}
		} else {
			/*`cpu` is buddy */
			for (i = 0; i < MAX_CPUS; i++) {
				if (i == cpu)
					continue;
				if ((t->buddy_mask | t->master_mask) & (1 << i)) {
					b = (struct xlr_vcpu_wakeup_info
						 *) (xlr_vcpu_wakeup_info) + i;
					b->buddy_mask = b->buddy_mask & ~(1 << cpu);
				}
			}
		}
	}
	reset_fields(t);
	cpu_online_map = cpu_online_map | (1U << cpu);
	t->cpu_status = CPU_RESOURCE_RELEASED;
}

int
xlr_lib_stop_vcpu(unsigned long mask)
{
	unsigned long current_used_vcpu = (~cpu_online_map) & (cpu_online_map);
	volatile struct xlr_vcpu_wakeup_info *t;
	int i;

	memory_reclaimed = 0;

	if ((mask & current_used_vcpu) != mask) {
		mask = mask & current_used_vcpu;
		if (mask) {
			printf
				("\nAll Specified vcpus are not in use, stopping %#lx vcpus\n",
				 mask);
		} else {
			printf("\nNone of the specified vcpu is running.\n");
			return -1;
		}
	}

	xlr_send_stop_ipi(mask);

	for (i = 0; i < XLR_NUM_CPU; i++) {
		if (!((1 << i) & mask))
			continue;
		t = xlr_vcpu_wakeup_info + i;

		if (t->cpu_status != CPU_WAIT_TO_RELEASE_RESOURCE)
			continue;
		/*Resource associated with this cpu can be reclaimed. */
		lib_reclaim_resources((struct xlr_vcpu_wakeup_info *) t, i);
	}

	printf("\nSent stop ipi to %#lx cpus", mask);
	printf("\n%#llx memory (in bytes) reclaimed\n",
		   (unsigned long long) memory_reclaimed);
	return 0;
}

int
xlr_load_userapp(unsigned int cpu, int nargs, char *cmdline_args[],
				 char *env_args[])
{
	struct xlr_vcpu_wakeup_info *t = (struct xlr_vcpu_wakeup_info *)
		xlr_vcpu_wakeup_info + cpu;
	unsigned char *argp = NULL;
	unsigned char *envp = NULL;
	struct psb_mem_map avail_mem_map;

	uint64_t phys_base =
		boot1_info->avail_memory_map & (uint64_t) 0x000000001fffffff;

	int j = 0, i = 0, k = 0;

	if (!t->valid) {
		printf("[%s]: Bad elf_entry\n", __FUNCTION__);
		return -1;
	}

	get_rest_vcpu_mask(cpu);
	t->mode = KSEG0_APP;
	t->master_cpu = cpu;
	t->master_mask = (0x1U << cpu);

	/* Update Memory Map
	 */
	get_current_avail_memory_map(&avail_mem_map);
	memcpy_phys_virt(phys_base, (void *) &avail_mem_map,
					 sizeof (avail_mem_map));

	for (i = 0; i < XLR_NUM_CPU; i++) {
		if (boot1_info->userapp_cpu_map & (1U << i)) {
			struct xlr_vcpu_wakeup_info *b =
				(struct xlr_vcpu_wakeup_info *)
				xlr_vcpu_wakeup_info + i;
			b->valid = 1;
			b->buddy_mask = boot1_info->userapp_cpu_map & ~(1U << cpu);
			b->mode = t->mode;
			b->master_cpu = t->master_cpu;
			b->master_mask = t->master_mask;
			b->printk_lock = t->printk_lock;
			b->map[0].phys = t->map[0].phys;
			b->map[0].page_size = t->map[0].page_size;
			for (k = 0; k < 16; k++)
				b->seg_info[k] = t->seg_info[k];
			setup_coredump_info(b, i);
		}
	}

	t->argc = nargs;
	memset(t->argv_buf, 0, XLR_ARGV_BUF_SIZE);
	for (j = 0, argp = t->argv_buf; j < nargs; ++j) {
		t->argv[j] = argp;
		strcpy((char *) argp, cmdline_args[j]);
		argp += (strlen(cmdline_args[j]) + 1);
	}
	memset(t->env_buf, 0, XLR_ENV_BUF_SIZE);
	for (j = 0, envp = t->env_buf; env_args[j] && j < 31; ++j) {
		t->envs[j] = envp;
		strcpy((char *) envp, env_args[j]);
		envp += (strlen(env_args[j]) + 1);
		t->envs[j + 1] = NULL;
	}
	t->env = j;

	/* Mark All vcpus as busy
	 */
	cpu_online_map = cpu_online_map & ~(boot1_info->userapp_cpu_map);

	xlr_lib_kseg0_start = -1;
	xlr_lib_kseg0_size = -1;

	if (kseg2_image) {
		printf
			("Loaded application, ELF_ENTRY @ 0x%lx (mapped from 0x%lx in KSEG2 mode)\n",
			 t->func, t->func | 0xc0000000);
	} else {
		printf("Loaded application, ELF_ENTRY @ 0x%lx\n", t->func);
	}
	return 0;
}

void alloc_per_cpu_mem_space()
{
    uint64_t size, isize = -1, addr;
    int ncpus = 0, smpmode = 0, index = NETOS_BL_KUSEG_MAP_ENTRIES, i;

    if (strcmp(app_mode,"smpload") == 0)
        smpmode = 1;

    ncpus = num_ones(cpu_online_map);

    /*
     *      * Alloc the exception stack , we need this to accessed as unmapped space 
     *           */
    size = NETOS_EXC_STACK_SZ * ncpus;

    if (smpmode) /* pte will be shared */
        size += sizeof(struct tlb_struct) * PHYS_VIRT_NUM_PTE;
    else
        size += sizeof(struct tlb_struct) * PHYS_VIRT_NUM_PTE * ncpus;

    /* 
     *      * find out the lowest phys block that can accomadate this, min is 2MB 
     *           */
    size = nlm_roundup(size, 0x200000);

    for (i = 0 ; i < NETOS_BL_KUSEG_MAP_ENTRIES; ++i) {
        if (!netos_bl_kuseg_mem_map[i].size)
            continue;
        if(netos_bl_kuseg_mem_map[i].size < size)
            continue;
        if(isize > netos_bl_kuseg_mem_map[i].size) {
            index = i;
            isize = netos_bl_kuseg_mem_map[i].size;
        }
    }
    if(index == NETOS_BL_KUSEG_MAP_ENTRIES) {
        printf("Mem alloc for stack and pte failed allocsize %ld ncpus %d\n", (long)size, ncpus);
        cpu_halt();
    }
    netos_bl_kuseg_mem_map[index].size -= size;
    addr = netos_bl_kuseg_mem_map[index].addr + netos_bl_kuseg_mem_map[index].size;

    excsp = NETOS_XKPHYS_CACHED | addr;

    ptebase = NETOS_XKPHYS_CACHED | (addr + NETOS_EXC_STACK_SZ * ncpus);

    //printf("excsp = %lx ptebase = %lx\n", (long)excsp, (long)ptebase);
}
