/*************************************************************************
 Copyright 2003-2010 Netlogic Microsystems ("Netlogic"). All rights
 reserved.
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are
 met:
 1. Redistributions of source code must retain the above copyright
    notice, this list of conditions and the following disclaimer.
 2. Redistributions in binary form must reproduce the above copyright
    notice, this list of conditions and the following disclaimer in
    the documentation and/or other materials provided with the
    distribution.
 THIS SOFTWARE IS PROVIDED BY Netlogic Microsystems ``AS IS'' AND
 ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL NETLOGIC OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 THE POSSIBILITY OF SUCH DAMAGE.
*******************************#NETL_2#**********************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "memblk.h"
#include "xlr_boot_lib.h"
#include "physaddrspace.h"

extern uint64_t max_app_phys_addr;
uint64_t kuseg_avail_mem = 0;
memblk_list_t kuseg_memblks[NTLBSZ];

void coalesce_memblk();
static struct xlr_kuseg_mem_blk *xlr_memblk_store = NULL;

unsigned int xlr_tlb_sizes[NTLBSZ] = {
	1 << 10,        //   1 KB
	4 << 10,        //   4 KB
	16 << 10,       //  16 KB
	64 << 10,       //  64 KB
	1 << 20,        //   1 MB
	4 << 20,        //   4 MB
	16 << 20,       //  16 MB
	64 << 20,       //  64 MB
	256 << 20       // 256 MB
};

unsigned long tlb_lower_align_size(uint64_t size)
{
	int i;
	for (i = NTLBSZ - 1; i >= 4; i--) {
		if (!(size % (2 * xlr_tlb_sizes[i])))
			return 2 * xlr_tlb_sizes[i];
	}
	return 0;
}

unsigned long tlb_align_size(uint64_t size)
{
	int i;

	for (i = 4; i < NTLBSZ ; i++) {
		if(size <= (2*xlr_tlb_sizes[i]))
			return (2*xlr_tlb_sizes[i]);
	}
	return 0;
}

unsigned int best_fit_tlb_size(uint64_t addr, uint64_t size)
{
	int i, pass;

	for (i = NTLBSZ - 1, pass = 0; i >= 4; --i, pass = 0) {
		pass += ((addr % (2 * xlr_tlb_sizes[i])) == 0);
		pass += (size >= (2 * xlr_tlb_sizes[i]));
		if (pass == 2)
			return xlr_tlb_sizes[i];
	}

	return 0;
}

void init_kuseg_memblks(void)
{
	int i;

	for (i = 0; i < NTLBSZ; ++i) {
		kuseg_memblks[i].align = 2 * xlr_tlb_sizes[i];
		list_init( &(kuseg_memblks[i].list));
	}

	return;
}

int add_memblk(uint64_t memblk_addr, uint64_t memblk_size)
{
	int i;
	memblk_t *memblk_new;
	unsigned int tlb_size;
	memblk_list_t *memblk_lv;

	if (!memblk_size)
		return 0;

	for ( ; memblk_size; memblk_size -= 2 * tlb_size, memblk_addr += 2 * tlb_size) {
		tlb_size = best_fit_tlb_size(memblk_addr, memblk_size);
		if (tlb_size == 0)
			return -1;

		memblk_new = (memblk_t *) malloc(sizeof(memblk_t));
		if (!memblk_new) {
			printf ("malloc() failure\n");
			return -1;
		}
		memblk_new->addr = memblk_addr;
		memblk_new->size = 2 * tlb_size;

		kuseg_avail_mem += memblk_new->size;

		for (i = NTLBSZ - 1; i >= 0; --i) {
			memblk_lv = &kuseg_memblks[i];

			if (memblk_lv->align == 2 * tlb_size) {
				list_t *list_ptr;
				memblk_t *memblk_v;
	
				list_iterate_forward(&(memblk_lv->list),list_ptr)
				{
					memblk_v = list_container_of(memblk_t, list_ptr, list);
	
					if (memblk_v->addr > memblk_addr) {
						list_add_front( &(memblk_v->list), &(memblk_new->list));
						goto end;
					}
				}
				list_add_front( &(memblk_lv->list), &(memblk_new->list));
end:
				break;
			}
		}
	}

	return 0;
}

static uint64_t _find_memblk(unsigned long size)
{
	int i;
	list_t *list_ptr;
	memblk_t *memblk_v;
	memblk_list_t *memblk_lv;
	uint64_t retval;

	retval = 0;

	for (i = 0; i < NTLBSZ; ++i) {
		if (size == kuseg_memblks[i].align)
			break;
	}
	if (i == NTLBSZ)
		return 0;

	for (; i < NTLBSZ; ++i) {
		memblk_lv = &kuseg_memblks[i];

		list_iterate_forward(&(memblk_lv->list), list_ptr)
		{
			list_delete(list_ptr);
			memblk_v = list_container_of(memblk_t, list_ptr, list);
	        retval = memblk_v->addr;
    		kuseg_avail_mem -= memblk_v->size;
    		add_memblk(memblk_v->addr + size, memblk_v->size - size);
    		free (memblk_v);
			return retval;
		}
	}

	return 0;
}

int store_memblk(uint64_t phys, uint64_t size)
{
    struct xlr_kuseg_mem_blk  *blk = NULL;
    struct xlr_kuseg_mem_blk *tmp = xlr_memblk_store;

    blk = (struct xlr_kuseg_mem_blk *)malloc(sizeof(struct xlr_kuseg_mem_blk));
    if (!blk)
        return -1;

    blk->phys = phys;
    blk->size = size;
    blk->next = NULL;

    if (!tmp)
        tmp = xlr_memblk_store = blk;
	else {
        while (tmp->next)
            tmp = tmp->next;
        tmp->next = blk;
    }

    return 0;
}

void restore_memblk_list()
{
    struct xlr_kuseg_mem_blk *blk = xlr_memblk_store;
    struct xlr_kuseg_mem_blk *tmp ;

    while (blk) {
        add_memblk(blk->phys, blk->size);
        tmp = blk;
        blk = blk->next;
        free(tmp);
    }

    xlr_memblk_store = NULL;

    coalesce_memblk();
}

int memblk_errno;

uint64_t find_memblk(unsigned long size)
{
    uint64_t phys = 0;

	memblk_errno = 0;
    do {
        phys = _find_memblk(size);
        if (phys == 0) {
            restore_memblk_list();
			memblk_errno = 1;
            return 0;
        }

        if ((phys + size) > max_app_phys_addr) {
            if (store_memblk(phys,size)) {
                restore_memblk_list();
				memblk_errno = 2;
                return 0;
            }
        }
		else {
            restore_memblk_list();
            return phys;
        }
    } while(1);

}

int display_kuseg_meminfo(void)
{
	list_t *list_index;
	memblk_t *memblk_v;
	memblk_list_t *memblk_lv;
	int i;

	printf ("Available application memory: 0x%llx\n", (unsigned long long) kuseg_avail_mem);

	printf ("------ Memory Blocks (aligned to TLB Page Sizes) ------\n");
	for (i = 0; i < NTLBSZ; ++i) {
		memblk_lv = &kuseg_memblks[i];

		list_iterate_forward( &(memblk_lv->list), list_index)
		{
			memblk_v = list_container_of(memblk_t, list_index, list);
			printf("@ addr: 0x%llx size: 0x%llx\n",
			       (unsigned long long) memblk_v->addr,
			       (unsigned long long) memblk_v->size);
		}
	}
	printf ("\n");

	return 0;
}

struct xlr_kuseg_mem_blk * free_kuseg_mem_blks()
{
	list_t *list_index;
	list_t *del_index = NULL;
	memblk_t *memblk_v;
	memblk_list_t *memblk_lv;
	int i;
	unsigned long long lowest_addr ;
	unsigned long long lowest_addr_size = 0x0;
	int pass;
	struct xlr_kuseg_mem_blk *start = NULL, *end = NULL;

	while (1) {
		lowest_addr = 0xffffffffffULL;
		pass = 0;
	 	for (i = 0; i < NTLBSZ; ++i) {
			memblk_lv = &kuseg_memblks[i];
			list_iterate_forward(&(memblk_lv->list),list_index)
			{
				memblk_v = list_container_of(memblk_t, list_index, list);
				if ((memblk_v->addr < lowest_addr)) {
					lowest_addr = memblk_v->addr;
					lowest_addr_size = memblk_v->size;
					del_index = list_index; //list_del(list_ptr);
					pass = 1;
				}
			}
		}
		if (!pass)
			break;

		memblk_v = list_container_of(memblk_t, del_index, list);
		kuseg_avail_mem -= memblk_v->size;
		list_delete(del_index);
		free (memblk_v);

		while (1) {
		restart:
			pass = 0;
		 	for (i = 0; i < NTLBSZ; ++i)  {
				memblk_lv = &kuseg_memblks[i];
				list_iterate_forward(&(memblk_lv->list),list_index)
				{
					memblk_v = list_container_of(memblk_t, list_index, list);
					if (memblk_v->addr == (lowest_addr + lowest_addr_size)) {
						lowest_addr_size += memblk_v->size;
						kuseg_avail_mem -= memblk_v->size;
						list_delete(list_index);
						free(memblk_v);
						pass = 1;
						goto restart;
					}
				}
			}
			if (!pass)
				break;
		}
		/*Add new entry, lowest_addr and lowest_addr_size*/
		if (!end)
			start = end = malloc(sizeof(struct xlr_kuseg_mem_blk));
		else {
			end->next = malloc(sizeof(struct xlr_kuseg_mem_blk));
			end = end->next;
		}
		end->phys = lowest_addr;
		end->size = lowest_addr_size;
		end->next = NULL;
	}
	return start;
}

void coalesce_memblk()
{
	struct xlr_kuseg_mem_blk *start = NULL, *end = NULL;

	/*
	 * Delete All Entries from memblk_list_t
	 */
	start = free_kuseg_mem_blks();

	/*
	 * Whole kuseg_mem list is empty.. Add all entries here.
	 */
	while (start)  {
		add_memblk(start->phys, start->size);
		end = start;
		start = start->next;
		free(end);
	}
}

void get_current_avail_memory_map(struct psb_mem_map *mem_map)
{
    struct xlr_kuseg_mem_blk *start = NULL, *end = NULL;
    int i=1;

    /*Delete All Entries from memblk_list_t*/
    start = free_kuseg_mem_blks();
    memset(mem_map,0,sizeof(*mem_map));

    mem_map->map[0].addr = 0x00100000;
    mem_map->map[0].size = 0x0c000000;
    mem_map->map[0].type = BOOT_MEM_RAM;

    /*Whole kuseg_mem list is empty.. Add all entries here.*/
    while(start){
        add_memblk(start->phys,start->size);
        if(i<32){
            mem_map->map[i].addr = start->phys;
            mem_map->map[i].size = start->size;
            mem_map->map[i].type = BOOT_MEM_RAM;
            i++;
        }
        end = start;
        start = start->next;
        free(end);
    }
    mem_map->nr_map = i;
}
